%    -*- Erlang -*- 
%    File:	vector.erl  (~jb/projects/erl_tracer/vector.erl)
%    Author:	Johan Bevemyr
%    Created:	Tue Sep 12 21:52:09 1995
%    Purpose:   Basic vector operations needed by ray-tracer.
%
 
-module(vector).
-export([vector_mult/2,vector_cross/2,vector_add/2,vector_sub/2,
	 vector_scale/2,vector_norm/1,vector_length/1]).
-compile(export_all).

-import(math,[sqrt/1]).


%
% A vector is represented as a tupple of three floats
%
%        {X,Y,Z}
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Scalar product of vectors.

vector_mult({U1,U2,U3},{V1,V2,V3}) ->
    V1*U1+V2*U2+V3*U3.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Cross product of vectors

vector_cross({U1,U2,U3},{V1,V2,V3}) ->
    {U2*V3-U3*V2,U3*V1-U1*V3,U1*V2-U2*V1}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Addition of vectors

vector_add({U1,U2,U3},{V1,V2,V3}) ->
    {U1+V1,U2+V2,U3+V3}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Vector subtraction

vector_sub({U1,U2,U3},{V1,V2,V3}) ->
    {U1-V1,U2-V2,U3-V3}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Vector scaling

vector_scale({U1,U2,U3}, S) ->
    {U1*S,U2*S,U3*S}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Normalising vector

vector_norm({U1,U2,U3}) ->
    S = 1/sqrt(U1*U1+U2*U2+U3*U3),
    {U1*S,U2*S,U3*S}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Vector length

vector_length(Vector) ->
    sqrt(vector_mult(Vector,Vector)).
