%% Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
%% File    : dict.erl
%% Author  : Robert Virding
%% Purpose : A simple dictionary interface.

-module(dict).

-export([new/0,is_key/2,dict_to_list/1]).
-export([fetch/2,find/2,fetch_keys/1,store/3,append/3,append_list/3,erase/2]).

%% new() -> Dictionary

new() ->
    [].

%% is_key(Key, Dictionary) -> Boolean

is_key(Key, [{Key,Val}|Dict]) -> true;
is_key(Key, [Pair|Dict]) ->
    is_key(Key, Dict);
is_key(Key, []) -> false.

%% dict_to_list(Dictionary) -> [{Key,Value}]

dict_to_list(Dict) -> Dict.

%% fetch(Key, Dictionary) -> Value

fetch(Key, [{K,Value}|D]) when Key > K ->
    fetch(Key, D);
fetch(Key, [{K,Value}|_]) when Key == K ->
    Value.

%% find(Key, Dictionary) -> {ok,Value} | error

find(Key, [{K,Value}|D]) when Key > K ->
    find(Key, D);
find(Key, [{K,Value}|_]) when Key == K ->
    {ok,Value};
find(Key, [{K,Value}|_]) when Key < K ->
    error;
find(Key, []) ->
    error.

%% fetch_keys(Dictionary) -> [Key]

fetch_keys([{Key,Val}|Dict]) ->
    [Key|fetch_keys(Dict)];
fetch_keys([]) ->
    [].

%% store(Key, Value, Dictionary) -> Dictionary'

store(Key, New, [{K,Old}|Dict]) when Key > K ->
    [{K,Old}|store(Key, New, Dict)];
store(Key, New, [{K,Old}|Dict]) when Key == K ->
    [{Key,New}|Dict];
store(Key, New, [{K,Old}|Dict]) when Key < K ->
    [{Key,New},{K,Old}|Dict];
store(Key, New, []) ->
	[{Key,New}].

%% append(Key, Value, Dictionary) -> Dictionary'

append(Key, New, [{K,Old}|Dict]) when Key > K ->
    [{K,Old}|append(Key, New, Dict)];
append(Key, New, [{K,Old}|Dict]) when Key == K ->
    [{Key,lists:append(Old, [New])}|Dict];
append(Key, New, [{K,Old}|Dict]) when Key < K ->
    [{Key,[New]},{K,Old}|Dict];
append(Key, New, []) ->
	[{Key,[New]}].

%% append_list(Key, ValueList, Dictionary) -> Dictionary'

append_list(Key, NewList, [{K,Old}|Dict]) when Key > K ->
    [{K,Old}|append_list(Key, NewList, Dict)];
append_list(Key, NewList, [{K,Old}|Dict]) when Key == K ->
    [{Key,lists:append(Old, NewList)}|Dict];
append_list(Key, NewList, [{K,Old}|Dict]) when Key < K ->
    [{Key,NewList},{K,Old}|Dict];
append_list(Key, NewList, []) ->
	[{Key,NewList}].

%% erase(Key, Dictionary) -> Dictionary'

erase(Key, [{K,Value}|Dict]) when Key > K ->
    [{K,Value}|erase(Key, Dict)];
erase(Key, [{K,Value}|Dict]) when Key == K ->
    Dict;
erase(Key, [{K,Value}|Dict]) when Key < K ->
    [{K,Value}|Dict];
erase(Key, []) ->
    [].
