%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% File     :   erlang.erl
%% Date	    :	15SEPT1993
%% Author   :   Claes Wikstrom
%% Purpose  :   System module
 
-module(erlang).

-export([apply/2,spawn/4,spawn_link/4,process_info/1]).
-export([crasher/5]).

apply({M,F},A) -> apply(M,F,A).

spawn(N,M,F,A) when N /= node() ->
    monitor_node(N, true),
    {net_kernel,N} ! {self(),spawn,M,F,A,group_leader()},
    receive
	{nodedown, N} -> R = spawn(erlang,crasher,[N,M,F,A,noconnection]);
	{spawn_reply, Pid} -> R = Pid
    end,
    monitor_node(N, false),
    R;
spawn(N,M,F,A) -> spawn(M,F,A).


spawn_link(N,M,F,A) when N /= node() ->
    monitor_node(N, true),
    {net_kernel,N} ! {self(),spawn_link,M,F,A,group_leader()},
    R = receive
	{nodedown, N} -> 
	    spawn_link(erlang,crasher,[N,M,F,A,noconnection]);
	{spawn_link_reply,Pid} -> 
	    Pid
    end,
    monitor_node(N, false),
    R;
spawn_link(N,M,F,A) -> spawn_link(M,F,A).

crasher(Node,Mod,Fun,Args,Reason) ->
    error_logger:error_msg('** Can not start ~w:~w,~w on ~w **~n',
			  [ Mod,Fun,Args,Node]),
    exit(Reason).



process_info(P) when pid(P) ->
    case process_info(P, registered_name) of
	undefined -> undefined;
	[] -> call_process_info(P, []);
	{X,Y} -> call_process_info(P,[{X,Y}])
    end.

call_process_info(P, Reg) ->
    Old = process_flag(pre_empt, false),
    L = [process_info(P, current_function),
	 process_info(P, initial_call),
	 process_info(P, status),
	 process_info(P, messages),
	 process_info(P, links),
	 process_info(P, dictionary),
	 process_info(P, trap_exit),
	 process_info(P, error_handler),
	 process_info(P, priority),
	 process_info(P, group_leader),
	 process_info(P, heap_size),
	 process_info(P, stack_size),
	 process_info(P, reductions) | Reg],
    process_flag(pre_empt, Old),
    L.
