%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File    : error_handler.erl
%% Author  : Joe Armstrong, Robert Virding, Magnus Froberg
%% Purpose : A simple error handler

-module(error_handler).

-export([undefined_function/3,undefined_global_name/2, stub_function/3]).
-import(error_logger,[error_msg/2]).


undefined_function(Module, Func, Args) ->
    case code:ensure_loaded(Module) of
	{module, Module} ->
	    case erlang:function_exported(Module, Func, length(Args)) of
		true ->
		    apply(Module, Func, Args);
		false ->
		    error_msg("** undefined function: ~w:~w~p **~n",
			       [Module,Func,Args]),
		    exit({undef,{Module,Func,Args}})
	    end;
	{module, Other} ->
	    error_msg("** autoload found module:~p in ~p.erl **~n",
		      [Other, Module]),
	    exit({undef,{Module,Func,Args}});
	{interpret, Module} ->
	    int:eval(Module, Func, Args);
	Other ->
	    error_msg("** cannot autoload: ~p ** ~n", [Module]),
	    exit({undef,{Module,Func,Args}})
    end.


undefined_global_name(Name, Message) ->
    error_msg("** undefined registered name: ~w **~n", [Name]),
    exit({badarg,{Name,Message}}).

stub_function(Mod, Func, Args) ->
    error_msg("** undefined local function: ~w:~w~w **~n",
			   [Mod,Func,Args]),
    exit({undef,{Mod,Func,Args}}).

    
	    




