-module(file).
-author('tobbe@serc.rmit.edu.au').
%% -------------------------------------------------------------------
%% Created:  30 Apr 1998 by tobbe@serc.rmit.edu.au
%% Function: The gerl-version of file.erl
%% -------------------------------------------------------------------
-vc('$Id: file.erl,v 1.8 2002/11/24 13:41:35 geoff Exp $ ').
-export([read_file/1,write_file/2,open/2,close/1]).

%%-export([get_cwd/0,set_cwd/1,delete/1,rename/2,make_dir/1,del_dir/1,
%%	 list_dir/1,file_info/1,read_file/1,write_file/2]).
%%-export([consult/1,path_consult/2,eval/1,path_eval/2]).
%%-export([path_open/3,open/2,close/1,position/2,truncate/1]).

pragma
#include <unistd.h>
#include <malloc.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include "erlib.h"
#include "typecheck.h"
#define NIL          0
#define READ         1
#define WRITE        2
#define READ_WRITE   3
#define BINARY       8
end pragma;
.

%% ----------------------------------------------------
%% Open a file.
%% NB: The returned file descriptor is not a the standard
%% Unix one. It contains information weather read data
%% should be regarded as binary or a list of bytes, i.e:
%% {ok, {Fd, BinaryFlag}} or throws an error.
%% Also, the file is always opened as a 'raw' file since
%% we (at the moment) don't have a centralized file server.

o_read()   -> 1.
o_write()  -> 2.
o_raw()    -> 4.
o_binary() -> 8.
%o_append() -> 16.

open(Fname, ModeList) ->
    open(Fname, ModeList, 0).

open(Fname, [read|T], Flags) ->
    open(Fname, T, Flags bor o_read());
open(Fname, [write|T], Flags) ->
    open(Fname, T, Flags bor o_write());
open(Fname, [raw|T], Flags) ->
    open(Fname, T, Flags bor o_raw());
open(Fname, [binary|T], Flags) ->
    open(Fname, T, Flags bor o_binary());
open(Fname, [], Flags) ->
    open_file(Fname, Flags bor o_raw()).

%% Either read or write has to be specified (or both)

open_file(Fname, Flags) when (Flags band 3) > 0 ->
pragma
  int fd, flags, mode=0644, rwmask=3;
  char *fn=NULL;

  fn = Elist_to_string(Fname, 1, "file2:open_file/2");

  TC(Flags, LT_number, 2, "file2:open_file/2");

if ((Flags.u.number & rwmask) == READ)
    flags = O_RDONLY;
  else if ((Flags.u.number & rwmask) == WRITE)
    flags = O_WRONLY | O_CREAT | O_TRUNC;
  else if ((Flags.u.number & rwmask) == READ_WRITE)
      flags = O_RDWR | O_CREAT | O_APPEND;
  else {
     return Eerror("missing flags");
  }
	
  if ((fd = open(fn, flags, mode)) < 0) 
  {
     free(fn);
     return Eerror(strerror(errno));
  }
  else if (Flags.u.number & BINARY)
  {
    Val _t1;
    free(fn);
#ifdef _ALLOCS_H
    tuplealloc(_t1,2);
    return Eok(Emake_tuple(_t1, EValint(fd), EValint(BINARY)));
#else
    return Eok(Emake_tuple(2, EValint(fd), EValint(BINARY)));
#endif
  }
  else
  {
    Val _t2;
    free(fn);
#ifdef _ALLOCS_H
    tuplealloc(_t2, 2);
    return Eok(Emake_tuple(_t2, EValint(fd), EValint(NIL)));
#else
    return Eok(Emake_tuple(2, EValint(fd), EValint(NIL)));
#endif
  }
end pragma;
.

%% ------------
%% Close a file

close({Fd,FdInfo}) ->
    close_file(Fd).

close_file(Fd) ->
pragma
  close(Fd.u.number);
  return Eokay();
end pragma;
.

read({Fd,Info}, Nbytes) ->
    read(Fd, Info, Nbytes).

read(Fd, Info, Nbytes) ->
pragma
  int n,fd,info,size;
  char *cp=NULL;

  fd   = Fd.u.number;
  info = Info.u.number;
  size = Nbytes.u.number;

  fflush(stdout);

  cp = (char *) malloc(size);
  if ((n = read(fd, cp, size)) < 0) {
    char *emsg=strerror(errno);				  
    close(fd);
    free(cp);				  
    perror("read");
    return Eerror(emsg);
  }
  // close(fd);

  if (info == BINARY) 
  {
    Val _xy;
    _xy = Emake_binary(cp, n);
    free(cp);
    return Eok(_xy);
  }
  else 
  {
    Val v;	
    //printf("file:read %d chars\n", n);
    v = Ebytes_to_list(cp, n);
    free(cp);	
    return Eok(v);
  }	
end pragma;
.

%% --------------------------------------------------
%% Read the specified file and return it as a binary.

read_file(Fname) ->
pragma
  int fd,size;
  struct stat s;
  char *cp=NULL,*fn=NULL;

  fn = Elist_to_string(Fname, 1, "file2:read_file/2");

  if (stat(fn, &s) != 0) 
     return Eerror(strerror(errno));
  else
     size = (int) s.st_size;

  if ((fd = open(fn, O_RDONLY)) < 0) 
     return Eerror(strerror(errno));
  else 
  {
     cp = (char *) malloc(size);
     if (read(fd, cp, size) < 0) 
     {
       char *emsg=strerror(errno);				  
       close(fd);
       free(cp);				  
       /* printf("failed to read file: %s\n", fn); */
       return Eerror(emsg);
     }
     else 
     {
       close(fd);
       /* printf("file read: returning binary of size %d\n", size); */
       return Eok(Emake_binary(cp, size));
     }	   
  }	
end pragma;
.

%% -----------------------------------
%% Take a binary and write it to file.

write_file(Fname, Bin) ->
pragma
  int fd, mode=0644;
  char *fn=NULL;

  fn = Elist_to_string(Fname, 1, "file2:write_file/2");

  if ((fd = open(fn, O_WRONLY | O_CREAT, mode)) < 0) 
     return Eerror(strerror(errno));
  else { 
     if (write(fd, Bin.u.binary->b, Bin.u.binary->size) < 0) {
       char *emsg=strerror(errno);				  
       close(fd);
       return Eerror(emsg);
     }
     else {
       close(fd);
       return Eokay();
     }
  }
end pragma;
.

append_file(Fd, Bin) when binary(Bin) ->
pragma
    int fd;

    fd = Fd.u.number;
    if (fd < 0) return Eerror("Illegal file descriptor");

    lseek(fd, 0, SEEK_END);
    if (write(fd, Bin.u.binary->b, Bin.u.binary->size) < 0) 
    {
       char *emsg=strerror(errno);				  
       return Eerror(emsg);
    }
    else 
    {
       return Eokay();
    }
end pragma;
;

append_file(Fd, At) when atom(At) ->
pragma
    int fd;

    fd = Fd.u.number;
    if (fd < 0) return Eerror("Illegal file descriptor");

    lseek(fd, 0, SEEK_END);
    if (write(fd, At.u.atom, strlen(At.u.atom)) < 0) 
    {
       char *emsg=strerror(errno);				  
       return Eerror(emsg);
    }
    else 
    {
       return Eokay();
    }
end pragma;
;

append_file(Fd, At) when list(At) ->
pragma
    int fd;
    int i;
    char * buf;

    fd = Fd.u.number;
    if (fd < 0) return Eerror("Illegal file descriptor");

    lseek(fd, 0, SEEK_END);
    buf = (char *) malloc(At.u.list->length * sizeof(char)+1);
    for (i = 0; i < At.u.list->length; i++)
    {
        buf[i] = Eindex_list(At, i).u.number;
    }

    if (write(fd, buf, At.u.list->length) < 0) 
    {
       char *emsg=strerror(errno);				  
       free(buf);
       return Eerror(emsg);
    }
    else 
    {
       free(buf);
       return Eokay();
    }
end pragma;
.


