%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : gen_tcp.erl
%%% Author  : Tony Rogvall <tony@erix.ericsson.se>
%%% Purpose : Generic TCP module
%%% Created : 27 Apr 1997 by Tony Rogvall <tony@erix.ericsson.se>

%%
%% Changed to supprt Gerl (Geoff's Erlang Compiler)
%% Embedded C code directly.
%%

-module(gen_tcp).
-author('tony@erix.ericsson.se').

-export([connect/3, connect/4, listen/2, accept/1, accept/2, close/1, 
        shutdown/2]).
-export([send/2, recv/2]). %, recv/3]).
-export([controlling_process/2]).

-include("inet_int.hrl").

pragma
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
end pragma;
.


%%
%% Connect a socket
%%
connect(Address, Port, Opts) -> 
    connect(Address,Port,Opts,infinity).

connect(Address, Port, Opts, Time) ->
%    Mod = mod(Opts),
%    case Mod:getaddr(Address) of
%	{ok,IP} ->
%	    case  Mod:getserv(Port) of
%		{ok,TP} -> Mod:connect(IP, TP, Opts, Time);
%		{error,einval} -> exit(badarg);
%		Error -> Error
%	    end;
%	{error,einval} -> exit(badarg);
%	Error -> Error
    throw('unimplemented connect/4')
.

%%
%% Listen on a tcp port
%% Returns a tuple { 'ok', Sock }
%%      or { 'EXIT', Reason }
%%
listen(Port, Opts) ->
pragma
    int s, ret;
    struct sockaddr_in sin;
    struct hostent *hp;
    int             tmp;
    char            host_name[256];
    Val _xxx, tp;

    if (gethostname(host_name, sizeof host_name) == -1) 
        Ethrowv("gethostname");

    hp = gethostbyname(host_name);
    if (hp == NULL) Ethrowv("gethostbyname");

    memset((char *) &sin, '\0', sizeof sin);
    memcpy((char *) &sin.sin_addr, hp->h_addr, hp->h_length);
    sin.sin_port = htons(Port.u.number);
    sin.sin_family = hp->h_addrtype;
    sin.sin_addr.s_addr = INADDR_ANY;

    s = socket(PF_INET, SOCK_STREAM, 0);

    if (bind(s, (struct sockaddr *) & sin, sizeof sin) == -1) 
    {
        if (errno == EADDRINUSE) 
            Ethrowv("eaddrinuse");
        else 
            Ethrowv("bind"); 
    }

    ret = listen(s, 128);
    // if (ret == -1) Ethrowv("listen");

    // printf("listen end\n");
    
#ifdef _ALLOCS_H
    return (tuplealloc(tp,2), Emake_tuple(tp, EValatom(_xxx, "ok"), EValint(s)));
#else
    return (Emake_tuple(2, EValatom("ok"), EValint(s)));
#endif
end pragma;
.

%    Mod = mod(Opts),
%    case Mod:getserv(Port) of
%	{ok,TP} ->
%	    Mod:listen(TP, Opts);
%	{error,einval} ->
%	    exit(badarg);
%	Other -> Other
%    end.

%%
%% Generic tcp accept
%%
%accept(S) when record(S, socket) ->
%    call(S, {accept, infinity}).
%
%accept(S, infinity) when record(S, socket) ->
%    call(S, {accept, infinity});
%accept(S, Time) when record(S, socket), integer(Time), Time >= 0 ->
%    call(S, {accept, Time}).

accept(Sock) ->
pragma
    struct sockaddr_in addr;
    int             length, new_socket;
    Val _xxx, tp;

    length = sizeof addr;
    new_socket = accept(Sock.u.number, (struct sockaddr *) &addr, &length);
    if (new_socket != -1) 
    {
#ifdef _ALLOCS_H
        return (tuplealloc(tp,2), Emake_tuple(tp, EValatom(_xxx, "ok"), EValint(new_socket)));
#else
        return (Emake_tuple(2, EValatom("ok"), EValint(new_socket)));
#endif
    }

    return Ethrowv("accept"); 
end pragma;
.


%%
%% Close
%%
%close(S) when record(S, socket) -> 
%    receive
%	{tcp_closed, S} -> ok
%    after 0 ->
%	    inet:close(S)
%    end.

close(Sock) ->
pragma
    // printf("close/1\n");
    return EValint(close(Sock.u.number));
end pragma;
.


%%
%% Shutdown
%%
shutdown(Sock, How) ->
pragma
    // printf("close/1\n");
    return EValint(shutdown(Sock.u.number,How.u.number));
end pragma;
.


%%
%% Send
%%
%send(S, Packet) when record(S, socket) ->
%    call(S, {send, Packet}).

send(S, Packet) ->
pragma
    return EValint(send(S.u.number, Packet.u.binary->b, Packet.u.binary->size, 0));
end pragma;
.


%%
%% Receive data from a socket (passive mode)
%%
%recv(S, Length) when  record(S,socket), integer(Length) ->
%    call(S, {recv, Length, infinity}).
%
%recv(S, Length, infinity) when record(S,socket), integer(Length) ->
%    call(S, {recv, Length, infinity});
%recv(S, Length, Time) when record(S,socket), integer(Length),
%                                integer(Time),Time >= 0 ->
%    call(S, {recv, Length, Time}).

recv(S, Length) ->
pragma
    char buf[2048];
    int num;
    Val tp,_xxx;
    
    // printf("in gen_tcp:recv/2\n");
    fflush(stdout);

    num = recv(S.u.number, buf, sizeof buf, 0);

    if (num == -1)
    {
        perror("recv");
#ifdef _ALLOCS_H
        return (tuplealloc(tp,2), Emake_tuple(tp, EValatom(_xxx, "EXIT"), EValint(num)));
#else
        return (Emake_tuple(2, EValatom("EXIT"), EValint(num)));
#endif
    }

    if (Length.u.number > num) num = Length.u.number;

#ifdef _ALLOCS_H
    return (tuplealloc(tp,2), Emake_tuple(tp, EValatom(_xxx, "ok"), EValmbinary(buf, num)));
#else
    return (Emake_tuple(2, EValatom("ok"), EValmbinary(buf, num)));
#endif
end pragma;
.


%%
%% Set controlling process
%%
%controlling_process(S, NewOwner) when record(S,socket),pid(NewOwner) ->
%    case call(S, {set_owner, NewOwner}) of
%	ok -> 
%	    sync_input(S, NewOwner),
%	    S#socket.pid ! {commit_owner, NewOwner},
%	    receive
%		{owner, NewOwner} -> ok
%	    end;
%	Error -> Error
%    throw('controlling_process/2 unimplemented')
%    end.

sync_input(S, Owner) ->
    receive
	{tcp, S, Data} ->
	    Owner ! {tcp, S, Data},
	    sync_input(S, Owner);
	{tcp_closed, S} ->
	    Owner ! {tcp_closed, S},
	    sync_input(S, Owner)
    after 0 -> 
	    ok
    end.

%% call socket server
%call(S, Request) ->
%    Tag = make_ref(),
%    Pid = S#socket.pid,
%    Pid ! {call, self(), Tag, Request},
%    receive
%	{Tag, Reply} -> Reply;
%	{'EXIT', Pid, Reason} ->
%	    {error, closed};
%	{tcp_closed, S} ->
%	    {error, closed}
%    throw('call/2 unimplemented')
%    end.

%% Get the tcp_module
%mod() -> inet_db:tcp_module().
%
%%% Get the tcp_module, but option tcp_module overrides
%mod(Opts) ->
%    case lists:keysearch(tcp_module, 1, Opts) of
%	{value, {_, Mod}} -> Mod;
%	_ -> inet_db:tcp_module()
%    end.
