%%
%% Modules relied upon: string, gen_tcp, lists, regexp
%%

-module(httpd).
-export([start/1,server/1,reqhandler/1]).

start(Port) ->
   %spawn(httpd, server, [Port]).
    server(Port).

server(Port) ->
   {ok, Lsock} = gen_tcp:listen(Port, [binary, {packet, 0}, 
      {active, false}]),
   serverloop(Lsock).

serverloop(Lsock) ->
   {ok, Sock} = gen_tcp:accept(Lsock),
   spawn(httpd,reqhandler,[Sock]),
   serverloop(Lsock).

reqhandler(Sock) ->
   ReqStr = getreq(Sock),
   [FirstArg, SecondArg | Tail] = string:tokens(ReqStr, " \n\t"),
   if 
      FirstArg =/= "GET" ->
         io:puts('FirstArg =/= GET\n'),
         gen_tcp:send(Sock, list_to_binary("<head><title>400 Bad Request</title></head>\r\n<body><h1>400 Bad Request</h1>\r\n</body>"));
      true ->
         %{ok, BaseName, _} = string:re_gsub(SecondArg, "/$|^$", "/index.html"),
         %{ok, File, _} = string:re_sub(BaseName, "^/+", ""),
         %sendfile(Sock, File)
         % io:puts('sending file back\n'),
         sendfile(Sock, SecondArg)
   end,
   gen_tcp:close(Sock).

getreq(Sock) ->
   getreq(Sock, []).

getreq(Sock, OrigStr) ->
   {ok, Pack} = gen_tcp:recv(Sock, 0),
   RecStr = binary_to_list(Pack),
   NewStr = lists:append(OrigStr, RecStr),
   Pos = string:index(NewStr, "\r\n"),
   if 
      Pos =/= 0 ->
         string:substr(NewStr, 1, Pos-1);
      true ->
         getreq(Sock, NewStr)
   end.

sendfile(Sock, Filename) ->
   case file2:read_file(Filename) of
      {ok, Binary} ->
         % io:puts('got binary representation of file\n'),
         gen_tcp:send(Sock, Binary);
      _ ->
         % io:puts('unable to read file\n'),
         gen_tcp:send(Sock, list_to_binary("<head><title>404 File Not Found</title></head>\r\n<body><h1>404 File Not Found</h1>\r\nThe requested URL was not found on this server.<p>\r\n</body>"))
   end.
