%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose:   Basic startup function boot/1
%% Very stupid init:boot() that does nothing but what it's told
%% Very general however ....

-module(init).
-export([get_args/0,get_flag/1,get_flags/0,boot/1,start_em/1]).

get_args() ->
    request(get_args).

get_flag(F) ->
    request({get_flag,F}).

get_flags() ->
    request(get_flags).

request(Req) ->
    init ! {self(),Req},
    receive 
	{init,Rep} -> 
	    Rep
    end.

boot(BootArgs) ->
    register(init, self()),
    process_flag(trap_exit, true),
    {Start,Flags,Args} = parse_boot_args(BootArgs),
    Loaded = load_files(Flags),
    spawn(init,start_em,[Start]), %% So the servers can get init's flags
    garbage_collect(), %% For the binaries :-(
    loop(Start,[{'-loaded', Loaded} | Flags],Args).

loop(Start,Flags,Args) ->
    receive
	{From,get_args} -> 
	    From ! {init,Args};
	{From,{get_flag,F}} ->
	    case search(F,Flags) of
		false -> 
		    From ! {init,notfound};
		{value,V} ->
		    From ! {init,V}
	    end;
	{From,get_flags} -> 
	    From ! {init,Flags};
	X ->
	    error_logger ! {init,X}
    end,
    loop(Start,Flags,Args).

%% Use apply here instead of spawn to ensure syncronicity for
%% those servers that wish to have it so.
%% Disadvantage: any thing started with -s that does not
%% eventually spawn will hang the startup routine.

start_em([S|Tail]) ->
    case whereis(user) of
	undefined -> 
	    ok;
	P when pid(P) ->			%Let's set the group_leader()
	    erlang:group_leader(P, self())
    end,
    start_it(S),
    start_em(Tail);
start_em([]) -> ok.

start_it([]) -> ok;
start_it([M]) ->
    catch apply(M,start,[]);
start_it([M,F]) ->
    catch apply(M,F,[]);
start_it([M,F|Args]) ->
    catch apply(M,F,[Args]).			%One variable len arg only

%% Parse the command line arguments and extract things to start, flags
%% and other arguments. We keep the relative of the groups.

parse_boot_args(Args) ->
    parse_boot_args(Args, [], [], []).

parse_boot_args([B|Bs], Ss, Fs, As) ->
    case check(B) of
	start_arg ->
	    {S,Rest} = get_args(Bs, []),
	    parse_boot_args(Rest, [S|Ss], Fs, As);
	flag ->
	    {F,Rest} = get_args(Bs, []),
	    parse_boot_args(Rest, Ss, [list_to_tuple([B|F])|Fs], As);
	arg ->
	    parse_boot_args(Bs, Ss, Fs, [B|As]);
	end_args ->
	    parse_boot_args(Bs, Ss, Fs, As)
    end;
parse_boot_args([], Start, Flags, Args) ->
    {reverse(Start),reverse(Flags),reverse(Args)}.

check('-s') -> start_arg;
check('--') -> end_args;
check(X) when atom(X) ->
    case atom_to_list(X) of
	[$-|Rest] -> flag;
	Chars     -> arg			%Even empty atoms
    end;
check(X) -> arg.				%This should never occur

get_args([B|Bs], As) ->
    case check(B) of
	start_arg -> {reverse(As), [B|Bs]};
	end_args -> {reverse(As), Bs};
	flag -> {reverse(As), [B|Bs]};
	arg ->
	    get_args(Bs, [B|As])
    end;
get_args([], As) -> {reverse(As),[]}.

%% We let init load the files that are necessary for
%% the boot process to continue. Have a load path of the current
%% directory and the standard library.

load_files(Flags) ->
    case {search('-root', Flags), search('-load', Flags)} of
	{{value, {'-root', Root}}, {value,Load0}} ->
	    Load = tl(tuple_to_list(Load0)),
	    P = open_port({spawn, 'binary_filer'}, [binary]),
	    load_files(P, ["./",concat([Root,"/lib/std/ebin/"])], Load);
	_ ->
	    []
    end.

load_files(Port, Path, [Mod|Mods]) ->
    File = load_file(Port, Path, Mod),
    [{Mod,File}|load_files(Port, Path, Mods)];
load_files(Port, Path, []) ->
    Port ! {self(), close},
    receive
	{Port, closed} -> ok
    end,
    receive
	{'EXIT', Port, _} -> ok
    end,
    [].

load_file(Port, [P|Path], Mod) ->
    File = concat([P,Mod,extension()]),
    Port ! {self(),{command,[$f|File]}},
    receive
	{Port,{data,Bin}} ->
	    case binary_to_list(Bin, 1, 1) of
		[$y] ->
		    {_,Code} = split_binary(Bin, 1),
		    case erlang:load_module(Mod, Code) of
			{module,Mod} -> File;
			Other ->
			    load_file(Port, Path, Mod)
		    end;
		Other ->
		    load_file(Port, Path, Mod)
	    end
    end;
load_file(Port, [], Mod) ->
    erlang:display({'** error when loading in boot **',Mod}),
    halt().

concat([A|T]) when atom(A) ->			%Atom
    append(atom_to_list(A), concat(T));
concat([C|T]) when C >= 0, C =< 255 ->
    [C|concat(T)];
concat([S|T]) ->				%String
    append(S, concat(T));
concat([]) ->
    [].

append([H|T], Z) ->
    [H|append(T, Z)];
append([], X) -> X.

reverse(X) ->
    reverse(X, []).

reverse([H|T], Y) ->
    reverse(T, [H|Y]);
reverse([], X) -> X.
			
search(Key, [H|T]) when tuple(H), element(1, H) == Key ->
    {value, H};
search(Key, [_|T]) -> search(Key, T);
search(Key, []) -> false.

extension() -> 
    case erlang:info(machine) of
        "JAM" -> ".jam";
        "VEE" -> ".vee";
        "BEAM" -> ".beam"
    end.
