%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : io.erl
%% Author  : Robert Virding
%% Purpose : Standard i/o server interface functions.

%% Modifications:
%%   930319 -- Changes code to add io groups
%%   930323 -- $\n changed to $\r in get_line

-module(io).

-export([put_chars/1,put_chars/2,nl/0,nl/1,
	 get_chars/2,get_chars/3,get_line/1,get_line/2]).
-export([write/1,write/2,read/1,read/2]).
-export([fwrite/1,fwrite/2,fwrite/3,fread/2,fread/3,
	 format/1,format/2,format/3]).
-export([scan_erl_seq/1,scan_erl_seq/2,scan_erl_seq/3,
	 scan_erl_form/1,scan_erl_form/2,scan_erl_form/3,
	 parse_erl_seq/1,parse_erl_seq/2, parse_erl_seq/3,
	 parse_erl_form/1,parse_erl_form/2,parse_erl_form/3]).
-export([request/1,request/2,requests/1,requests/2]).

-export([parse_exprs/2]).

parse_exprs(X,Y) -> parse_erl_seq(X,Y).

%%pragma
%%#include <stdio.h>
%%end pragma;

%%
%% Temporary debugging stuff
%%

puts(Chars) ->
pragma
    int i;
    // printf("#%s#\n", Chars.u.atom);
    i = puts(Chars.u.atom);
    fflush(stdout);
    return EValint(i);
end pragma;
.

puti(X) ->
pragma
    int i;
    i = printf("%d\n", X.u.number);
    fflush(stdout);
    return X;
end pragma;
.


%%
%% User interface.
%%

%% Writing and reading characters.

put_chars(Chars) ->
    put_chars(default_output(), Chars).

%    request(Io, {put_chars,Chars}).
put_chars(Io, Chars) ->
pragma 
    return EValint(fputs(Elist_to_string(Chars, 2, "io:put_chars/2"), (FILE *)Io.u.any));
end pragma;
.

nl() ->
    nl(default_output()).

%    request(Io, {put_chars,io_lib:nl()}).
nl(Io) ->
    put_chars(io_lib:nl()).

get_chars(Prompt, N) ->
    get_chars(default_input(), Prompt, N).

get_chars(Io, Prompt, N) ->
    throw('Unimplemented io:get_chars/2\n'),
    request(Io, {get_chars,Prompt,N}).

get_line(Prompt) ->
    get_line(default_input(), Prompt).

get_line(Io, Prompt) ->
    throw('Unimplemented io:get_line/2\n'),
    request(Io, {get_line,Prompt}).

%% Writing and reading Erlang terms.

write(Term) ->
    write(default_output(), Term).

write(Io, Term) ->
    throw('Unimplemented io:write/2\n'),
    request(Io, {write,Term}).

read(Prompt) ->
    read(default_input(), Prompt).

read(Io, Prompt) ->
    throw('Unimplemented io:read/2\n'),
    case request(Io, {get_until,Prompt,erl_scan,tokens,[1]}) of
	{ok,Toks,EndLine} ->
	    erl_parse:parse_term(Toks);
	{error,E,EndLine} ->
	    {error,E};
	{eof,EndLine} ->
	    eof;
	Other ->
	    Other
    end.

%% Formatted writing and reading.

fwrite(Format) ->
    fwrite(Format, []).

fwrite(Format, Args) ->
    fwrite(default_output(), Format, Args).

fwrite(Io, Format, Args) ->
    throw('Unimplemented io:fwrite/3\n'),
    request(Io, {fwrite,Format,Args}).

fread(Prompt, Format) ->
    fread(default_input(), Prompt, Format).

fread(Io, Prompt, Format) ->
    throw('Unimplemented io:fread/3\n'),
    request(Io, {fread,Prompt,Format}).

format(Format) ->
    format(Format, []).

format(Format, Args) ->
    format(default_output(), Format, Args).

%    request(Io, {format,Format,Args}).
format(Io, Format, Args) ->
%    put_chars(Format)
pragma
#if 1
#include <stdio.h>

    int i, x, z, c = 0;
    char buf[32];
    Val arg;

    if (!(Format.tag & LT_list))
        Ethrowv("Illegal type %d (arg 2) to io:format/3\n",
            Format.tag);

    for (i = 0; i < Format.u.list->length; i++)
    {
        x = Eindex_list(Format, i).u.number;
        switch (x)
        {
            case '~':
                // formatting character
                i++;
                z = Eindex_list(Format, i).u.number;
                arg = Eindex_list(Args, c++);

                switch (z)
                {
                    case '~':
                        putchar('~');
                        break;
                    case 'c':
                        break;
                    case 'f':
                    case 'e':
                    case 'g':
                        // floating point stuff
                        sprintf(buf, "%f", arg.u.real);
                        fputs(buf, (FILE *)Io.u.any);
                        break;
                    case 's':
                        // string syntax
                        break;
                    case 'w':
                        // write standard with syntax
                        EPrintvalue((FILE *)Io.u.any, arg);
                        break;
                    case 'p':
                        // new line separate argument
                        EPrintvalue((FILE *)Io.u.any, arg);
                        break;
                    case 'n':
                        putchar(10);
                        break;
                    default:
                        putchar(x);
                }
                break;
            default:
                putchar(x);

        }
    }
    return EValcopy(Format);

#endif

end pragma;
    .

%% Scanning Erlang code.

scan_erl_seq(Prompt) ->
    scan_erl_seq(default_input(), Prompt, 1).

scan_erl_seq(Io, Prompt) ->
    scan_erl_seq(Io, Prompt, 1).

scan_erl_seq(Io, Prompt, Pos0) ->
    throw('Unimplemented io:scan_erl_seq/3\n'),
    request(Io, {get_until,Prompt,erl_scan,tokens,[Pos0]}).

scan_erl_form(Prompt) ->
    scan_erl_form(default_input(), Prompt, 1).

scan_erl_form(Io, Prompt) ->
    scan_erl_form(Io, Prompt, 1).

scan_erl_form(Io, Prompt, Pos0) ->
    throw('Unimplemented io:scan_erl_form/3\n'),
    request(Io, {get_until,Prompt,erl_scan,tokens,[Pos0]}).

%% Parsing Erlang code.

parse_erl_seq(Prompt) ->
    parse_erl_seq(default_input(), Prompt, 1).

parse_erl_seq(Io, Prompt) ->
    parse_erl_seq(Io, Prompt, 1).

parse_erl_seq(Io, Prompt, Pos0) ->
    throw('Unimplemented io:parse_erl_seq/3\n'),
    case request(Io, {get_until,Prompt,erl_scan,tokens,[Pos0]}) of
	{ok,Toks,EndPos} ->
	    case erl_parse:parse_exprs(Toks) of
		{ok,Exprs} -> {ok,Exprs,EndPos};
		{error,E} -> {error,E,EndPos}
	    end;
	Other ->
	    Other
    end.

parse_erl_form(Prompt) ->
    parse_erl_form(default_input(), Prompt, 1).

parse_erl_form(Io, Prompt) ->
    parse_erl_form(Io, Prompt, 1).

parse_erl_form(Io, Prompt, Pos0) ->
    throw('Unimplemented io:parse_erl_form/3\n'),
    case request(Io, {get_until,Prompt,erl_scan,tokens,[Pos0]}) of
	{ok,Toks,EndPos} ->
	    case erl_parse:parse_form(Toks) of
		{ok,Exprs} -> {ok,Exprs,EndPos};
		{error,E} -> {error,E,EndPos}
	    end;
	Other ->
	    Other
    end.

%% Miscellaneous functions.

request(Request) ->
    request(default_output(), Request).

request(Io, Request) ->
    io_request(Io, io_request(Request)),
    wait_io_reply(Io).

requests(Requests) ->				%Requests as atomic action
    requests(default_output(), Requests).

requests(Io, Requests) ->			%Requests as atomic action
    io_request(Io, {requests,io_requests(Requests)}),
    wait_io_reply(Io).

%    group_leader().
default_input() ->
pragma
    Val ret;
    ret.u.any = (void *) stdin;
    ret.tag = LT_atom;
    return ret;
end pragma;
.

%    group_leader().
default_output() ->
pragma
    Val ret;
    ret.u.any = (void *) stdout;
    ret.tag = LT_atom;
    return ret;
end pragma;
.

io_request(standard_io, Request) ->
    group_leader() ! {io_request,self(),standard_io,Request};
io_request(Io, Request) ->
    case catch Io ! {io_request,self(),Io,Request} of
	{'EXIT', R} -> {error, arguments};
	X -> X
    end.

wait_io_reply(From) ->
    receive
	{io_reply,From,Reply} ->
	    Reply;
	{'EXIT',From,Reason} ->			%In case we are trapping exits
	    {error,terminated}
   end.

%% io_requests(Requests)
%%  Transform requests into correct i/o server messages. Only handle the
%%  one we KNOW must be changed, others, including incorrect ones, are
%%  passed straight through. Perform a flatten on the request list.

io_requests(Rs) ->
    io_requests(Rs, [], []).

io_requests([{requests,Rs1}|Rs], Cont, Tail) ->
    io_requests(Rs1, [Rs|Cont], Tail);
io_requests([R|Rs], Cont, Tail) ->
    [io_request(R)|io_requests(Rs, Cont, Tail)];
io_requests([], [Rs|Cont], Tail) ->
    io_requests(Rs, Cont, Tail);
io_requests([], [], Tail) ->
    [].

io_request({write,Term}) ->
    {put_chars,io_lib,write,[Term]};
io_request({format,Format,Args}) ->
    {put_chars,io_lib,format,[Format,Args]};
io_request({fwrite,Format,Args}) ->
    {put_chars,io_lib,fwrite,[Format,Args]};
io_request(nl) ->
    {put_chars,io_lib:nl()};
io_request({get_chars,Prompt,N}) ->
    {get_until,Prompt,io_lib,collect_chars,[N]};
io_request({get_line,Prompt}) ->
    {get_until,Prompt,io_lib,collect_line,[]};
io_request({fread,Prompt,Format}) ->
    {get_until,Prompt,io_lib,fread,[Format]};
io_request(R) ->				%Pass this straight through
    R.
