
-module(list).
-export([insert/2,insertc/3,count/2,delete/2,deleten/3,addnhead/3,addntail/3,
rep/2, append/2,match/2,addtoalllist/2,insertallplaces/2,perm/1,mapcarg/3,
tail/1,progtail/1,diff/2,testin/2]).

% simple insert routine
insert([],I) ->
	[I];
insert([H|[]],I) ->
	if 
		H < I -> [H,I];
		true -> [I,H]
	end;
insert([H|T],I) ->
	if
		I < H -> [I,H|T];
		true -> [H|insert(T, I)]
	end.

% More interesting insert routine which shifts the test routine into a 
% function allowing for polymorphic functions
insertc([],I,Genfn) ->
	[I];
insertc([H|[]],I,Genfn) ->
	X = apply(Genfn, [H, I]),
	if 
		X == 0 -> [H,I];
		true -> [I,H]
	end;
insertc([H|T],I,Genfn) ->
	X = apply(Genfn, [I, H]),
	if
		X == 0 -> [I,H|T];
		true -> [H|insert(T, I)]
	end.

% Count the number of Items found in List and return the number

count([H|T],Item) ->
	if 
		H == Item -> count(T,Item) + 1;
		true ->  count(T,Item)
	end;
count([], Item) ->
	0.

% Delete first matching item from list

delete([H|T],Item) ->
	if 
		H == Item -> T;
		true -> [H|delete(T,Item)]
	end;
delete([],Item) ->
	[].

% Delete first n matching items from list

deleten(List,Item,0) ->
	List;
deleten([H|T],Item,1) ->
	if 
		H == Item -> T;
		true -> [H|deleten(T,Item,1)]
	end;
deleten([H|T],Item,N) ->
	if 
		H == Item -> deleten(T,Item,N-1);
		true -> [H|deleten(T,Item,N)]
	end;
deleten([],Item,N) ->
	[].

% Add n items to the head of the list

addnhead(List, Item, 0) ->
	List;
addnhead(List, Item, N) ->
	[Item | addnhead(List, Item, N-1)]. 

% Add n items to the tail of the list

addntail(List, Item, N) ->
	T = rep(Item,N),
	append(List, T).

% Make a list composed of N lots of items

rep(Item, 0) ->
	[];
rep(Item, N) ->
	[Item | rep(Item, N-1)].

% Append

append([H|T], List) ->
	[H|append(T, List)];
append([], List) ->
	List.

% Simple check to see if a list contains an item
match([],Item) ->
	0;
match([H|T],Item) ->
	if 
		H == Item -> 1;
		true -> match(T,Item)
	end.

% add item to all lists contained in list [H|T] at front of list

addtoalllist([H|T], Item) ->
	[[Item|H]|addtoalllist(T, Item)];
addtoalllist([],Item) ->
	[].

% Insert Item into all places in the list

insertallplaces(L, Item) ->
	iap([],L,Item).

iap(L, [H|T], Item) -> 
	X = append(L,[H]),
	Y = append(append(L,[Item]),[H|T]),
	[Y | iap(X,T,Item)];
iap(L, [], Item) ->
	[append(L,[Item])].

% Perm takes a list and produces a list of lists of 
% all the permuted elements of that list

perm(L) ->
	perm([],L).

perm(R,[A]) ->
	[[A]];
perm(R, [H|T]) ->
	L = append(R,[H]),
	mapcarg({list,insertallplaces},perm(L,T),H).

mapcarg({Module, Func}, [H|T],Item) ->
	append(apply({Module,Func},[H,Item]), mapcarg({Module, Func}, T, Item));
mapcarg({Module, Func}, [], Item) ->
	[].

% Find the last element of a list

tail([H]) ->
	H;
tail([H|T]) ->
	tail(T).

% Return a tuple consisting of the last element of a list and 
% the remainder of the list

progtail(L) ->
	progtail([],L).

progtail(R,[H]) ->
	{H, R};
progtail(R,[H|T]) ->
	progtail(append(R,[H]), T).

% find the difference list between two lists
% First cut
%diff([Hs|Ts], [Hd|Td]) ->
%	if 
%		Hs == Hd -> diff(Ts, Td);
%		true -> [Hs | diff(Ts, [Hd|Td])]
%	end;
%diff(X, []) ->
%	X;
%diff([], _) ->
%	[].

diff([Hs|Ts], Ld) ->
	X = testin(Ld, Hs),
	if 
		X == 1 -> diff(Ts, delete(Ld,Hs));
		true -> [Hs | diff(Ts, Ld)]
	end;
diff(X, []) ->
	X;
%diff([], _) ->
%	[].

% test if item is in list
testin([], Item) ->
	0;
testin([Hs|Ts], Item) ->
	if 
		Hs == Item -> 1;
		true -> testin(Ts, Item)
	end.

