%% File    : random.erl
%% Author  : Joe Armstrong, Robert Virding
%% Purpose : Reasonable random number generator.

%%  The method is attributed to B.A. Wichmann and I.D.Hill
%%  See "An efficient and portable pseudo-random number generator",
%%  Journal of Applied Statistics. AS183. 1982. Also Byte March 1987.

-module(random).

-export([seed/0,seed/3,uniform/0,uniform/1]).

pragma
#include <time.h>
end pragma;
.

%% seed()
%%  Seed random number generation with default values

%%    seed(3172, 9814, 20125).
seed() ->
pragma
    srandom(time(0));
    return EValint(1);
end pragma;
.

%% seed(A1, A2, A3) 
%%  Seed random number generation 

seed(A1, A2, A3) ->
    put(random_seed, 
	{abs(A1) rem 30269, abs(A2) rem 30307, abs(A3) rem 30323}).

%% uniform()
%%  Returns a random float between 0 and 1.

uniform() ->
pragma
    return EValint(random());
end pragma;
.

%%    {A1, A2, A3} = get(random_seed),
%%   B1 = (A1*171) rem 30269,
%%  B2 = (A2*172) rem 30307,
%% B3 = (A3*170) rem 30323,
%%  put(random_seed, {B1,B2,B3}),
%%    R = A1/30269 + A2/30307 + A3/30323,
%%   R - trunc(R).

%% uniform(N) -> I
%%  Given an integer N >= 1, uniform(N) returns a random integer
%%  between 1 and N.
%%    trunc(uniform() * N) + 1.

uniform(N) ->
pragma
    return EValint(random() % N.u.number);
end pragma;
.
