%%
%% $Id: socket.erl,v 1.3 1998/09/07 08:59:18 geoff Exp $
%%
%% Socket object
%% Basically a container for all the C stuff
%%
%% $Source: /home/cvs/gerl/erlib/socket.erl,v $
%% $Revision: 1.3 $
%% $Date: 1998/09/07 08:59:18 $
%% $State: Exp $
%%
%% Author: Geoff Wong
%%

-module(socket).
-export([af_inet/0, af_unix/0, sock_stream/0, sock_dgram/0, sock_raw/0,
    create_socket/3, accept/1, bind/2, close/1, connect/2,  fcntl/3, ioctl/3,
    recv/1, send/2]).

pragma

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <strings.h>
#ifdef SunOS
#include <inet/tcp.h>
#else
#include <netinet/tcp.h>
#endif

#define MAX_BUFFER 65536

end pragma;
.


%% A bunch of constants.

af_inet() ->
pragma
    return EValint(AF_INET);
end pragma;
.

af_unix() ->
pragma
    return EValint(AF_UNIX);
end pragma;
.

sock_stream() ->
pragma
    return EValint(SOCK_STREAM);
end pragma;
.

sock_dgram() ->
pragma
    return EValint(SOCK_DGRAM);
end pragma;
.

sock_raw() ->
pragma
    return EValint(SOCK_RAW);
end pragma;
.

f_setfl() ->
pragma
    return F_SETFL;
end pragma;
.

o_ndelay() ->
pragma
    return O_NDELAY;
end pragma;
.




%%
%% Should we have a `record'?
%%
%% { socket_no, domain, protocol }
%%


%% create a socket of a certain name 
create_socket(Domain, Type, Protocol) ->
pragma
    int socket_no;

    socket_no = socket(Domain->u.number, Type->u.number, Protocol->u.number);

    return Emake_tuple(3, EValint(socket_no), Domain, Type);
end pragma;
    .

accept({Socket, Domain, Type}) ->
pragma
    struct sockaddr_in addr;
    int             length, new_socket;

    /* make sure we're bound to a socket */
    if (Socket->u.number == 0) 
        Ethrowv("Accepting on an unbound socket");

    /* can only accept on stream sockets */
    if (Type->u.number != SOCK_STREAM) 
        Ethrowv("Unable to accept on non-stream sockets");

    /* accept connections */
    length = sizeof addr;
    new_socket = accept(Socket->u.number,(struct sockaddr *) &addr, &length);

    if (new_socket > 0)
        return Emake_tuple(3, EValint(new_socket), Domain, Type);

    return EValint(0);
end pragma;
    .

%% Bind our socket to a name
%% Should use Domain properly?
bind({Socket,Domain,Type}, Port ) ->
pragma
    if (Domain->u.number == AF_INET)
    {
        struct sockaddr_in sin;
        memset(&sin, 0, sizeof sin);

        sin.sin_port = htons(Port->u.number);
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = INADDR_ANY;

        if (bind(Socket->u.number, (struct sockaddr *) &sin, sizeof sin) == -1)
        {
            return EValbool(0);
        }
        return EValbool(1);
    }
    else
    {
        struct sockaddr_un sin;
        int len;

        memset((char *) &sin, '\0', sizeof sin);
        sin.sun_family = AF_UNIX;
        strcpy(sin.sun_path, Port->u.atom);
        len = sizeof(sin) - sizeof(sin.sun_path) + strlen(sin.sun_path);

        if (bind(Socket->u.number, (struct sockaddr *) &sin, len) == -1)
        {
            return EValbool(0);
        }
        return EValbool(1);
    }
    Ethrowv("Unknown domain (%d) in bind\n", Domain->u.number);
end pragma;
    .

close({Socket,Domain,Type}) ->
pragma
    int ret;
    ret = close(Socket->u.number);
    if (ret == -1)
        Ethrowv("Error in socket_close/1: %d", errno);
    return EValbool(1);
end pragma;
    .

%% Connect to a socket or port
%%

%%connect({Socket,Domain,Type}, { Saddress, Port } ) ->
%%    throw('connect unimplemented')
%%    ;

%% Connect()s returns 0 on failure
%%            returns 1 on success
connect({Socket,Domain,Type}, { Iaddr, Port }) ->
pragma
    if (Domain->u.number == AF_INET)
    {
        struct sockaddr_in sin;

        sin.sin_port = htons(Port->u.number);
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = Iaddr->u.number;
        if (connect(Socket->u.number, (struct sockaddr *) &sin, sizeof sin) == -1)
        {
            Ethrowv("Failed to connect error: %d", errno);
        }
        return EValbool(1);
    }
    else if (Domain->u.number == AF_UNIX)
    {
        struct sockaddr_un server;
        int len;

        server.sun_family = AF_UNIX;
        strcpy(server.sun_path, Iaddr->u.atom);
        len = sizeof(server) - sizeof(server.sun_path) + strlen(server.sun_path);
        /* waiting for a connection */
        if (connect(Socket->u.number, (struct sockaddr *)&server, len) == -1)
        {
            Ethrowv("Failed to connect error: %d", errno);
        }
        return EValint(1);
    }
    return EValbool(0);
end pragma;
    ;

connect({Socket,Domain,Type}, Saddress) ->
    throw('connect unimplemented')
    .

connectto(Socket, { Saddress, Port }) ->
    throw('Connecto unimplemented')
    .

getsockopt(Socket, Level, Optname, Buflen) ->
    throw('Getsockopt unimplemented')
    .

%% Do ioctl/fcntl on a socket (very system dependent).
fcntl({Socket,_,_}, Cmd, Arg) ->
pragma
    int ret;
    ret = fcntl(Socket->u.number, Cmd->u.number, Arg->u.number);
    if (ret == -1)
        Ethrowv("fnctl error: %d", errno);
    return EValint(ret);
end pragma;
    .

ioctl({Socket,_,_}, Cmd, Arg) ->
pragma
    int ret;
    ret = ioctl(Socket->u.number, Cmd->u.number, Arg->u.number);
    if (ret == -1)
        Ethrowv("fnctl error: %d", errno);
    return EValint(ret);
end pragma;
    .

%% listen for a stream connection on our socket
listen({Socket,_,_}, Backlog) ->
pragma
    return listen(Socket->u.number, Backlog->u.number);                    
end pragma;
    .

%% Returns a list of items received.
recv({Socket,_,_}) ->
pragma
    int x;
    Val _tx1;
    char * buf, * ret;
    buf = (char *) malloc((MAX_BUFFER+1)*sizeof(char));
    x = recv(Socket->u.number, buf, MAX_BUFFER, 0);
    if (x == -1)
    {
        free(buf);
        Ethrowv("recv failed: %d", errno);
    }
    buf[x] = 0;
    ret = string_copy(buf);
    free(buf);
    return EValstring(_tx1, ret);
end pragma;
    .

%% Returns: { socketinfo, recvinfo }
%% Store the return info in the given socket
recvfrom({Socket,_,_}) ->
    throw("recvfrom unimplemented")
    .

send({Socket,_,_}, What) ->
pragma
    int x;

    x = send(Socket->u.number, What->u.atom, strlen(What->u.atom), 0);
    return EValint(x);                                                                  
end pragma;
    .

sendto({Socket, Domain, Type}, What) ->
pragma
    int x;
    struct sockaddr_in sin;
    int len = sizeof(struct sockaddr_in);

    sin.sin_family = Domain->u.number;
#if 0
    /* FIX THIS */
    sin.sin_addr.s_addr = Socket->int_address;
    sin.sin_port = htons(Socket->port);
#endif

    x = sendto(Socket->u.number, What->u.atom, strlen(What->u.atom), 
                0, (struct sockaddr *)&sin, len);
    return EValint(x);
end pragma;
    .

%%
%% Name: setblocking
%% Purpose: set a socket into blocking/nonblocking mode
%% Parameters: a socket 3-tuple, (bool) flags = false non-blocking mode
%%              true blocking mode
%%
setblocking({Socket,_,_}, Flags) ->
pragma
    if (Flags->u.number == 0) 
    {
#if 0
        if (ioctl(socket_no, FIONBIO, &tmp) == -1) 
        {
                Ethrowv("failed to ioctl FIONBIO: %d", errno);
        }
#else                           /* sun */
        if (fcntl(Socket->u.number, F_SETFL, O_NDELAY) == -1) 
        {
                Ethrowv("failed to fcntl O_NDELAY: %d", errno);
        }
#endif                          /* sun */
    }
    return EValint(1); 
end pragma;
    .

setsockopt(Socket, Level, Optname, Value) ->
pragma
    int tmp = Value->u.number;
    if (setsockopt(Socket->u.number, Level->u.number, 
        Optname->u.number, (char *)&tmp, sizeof(tmp))==-1)
    {
            Ethrowv("setsockopt failed: %d", errno);
    }
    return EValbool(1);
end pragma;
    .

%%
%% Name: shutdown
%% Purpose: shutdown a socket gracefully 
%% Parameters: a socket 3-tuple and a number showing how (from C man page)
%%
shutdown({Socket,_,_}, How) ->
pragma
    return shutdown(Socket->u.number, How->u.number);
end pragma;
    .


%%
%% Name: select_all
%% Purpose: to select on a group of open sockets looking for input
%%        _all means - check all descriptors in the list
%% Parameters: Connected - list of open sockets,
%%  { Secs, Usecs } - a timing tuple
%%
select_all(Connected, { Secs, Usecs }) ->
pragma
    struct timeval timeout; 
    fd_set readfds;
    int max, res;
    int count = 0;
    Val ret;

    FD_ZERO(&readfds);

    for (i = 0; i < Connected->u.list->size; i++)
    {
        T = Connected->u.list->e[i]->u.number;
        FD_SET(T, &readfds);
        if (T >= max) max = T+1;
    }

    timeout.tv_sec = Sec->u.number;
    timeout.tv_usec = Usec->u.number;

    res = select(max, &readfds, 0, 0, &timeout);

    if (res == -1)
    {
        Ethrowv("select error: %d\n", errno);
    }

    ret = Elistalloc(res);
    for (i = 0; i < Connected->u.list->size; i++)
    {
        if (FD_ISSET(Connected->u.list->e[i]), &readfds)
        {
            EValCopyin(ret->u.list->e[count++], Connected->u.list->e[i]);
        }
    }
    return ret;
end pragma;
    .

