%% Copyright (C) 1992, Ellemtel Telecommunications Systems Laboratories
%% File    : string.erl
%% Author  : Robert Virding, Torbjrn Trnkvist
%% Purpose : String library.

-module(string).

-export([len/1,equal/2,concat/2,chr/2,rchr/2,index/2,
	 substr/2,substr/3,tokens/2,chars/2,chars/3]).
-export([copies/2,words/1,words/2,strip/1,strip/2,strip/3,index/2,
	 sub_word/2,sub_word/3,left/2,left/3,right/2,right/3,
	 sub_string/2,sub_string/3,centre/2,centre/3]).
-export([re_sh_to_awk/1,re_parse/1,re_match/2,re_sub/3,re_gsub/3,re_split/2]).

-import(lists,[append/2,reverse/1,reverse/2,member/2]).

%% Robert's bit

%% len(String)
%%  Return the length of a string.

len(S) ->
    length(S).

%% equal(String1, String2)
%%  Test if 2 strings are equal.

equal(S, S) -> true;
equal(_, _) -> false.

%% concat(String1, String2)
%%  Concatenate 2 strings.

concat(S1, S2) ->
    append(S1, S2).

%% chr(String, Char)
%% rchr(String, Char)
%%  Return the first (last) index of the character in a string.

chr(S, C) ->
    chr(S, C, 1).

chr([C|Cs], C, I) ->
    I;
chr([_|Cs], C, I) ->
    chr(Cs, C, I+1);
chr([], C, I) ->
    0.

rchr(S, C) ->
    rchr(S, C, 1, 0).

rchr([C|Cs], C, I, L) ->			%Found one, now find next!
    rchr(Cs, C, I+1, I);
rchr([_|Cs], C, I, L) ->
    rchr(Cs, C, I+1, L);
rchr([], C, I, L) ->
    L.

%% index(String, SubString)
%%  Return the index of the sub-string in a string.

index(S, Sub) ->
    index1(S, Sub, 1).

%% index1(String, SubStr, Index)
%% index2(String, SubStr, StartIndex, StartString, StartSub)
%%  Outside/inside a sub-string.

index1([C|S], [C|Sub], I) ->
    index2(S, Sub, I, S, [C|Sub]);
index1([_|S], Sub, I) ->
    index1(S, Sub, I+1);
index1([], Sub, I) ->
    0.

index2([C|S], [C|Sub], I, StS, StSub) ->
    index2(S, Sub, I, StS, StSub);
index2(S, [], I, StS, StSub) ->			%Found it
    I;
index2([_|S], Sub, I, StS, StSub) ->
    index1(StS, StSub, I+1);
index2([], Sub, I, StS, StSub) ->
    index1(StS, StSub, I+1).

%% substr(String, Start)
%% substr(String, Start, Length)
%%  Extract a sub-string from String.

substr(S, I) when I >= 1 ->
    substr1(S, I, 1000000).			%Try for a long string

substr(S, I, L) when I >= 1, L >= 0 ->
    substr1(S, I, L).

%% substr1(String, Start, Length)
%% substr2(String, Length)
%%  Functions that do the work before/in the bit extracted.

substr1(S, 1, L) -> substr2(S, L);
substr1([_|Ss], I, L) -> substr1(Ss, I-1, L).

substr2(S, 0) -> [];
substr2([C|Ss], L) -> [C|substr2(Ss, L-1)];
substr2([], L) -> [].				%Be nice!

%% tokens(String, Seperators).
%%  Return a list of tokens seperated by characters in Seperators.

tokens(S, Seps) ->
    tokens1(S, Seps, []).

tokens1([C|S], Seps, Toks) ->
    case member(C, Seps) of
	true ->
	    tokens1(S, Seps, Toks);
	false ->
	    tokens2(S, Seps, Toks, [C])
    end;
tokens1([], Seps, Toks) ->
    reverse(Toks).

tokens2([C|S], Seps, Toks, Cs) ->
    case member(C, Seps) of
	true ->
	    tokens1(S, Seps, [reverse(Cs)|Toks]);
	false ->
	    tokens2(S, Seps, Toks, [C|Cs])
    end;
tokens2([], Seps, Toks, Cs) ->
    reverse([reverse(Cs)|Toks]).

chars(C, N) ->
    chars(C, N, []).

chars(C, N, Tail) when N > 0 ->
    [C|chars(C, N-1, Tail)];
chars(C, 0, Tail) ->
    Tail.

%% Torbjrn's bit.

%%% COPIES %%%

copies(_,0) -> [];
copies(CharList,Num) when list(CharList) ->
    append(CharList,copies(CharList,Num-1)).

%%% WORDS %%%

words(String) -> words(String,$ ).

words(String,Char) when integer(Char) ->
    w_count(strip(String,both,Char),Char,0).

w_count([],_,Num) -> Num+1;
w_count([H|T],H,Num) -> w_count(strip(T,left,H),H,Num+1);
w_count([H|T],Char,Num) -> w_count(T,Char,Num).

%%% SUB_WORDS %%%

sub_word(String,Index) -> sub_word(String,Index,$ ).

sub_word(String,Index,Char) when integer(Index),integer(Char) ->
    case words(String,Char) of
	Num when Num < Index ->
	    [];
	Num ->
	    s_word(strip(String,left,Char),Index,Char,1,[])
    end.

s_word([],_,_,_,Res) -> reverse(Res);
s_word([Char|T],Index,Char,Index,Res) -> reverse(Res);
s_word([H|T],Index,Char,Index,Res) -> s_word(T,Index,Char,Index,[H|Res]);
s_word([Char|T],Stop,Char,Index,Res) when Index < Stop -> 
    s_word(strip(T,left,Char),Stop,Char,Index+1,Res);
s_word([H|T],Stop,Char,Index,Res) when Index < Stop -> 
    s_word(T,Stop,Char,Index,Res).

%%% STRIP %%%

strip(String) -> strip(String, both).

strip(String, left) -> strip_left(String, $ );
strip(String, right) -> strip_right(String, $ );
strip(String, both) ->
    strip_right(strip_left(String, $ ), $ ).

strip(String, right, Char) -> strip_right(String, Char);
strip(String, left, Char) -> strip_left(String, Char);
strip(String, both, Char) ->
    strip_right(strip_left(String, Char), Char).

strip_left([Sc|S], Sc) ->
    strip_left(S, Sc);
strip_left([C|S], Sc) -> [C|S];
strip_left([], Sc) -> [].

strip_right([Sc|S], Sc) ->
    strip_right_1(S, Sc, 1);
strip_right([C|S], Sc) ->
    [C|strip_right(S, Sc)];
strip_right([], Sc) ->
    [].

strip_right_1([Sc|S], Sc, Scn) ->
    strip_right_1(S, Sc, Scn+1);
strip_right_1([C|S], Sc, Scn) ->
    chars(Sc, Scn, [C|strip_right(S, Sc)]);
strip_right_1([], Sc, Scn) ->
    [].

%%% LEFT %%%

left(String,Len) -> left(String,Len,$ ).

left(String,Len,Char) when integer(Char) ->
    case length(String) of
	Num when Num == Len ->
	    String;
	Num when Num > Len ->
	    l_cut(String,Len);
	Num when Num < Len ->
	    l_pad(String,Len-Num,Char)
    end.

l_cut(S, L) -> substr2(S, L).			%Extracts string length L

l_pad(String, Num, Char) ->
    append(String, chars(Char, Num, [])).

%%% RIGHT %%%

right(String,Len) -> right(String,Len,$ ).

right(String,Len,Char) when integer(Char) ->
    case length(String) of
	Num when Num == Len ->
	    String;
	Num when Num > Len ->
	    r_cut(String,Num-Len);
	Num when Num < Len ->
	    r_pad(String,Len-Num,Char)
    end.

r_cut(S, Start) -> substr1(S, Start+1, 100000).

r_pad(String, Num, Char) -> chars(Char, Num, String).

%%% CENTRE %%%

centre(String,Len) -> centre(String,Len,$ ).

centre(String,0,_) -> [];        % Strange cases when to centre a string
centre(String,Len,Char) ->
    Slen = length(String),
    if
	Len =< Slen ->
	    T = trunc((Slen-Len) / 2),
	    N = Slen- T,
	    r_cut(l_cut(String, N), N-Len);
	true ->
	    N = trunc((Len-Slen)/2),
	    r_pad(l_pad(String,N,Char),N+(Len-Slen-N*2),Char)
    end.
	    

%%% SUB_STRING %%%

sub_string(String, Start) when Start >= 1 ->
    substr1(String, Start, 100000).

sub_string(String, Start, Stop) when Start >= 1, Stop >= Start ->
    substr1(String, Start, Stop - Start + 1).

%% The Regular Expression Matcher
%%
%% This section implements regex matching of strings. The regexes supported
%% those which exist in AWK and the interface is AWK-like.

%% re_sh_to_awk(ShellRegExp)
%%  Convert a sh style regex into a full AWK one. The main difficulty is
%%  getting character sets right as the conventions are different.

re_sh_to_awk(Sh) ->
    [$^|re_sh_to_awk_1(Sh)].			%Fix the beginning

re_sh_to_awk_1([$.|Sh]) ->			%This must be made non-special
    [$\\,$.|re_sh_to_awk_1(Sh)];
re_sh_to_awk_1([$*|Sh]) ->			%This matches any string
    [$.,$*|re_sh_to_awk_1(Sh)];
re_sh_to_awk_1([$?|Sh]) ->			%This matches any character
    [$.|re_sh_to_awk_1(Sh)];
re_sh_to_awk_1([$[,$^,$]|Sh]) ->		%This takes careful handling
    [$\\,$^|re_sh_to_awk_1(Sh)];
re_sh_to_awk_1([$[,$^|Sh]) ->
    [$[|re_sh_to_awk_2(Sh, true)];
re_sh_to_awk_1([$[,$!|Sh]) ->
    [$[,$^|re_sh_to_awk_2(Sh, false)];
re_sh_to_awk_1([$[|Sh]) ->
    [$[|re_sh_to_awk_2(Sh, false)];
re_sh_to_awk_1([C|Sh]) ->
    [C|re_sh_to_awk_1(Sh)];
re_sh_to_awk_1([]) -> [$$].			%Fix the end

re_sh_to_awk_2([$]|Sh], UpArrow) ->
    [$]|re_sh_to_awk_3(Sh, UpArrow)];
re_sh_to_awk_2(Sh, UpArrow) ->
    re_sh_to_awk_3(Sh, UpArrow).

re_sh_to_awk_3([$]|Sh], true) ->
    [$^,$]|re_sh_to_awk_1(Sh)];
re_sh_to_awk_3([$]|Sh], false) ->
    [$]|re_sh_to_awk_1(Sh)];
re_sh_to_awk_3([C|Sh], UpArrow) ->
    [C|re_sh_to_awk_3(Sh, UpArrow)];
re_sh_to_awk_3([], true) ->
    [$^|re_sh_to_awk_1([])];
re_sh_to_awk_3([], false) ->
    re_sh_to_awk_1([]).

%% re_match(String, RegExp)
%%  Try to match a string with regular expression. Accept pre-parsed regular
%%  expressions.

re_match(String, {regexp,RE}) ->		%We have already parsed it.
    re_match_apply(String, RE);
re_match(String, RegExp) ->
    case re_parse1(RegExp) of
	{error,What,Where} ->
	    {error,What};
	RE ->
	    re_match_apply(String, RE)
    end.

%% re_match_apply(String, RegExp)
%% re_match_apply(MatchResult, String, RegExp, Index)
%%  Successively step down the string trying to match the regular expression
%%  until a match is found.

re_match_apply(String, RE) ->
    re_match_apply(re_apply(RE, String, bos), String, RE, 1).

re_match_apply({match,Cs,St,S}, S0, RE, I) ->
    {match,reverse(Cs, []),I};
re_match_apply(nomatch, [C|S], RE, I) ->	%More string to try
    re_match_apply(re_apply(RE, S, mos), S, RE, I+1);
re_match_apply(nomatch, [], RE, I) ->		%That's it
    nomatch.

%% re_sub(String, RegExp, Replace)
%%  Substitute the first match of the regular expression RegExp with the
%%  string Replace in String. Accept pre-parsed regular expressions.

re_sub(String, {regexp,RE}, Rep) ->
    {ok,re_sub_apply(String, RE, Rep)};
re_sub(String, RegExp, Rep) ->
    case re_parse1(RegExp) of
	{error,What,Where} ->
	    {error,What};
	RE ->
	    {ok,re_sub_apply(String, RE, Rep)}
    end.

%% re_sub_apply(RegExp, String, Replace)
%% re_sub_apply(MatchResult, String, RegExp, Replace)
%%  Successively step down the string trying to match the regular expression
%%  until a match is found, then replace it.

re_sub_apply(String, RE, Rep) ->
    re_sub_apply(re_apply(RE, String, bos), String, RE, Rep).

re_sub_apply({match,Cs,St,S}, S0, RE, Rep) ->
    re_sub_repl(Rep, reverse(Cs, []), S);
re_sub_apply(nomatch, [C|S], RE, Rep) ->
    [C|re_sub_apply(re_apply(RE, S, mos), S, RE, Rep)];
re_sub_apply(nomatch, [], RE, Rep) ->
    [].

re_sub_repl([$&|Rep], Cs, Rest) ->
    append(Cs, re_sub_repl(Rep, Cs, Rest));
re_sub_repl([$\\,$&|Rep], Cs, Rest) ->
    [$&|re_sub_repl(Rep, Cs, Rest)];
re_sub_repl([C|Rep], Cs, Rest) ->
    [C|re_sub_repl(Rep, Cs, Rest)];
re_sub_repl([], Cs, Rest) ->
    Rest.

%% re_gsub(String, RegExp, Replace)
%%  Substitute every match of the regular expression RegExp with the
%%  string New in String. Accept pre-parsed regular expressions.

re_gsub(String, {regexp,RE}, Rep) ->
    re_gsub_apply(String, RE, Rep);
re_gsub(String, RegExp, Rep) ->
    case re_parse1(RegExp) of
	{error,What,Where} ->
	    {error,What};
	RE ->
	    {ok,re_gsub_apply(String, RE, Rep)}
    end.

re_gsub_apply(String, RE, Rep) ->
    re_gsub_apply(re_apply(RE, String, bos), String, RE, Rep).

re_gsub_apply({match,[],St,[]}, S0, RE, Rep) ->
    %% Have not moved so must move ourselves.
    re_sub_repl(Rep, [], []);
re_gsub_apply({match,[],St,[C|S]}, S0, RE, Rep) ->
    %% Have not moved so must move ourselves.
    re_sub_repl(Rep, [], [C|re_gsub_apply(re_apply(RE, S, mos), S, RE, Rep)]);
re_gsub_apply({match,Cs,St,S}, S0, RE, Rep) ->
    re_sub_repl(Rep, reverse(Cs, []),
		re_gsub_apply(re_apply(RE, S, mos), S, RE, Rep));
re_gsub_apply(nomatch, [C|S], RE, Rep) ->
    [C|re_gsub_apply(re_apply(RE, S, mos), S, RE, Rep)];
re_gsub_apply(nomatch, [], RE, Rep) ->
    [].

%% re_split(String, RegExp)
%%  Split a string into substrings where the RegExp describes the field
%%  seperator. The RegExp " " is specially treated, as are other single
%%  character REs.

re_split(String, {regexp,RE}) ->
    re_split_apply(String, RE, false);
re_split(String, "") ->				%Just pass it straight through
    {ok,[String]};
re_split(String, " ") ->			%This is really special
    case re_split_apply(String, re_parse1("[ \t]+"), true) of
	[[]|Ss] ->
	    {ok,Ss};
	Ss ->
	    {ok,Ss}
    end;
re_split(String, [C]) ->			%Single characters are NOT REs
    {ok,re_split_apply(String, [C], false)};	%We KNOW
re_split(String, RegExp) ->
    case re_parse1(RegExp) of
	{error,What,Where} ->
	    {error,What};
	RE ->
	    {ok,re_split_apply(String, RE, false)}
    end.

re_split_apply(String, RE, Trim) ->
    re_split_apply(re_apply(RE, String, bos), String, RE, [], Trim).

re_split_apply({match,Cs,St,S}, S0, RE, Sub, T) ->
    [reverse(Sub, [])|re_split_apply(re_apply(RE, S, mos), S, RE, [], T)];
re_split_apply(nomatch, [C|S], RE, Sub, T) ->
    re_split_apply(re_apply(RE, S, mos), S, RE, [C|Sub], T);
re_split_apply(nomatch, [], RE, [], true) ->
    [];
re_split_apply(nomatch, [], RE, Sub, T) ->
    [reverse(Sub, [])].

%% re_parse(RegExpString)
%%  Try to parse a regular expression.

re_parse(RegExp) ->
    case re_parse1(RegExp) of
	{error,What,Where} ->
	    {error,What};
	Parse ->
	    {regexp,{regexp,Parse}}
    end.

re_parse1(Cs0) ->
    case catch re_parse1(Cs0, []) of
	{RE,[]} ->
	    RE;
	{RE,[C|Cs]} ->
	    {error,[C],Cs};
	Error ->				%All errors
	    Error
    end.

%% Special characters.
re_parse1([$.|Cs], Stack) ->
    re_parse1(Cs, [char|Stack]);
re_parse1([$*|Cs], [RE|Stack]) ->
    re_parse1(Cs, [{span_0,[RE]}|Stack]);
re_parse1([$+|Cs], [RE|Stack]) ->
    re_parse1(Cs, [{span_1,[RE]}|Stack]);
re_parse1([$?|Cs], [RE|Stack]) ->
    re_parse1(Cs, [{span_0_1,[RE]}|Stack]);
re_parse1([$[,$^|Cs0], Stack) ->
    {Set,Cs} = char_set(Cs0),
    re_parse1(Cs, [{compl_set,Set}|Stack]);
re_parse1([$[|Cs0], Stack) ->
    {Set,Cs} = char_set(Cs0),
    re_parse1(Cs, [{char_set,Set}|Stack]);
re_parse1([$(|Cs0], Stack) ->
    case re_parse1(Cs0, []) of
	{RE,[$)|Cs]} ->
	    re_parse1(Cs, [{group,RE}|Stack]);
	{RE,Cs} ->
	    throw({error,")",Cs})
    end;
re_parse1([$)|Cs], Stack) ->
    {reverse(Stack, []),[$)|Cs]};
re_parse1([$||Cs0], Stack) ->
    {RE,Cs} = re_parse1(Cs0, []),
    re_parse1(Cs, [{alt,reverse(Stack, []),RE}]);
re_parse1([$\\,C|Cs], Stack) ->
    re_parse1(Cs, [C|Stack]);

%% These are only valid at the beginning and end of strngs.
re_parse1([$^|Cs], Stack) ->
    re_parse1(Cs, [bos|Stack]);
re_parse1([$$|Cs], Stack) ->
    re_parse1(Cs, [eos|Stack]);

%% Any non-special character.
re_parse1([C|Cs], Stack) ->
    re_parse1(Cs, [C|Stack]);
re_parse1([], Stack) ->
    {reverse(Stack, []),[]}.

char_set([$]|Cs]) ->
    char_set1(Cs, [$]]);
char_set(Cs) ->
    char_set1(Cs, []).

char_set1([C1,$-,C2|Cs], Set) when C2 /= $], C1 =< C2 ->
    char_set1(Cs, [{C1,C2}|Set]);
char_set1([$]|Cs], Set) ->
    {Set,Cs};
char_set1([C|Cs], Set) ->
    char_set1(Cs, [C|Set]);
char_set1([], Set) ->
    throw({error,"]",[]}).

%% re_apply(RE, String)
%%  Apply a regular expression to a string and test if it matches.
%%  The match is only attempted at the START of the string, the caller
%%  must recursively traverse the string until a match is detected.
%%  N.B. The list of matched characters is reversed.

re_apply(RE, String, State) ->
    re_apply(RE, String, State, []).

%% re_apply(RE, String, State, CharsMatched)

re_apply([char|RE], [C|S], St, Cs) when C /= $\n ->	%.
    re_apply(RE, S, St, [C|Cs]);
re_apply([{span_0,RE1}|RE], S0, St, Cs0) ->	%*
    re_apply_span_0(re_apply(RE1, S0, St, Cs0), RE1, RE, S0, St, Cs0);
re_apply([{span_1,RE1}|RE], S0, St0, Cs0) ->	%+
    case re_apply(RE1, S0, St0, Cs0) of
	{match,Cs,St,S} ->
	    re_apply_span_0(re_apply(RE1, S, St, Cs), RE1, RE, S, St, Cs);
	nomatch ->
	    nomatch
    end;
re_apply([{span_0_1,RE1}|RE], S0, St0, Cs0) ->	%?
    case re_apply(RE1, S0, St0, Cs0) of
	{match,Cs,St,S} ->			%Matched one time
	    re_apply(RE, S, St, Cs);
	nomatch ->				%Matched zero times
	    re_apply(RE, S0, St0, Cs0)
    end;
re_apply([{compl_set,Set}|RE], [C|S], St, Cs) ->	%[^ ... ]
    case in_set(Set, C) of
	false ->
	    re_apply(RE, S, St, [C|Cs]);
	true ->
	    nomatch
    end;
re_apply([{char_set,Set}|RE], [C|S], St, Cs) ->	%[ ... ]
    case in_set(Set, C) of
	true ->
	    re_apply(RE, S, St, [C|Cs]);
	false ->
	    nomatch
    end;
re_apply([{group,RE1}|RE], S, St, Cs) ->	%( ... )
    re_apply_cont(re_apply(RE1, S, St, Cs), RE);
re_apply([{alt,Lhs,Rhs}|RE], S0, St0, Cs0) ->	%|
    case re_apply(Lhs, S0, St0, Cs0) of
	{match,Cs,St,S} ->
	    re_apply(RE, S, St, Cs);
	nomatch ->
	    re_apply_cont(re_apply(Rhs, S0, St0, Cs0), RE)
    end;
re_apply([bos|RE], S, bos, Cs) ->		%^
    re_apply(RE, S, mos, Cs);
re_apply([eos], [], St, Cs) ->			%$
    re_apply([], [], eos, Cs);
re_apply([eos], [$\n], St, Cs) ->		%Special hack for get_line
    re_apply([], [], eos, Cs);
re_apply([C|RE], [C|S], St, Cs) ->
    re_apply(RE, S, mos, [C|Cs]);
re_apply([], S, St, Cs) ->
    {match,Cs,St,S};
re_apply(RE, S, St, Cs) ->
    nomatch.

%% re_apply_scan_0(MatchResult, ScanRE, RestRE, String, State, MatchStack)

re_apply_span_0({match,Cs,St,S}, RE1, RE, S0, St0, Cs0) ->
    %% Must set up backtrack point for every span match.
    case re_apply_span_0(re_apply(RE1, S, St, Cs), RE1, RE, S, St, Cs) of
	{match,Cs1,St1,S1} ->
	    {match,Cs1,St1,S1};
	nomatch ->
	    re_apply(RE, S0, St0, Cs0)
    end;
re_apply_span_0(nomatch, RE1, RE, S0, St0, Cs0) ->
    re_apply(RE, S0, St0, Cs0).

%% re_apply_cont(MatchResult, RestRE)

re_apply_cont({match,Cs,St,S}, RE) ->
    re_apply(RE, S, St, Cs);
re_apply_cont(nomatch, RE) ->
    nomatch.

in_set([{Min,Max}|Set], C) when C >= Min, C =< Max ->
    true;
in_set([C|Set], C) ->
    true;
in_set([_|Set], C) ->
    in_set(Set, C);
in_set([], C) ->
    false.
