//
// $Id: Socket.h,v 1.1 1998/05/08 05:49:25 geoff Exp $
//
// Socket object
// Basically a container for all the C stuff
//
// $Source: /home/cvs/gerl/lib/Socket.h,v $
// $Revision: 1.1 $
// $Date: 1998/05/08 05:49:25 $
// $State: Exp $
//
// Author: Geoff Wong
//
#ifndef _SOCKET_H
#define _SOCKET_H

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <strings.h>
#ifdef FreeBSD
#include <netinet/tcp.h>
#else
#include <inet/tcp.h>
#endif
#include "str.h"

class Socket 
{
    int     socket_no;
    unsigned long    int_address;
    int     string_address;
    int     port;
    int     family;
    int     socket_type;
    int     protocol;

    char *     hostname;
    char *  error;

public:
    Socket();
    Socket(int domain, int type, int prot);
        // create a socket of a certain name 
    Socket * Accept();
    int Bind(int port_number);
        // bind our socket to a name
    int Bind(const char * name);
    int Close();
    int Connect();
        // connect assuming we have the stuff in memory
    int Connect(const char * address);
    int Connect(const char * address, unsigned int port);
    int Connect(unsigned long addr, unsigned int port);
    void Connectto(const char * address, unsigned int port);
        // Connect()s returns 0 on failure
        //            returns 1 on success

    int Fileno();
    char * Getpeername();
    char * Getsockopt(int level, int optname, int buflen);
    int Listen(int backlog);        
        // listen for a stream connection on our socket
    char * Recv(int bufsize);
        // Returns the string received.
        // The string is allocated (and must be deleted)
    char * Recvfrom(int bufsize, Socket *);
        // Store the return info in the given socket
    int Send(const char * what);
    int Sendto(const char * what);
    int Setblocking(int flags);
        // set blocking/nonblock mode for the socket
    int Setsockopt(int level, int optname, char * value);
    int Shutdown(int how);
        // shut me down gracefully 
    void Set_error(const char * how);
    char * Query_error();
    char * Ip_number();
    char * Hostname();
};

#define MAX_BUFF 65536

#endif
