

/*
** $Id: basic.c,v 1.80 2002/09/08 05:41:50 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/basic.c,v $
** $Revision: 1.80 $
** $Date: 2002/09/08 05:41:50 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1997, 1998
** geoff@serc.rmit.edu.au
**
**  . should be a shared library on target system
**  . should be able to allocate values on the C stack as 
**    well as the heap
**
*/

#include <assert.h>
#include <string.h>
#include "erlib.h"
#include "typecheck.h"
#include "str.h"

/* Support stuff */

#ifdef _ALLOCS_H
/* Val __vvv;      *//* an ugly allocation hack */
#endif

Val valalloc(Etype t)
{
    Val ret;
    ret.tag = t;
    ret.u.any = 0;
    return ret;
}


Val EValcopyin(Val dest, Val src)
{
    dest.tag = src.tag;
    dest.u.any = src.u.any;
    return dest;
}

inline Val EValint(int x)
{
    Val ret;

    ret.tag = LT_number;
    ret.u.number = x;

    return ret;
}

#ifndef _ALLOCS_H
inline Val EValatom(const char * atom)
{
    Val ret;

    ret.tag = LT_atom | LT_on_heap;
    ret.u.atom = string_copy(atom);

    return ret;
}

inline Val EValnatom(const char * atom, int len)
{
    Val ret;

    ret.tag = LT_atom;
    ret.u.atom = malloc(sizeof (char) * len + 1);
    if (ret.u.atom == NULL) Ethrowv("Out of memory (EValnatom)\n");
    memcpy(ret.u.atom, atom, len);
    ret.u.atom[len] = '\0';

/* fprintf(stderr, "Valnatom (%d): %s\n", len, ret->u.atom); */
    return ret;
}

#endif

/* Used to copy return values */
void Efre(Val v)
{

    if (v.tag & LT_on_heap) switch (v.tag & LT_single)
    {
        case LT_tuple:
        case LT_pid:
#if 0
    int i;
            for (i = 0; i < v.u.tuple->size; i++)
            {
                Efre(v.u.tuple->e[i]);
            }
            free(v.u.tuple);
#endif
            tuplefree(v);

            break;
        case LT_list:
#ifdef _CONS_H
            break;
            {
                _Lelement * le, * ole;
                le = v.u.list;
                if (le != NULL && (le->e.tag & LT_on_heap))
                {
                    ole = le;   
                    ole->e.tag |= LT_on_heap;
                    Efre(ole->e);
                    le = le->next;
                    free(ole);
                    while (le && !(le->e.tag & LT_on_heap))
                    {
                        ole = le;   
                        ole->e.tag |= LT_on_heap;
                        Efre(ole->e);
                        le = le->next;
                        free(ole);
                    }
                }
            }
#else
            listfree(v);
#endif
            break;
        case LT_string:
            Ethrowv("EValcopy: can't free LT_string\n");
            break;
        case LT_atom:
            free(v.u.atom);
            break;
        case LT_binary:
            free(v.u.binary);
            break;
        default:
            break;
    }
}

inline Val EValmatom(const char * atom, int len)
{
    Val ret;

    ret.tag = LT_atom | LT_on_heap;
    ret.u.atom = malloc(sizeof (char) * len + 1);
    if (ret.u.atom == NULL) Ethrowv("Out of memory (EValmatom)\n");
    memcpy(ret.u.atom, atom, len);
    ret.u.atom[len] = '\0';

#if 0
 fprintf(stderr, "Valmatom (%d): %s\n", len, ret->u.atom);
#endif
    return ret;
}


Val EValcopy(Val v)
{
    Val ret;

    switch (v.tag & LT_single)
    {
        case LT_tuple:
        case LT_pid:
            v.u.tuple->ref++;
            ret = v;

#if 0
#if 1
            memcpy(ret.u.tuple, v.u.tuple, sizeof(_Tuple) + sizeof(Val) * (v.u.tuple->size-1));
#else
    int i;
            for (i = 0; i < v.u.tuple->size; i++)
                ret.u.tuple->e[i] = EValcopy(v.u.tuple->e[i]);
#endif
#endif

            break;
        case LT_list:
#ifdef _CONS_H
            ret = v;
            if (ret.u.list && (!(ret.u.list->e.tag & LT_gc_only)))
            {
                ret.tag = ret.tag | LT_on_heap;
                if (ret.u.list->e.tag & LT_on_heap)
                {
                    ret.u.list->e.tag = (ret.u.list->e.tag & LT_single) | LT_gc_only;
                }
                else if (!(ret.u.list->e.tag & LT_gc_only))
                {
                    ret.u.list->e.tag = ret.u.list->e.tag | LT_on_heap;
                }
            }
            /* should we actually set the tag to LT_on_heap and  */
            /* the head of the list to the same (so it doesnt get freed). */
#else
#if 1
            v.u.list->ref++;
            ret = v;
#else
            // Maybe should be this - at least for a value that
            // will be returned (maybe I need LT_returned?)
            if (v.tag & LT_on_heap)
            {
                v.u.list->ref++;
                ret = v;
            }
            else
            {
                // note: correspond changes to erlang.y (pattern_tail)
                // required  so it knows to try to free copies of these
                if (v.u.list->length == 0) return v;
                ret.tag = v.tag | LT_on_heap;
                ret.u.list = listalloc();
                memcpy(ret.u.list, v.u.list, sizeof(_List));
                ret.u.list->e->ref++;
                ret = v;
            }
#endif

            //ret = listmalloc(0);
            //if (ret.u.list->e) ret.u.list->e->ref++;
#endif
            break;
        case LT_string:
            Ethrowv("EValcopy: can't copy LT_string\n");
            break;
        case LT_atom:
            ret = EValmatom(v.u.atom, strlen(v.u.atom));
            break;
        case LT_binary:
            /* ret = EValbinary(v.u.binary->size); */
            ret = EValmbinary(v.u.binary->b, v.u.binary->size);
            ret.tag = v.tag;
            break;
        default:
            return v;    
#if 0
            ret = valalloc(v.tag);
            ret.u.any = v.u.any;
            break;
#endif
    }
    /* if (v.tag & LT_on_heap) Efre(v); */
    return ret;
}

inline Val EValreal(float x)
{
    Val ret;

    ret.tag = LT_real;
    ret.u.real = x;

    return ret;
}

inline Val EValbool(int x)
{
    Val ret;

    ret.tag = LT_bool;
    ret.u.number = x;

    return ret;
}

Val Emultiply(Val l, Val r)
{

    if (l.tag & LT_number)
    {
        if (r.tag & LT_real) return EValreal(l.u.number * r.u.real);
        else if (r.tag & LT_number) return EValint(l.u.number * r.u.number);
        else TC(r, LT_number | LT_real, 2, "\"*\"");
    }
    else if (l.tag & LT_real)
    {
        if (r.tag & LT_real) return EValreal(l.u.real * r.u.real);
        else if (r.tag & LT_number) return EValreal(l.u.real * r.u.number);
        else TC(l, LT_number | LT_real, 1, "\"*\"");
    }
    TC(l, LT_number | LT_real, 1, "\"*\"");

    return l;
}

inline Val   Emultiplyvi(Val l, int n)
{
    if ((l.tag & LT_single) == LT_number)
    {
        return EValint(l.u.number * n);
    }
    else
    {
        return EValreal(l.u.real * n);
    }
}

inline Val Emultiplyiv(int l, Val n)
{
    if ((n.tag & LT_single) == LT_number)
    {
        return EValint(n.u.number * l);
    }
    else
    {
        return EValreal(n.u.real * l);
    }
}

inline Val Emultiplyrv(float l, Val n)
{
    if ((n.tag & LT_single) == LT_number)
    {
        return EValreal((float)n.u.number * l);
    }
    else
    {
        return EValreal(n.u.real * l);
    }
}

inline Val Emultiplyvr(Val n, float l)
{
    if ((n.tag & LT_single) == LT_number)
    {
        return EValreal((float)n.u.number * l);
    }
    else
    {
        return EValreal(n.u.real * l);
    }
}

inline Val Emultiplyii(int l, int r)
{
    return EValint(l * r);
}

Val    Eminus(Val l, Val r)
{
#if 0
    TC(l, LT_number | LT_real, 1, "\"-\"");
    TC(r, LT_number | LT_real, 2, "\"-\"");
    TCEQ(l,r, "\"-\"");
#endif

    if ((l.tag & LT_single) == LT_number)
    {
        return EValint(l.u.number - r.u.number);
    }
    else
    {
        return EValreal(l.u.real - r.u.real);
    }
}

inline Val    Eminusvi(Val l, int r)
{
    return EValint(l.u.number - r);
}

inline Val    Eminusiv(int l, Val r)
{
    return EValint(l - r.u.number);
}

inline Val    Eminusir(int l, float r)
{
    return EValreal((float)l - r);
}

inline Val    Eminusri(float l, int r)
{
    return EValreal(l - (float)r);
}

inline Val    Eminusvr(Val l, float r)
{
    return EValreal(l.u.real - r);
}

inline Val    Eminusrv(float l, Val r)
{
    return EValreal(l - r.u.real);
}

inline Val    Eminusrr(float l, float r)
{
    return EValreal(l - r);
}

inline Val    Eminusii(int l, int r)
{
    return EValint(l - r);
}

Val    Edivide(Val l, Val r)
{
    float a = 0.0, b = 0.0;

#if 0
    TC(l, LT_number | LT_real, 1, "\"/\"");
    TC(r, LT_number | LT_real, 2, "\"/\"");
#endif

    if (l.tag & LT_number)
    {
        a = (float) l.u.number;
    }
    else if (l.tag & LT_real)
    {
        a = l.u.real;
    }
    else TC(l, LT_number | LT_real, 1, "\"/\"");
    
    if (r.tag & LT_number)
    {
        b = (float) r.u.number;
    }
    else if (r.tag & LT_real)
    {
        b = r.u.real;
    }
    else TC(r, LT_number | LT_real, 2, "\"/\"");

#if 0
    if (b == 0.0)
    {
        Ethrowv("Division by zero error\n");
    }
#endif

    return EValreal(a / b);
}

inline Val   Edividevi(Val l, int n)
{
    float a;

    if ((l.tag & LT_single) == LT_number)
    {
        a = (float) l.u.number;
    }
    else
    {
        a = l.u.real;
    }
    return EValreal(a / n);
}

inline Val   Edividevr(Val l, float n)
{
    float a;

    if ((l.tag & LT_single) == LT_number)
    {
        a = (float) l.u.number;
    }
    else
    {
        a = l.u.real;
    }
    return EValreal(a / n);
}

inline Val   Edividerv(float l, Val n)
{
    float a;

    if ((n.tag & LT_single) == LT_number)
    {
        a = (float) n.u.number;
    }
    else
    {
        a = n.u.real;
    }
    return EValreal(l / a);
}

inline Val   Edivideri(float l, int n)
{
    return EValreal(l/(float)n);
}

inline Val   Edivideir(int l, float n)
{
    return EValreal((float)l/n);
}

Val    Ediv(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"div\"");
    TC(r, LT_number, 2, "\"div\"");
    TCEQ(l,r, "\"div\"");
#endif

    return EValint(l.u.number / r.u.number);
}

inline Val    Edivvi(Val l, Val r)
{
    return EValint(l.u.number / r.u.number);
}

inline Val    Ediviv(Val l, Val r)
{
    return EValint(l.u.number / r.u.number);
}

Val    Eplus(Val l, Val r)
{
#if 0
    TC(l, LT_number | LT_real, 1, "\"+\"");
    TC(r, LT_number | LT_real, 2, "\"+\"");
    TCEQ(l,r, "\"+\"");
#endif

    if ((l.tag & LT_single) == LT_number)
    {
        return EValint(l.u.number + r.u.number);
    }
    else
    {
        return EValreal(l.u.real + r.u.real);
    }
}

inline Val    Eplusvi(Val l, int r)
{
    return EValint(l.u.number + r);
}

inline Val    Eplusiv(int l, Val r)
{
    return EValint(l + r.u.number);
}

inline Val    Eplusvr(Val l, float r)
{
    return EValreal(l.u.real + r);
}

inline Val    Eplusrv(float l, Val r)
{
    return EValreal(l + r.u.real);
}

inline Val    Eplusrr(float l, float r)
{
    return EValreal(l + r);
}

/*
 * Enotequal - ugly overloaded function
 *  should be replaced in most places in the compiler with a function
 *  which has more implicit type info (and hence less "if" overhead)
 * Returns: 0 on equal, 1 for notequal
 */
inline int Enotequal(Val l, Val r)
{
    return !(Eequal(l,r));
}

inline int Enotequalvl(Val v, List l)
{
    if (!(v.tag & LT_list)) return 0;
    return !listcmp(v.u.list, l);
}

inline int Enotequalvi(Val l, int r)
{
    return (l.u.number != r);
}

inline int Enotequalva(Val l, char * r)
{
    return (strcmp(l.u.atom, r));
}

int Elessequalvi(Val l, int r)
{
    int ll;
    if (l.tag & LT_number)
        ll = l.u.number;
    else if (l.tag & LT_real)
        ll = (int) l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll <= r);
}

inline int Elessequalrr(float r, float l)
{
    return (r <= l);
}

inline int Eequalii(int r, int l)
{
    return (r == l);
}

inline int Elessequalii(int r, int l)
{
    return (r <= l);
}

int    Elessequalvr(Val l, float r)
{
    float ll;
    if (l.tag & LT_number)
        ll = (float) l.u.number;
    else if (l.tag & LT_real)
        ll = l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll <= r);
}

int    Elessequal(Val l, Val r)
{
    /* FIX: all types have some bizarre lexical ordering */
#if 0
    TC(l, LT_number | LT_real | LT_string, 1, "\"<=\"");
    TC(r, LT_number | LT_real | LT_string, 2, "\"<=\"");
    TCEQ(l,r, "\"<=\"");
#endif

    switch(l.tag)
    {
        case LT_number:
            if (l.u.number <= r.u.number) return (1);
            break;
        case LT_real:
            if (l.u.real <= r.u.real) return (1);
            break;
        case LT_string:
            /* FIX: verify <= on strcmp */
            if (strcmp(l.u.string, r.u.string) <= 0) return (1);
            break;
        case LT_atom:
            /* FIX: verify <= on strcmp */
            if (strcmp(l.u.atom, r.u.atom) <= 0) return (1);
            break;
        case LT_tuple:
            if (l.u.tuple->size <= r.u.tuple->size) return (1);
            break;
        case LT_list:
            {
#ifdef _CONS_H
                _Lelement * le = l.u.list;
                _Lelement * re = r.u.list;
                while (le && re) 
                {
                    le = le->next;
                    re = re->next;
                }
                if (le == NULL) return 1;
#else
            {
                int ll = 0, rl = 0;
                if (l.u.list) ll = l.u.list->length;
                if (r.u.list) rl = r.u.list->length;
                if (ll <= rl) return (1);
            }
#endif
            }
            break;
        default:
            break;
    }
    return (0);
}

int    Elessthanvi(Val l, int r)
{
    int ll;

    if (l.tag & LT_number)
        ll = l.u.number;
    else if (l.tag & LT_real)
        ll = (int) l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll < r);
}

int    Elessthanvr(Val l, float r)
{
    float ll;
    if (l.tag & LT_number)
        ll = (float) l.u.number;
    else if (l.tag & LT_real)
        ll = l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll < r);
}

int    Elessthan(Val l, Val r)
{
#if 0
    TC(l, LT_number | LT_real | LT_string, 1, "\"<\"");
    TC(r, LT_number | LT_real | LT_string, 2, "\"<\"");
    TCEQ(l,r, "\"<\"");
#endif

    switch(l.tag)
    {
        case LT_number:
            if (l.u.number < r.u.number) return (1);
            break;
        case LT_real:
            if (l.u.real < r.u.real) return (1);
            break;
        case LT_string:
            /* FIX: verify < on strcmp */
            if (strcmp(l.u.string, r.u.string) < 0) return (1);
            break;
        case LT_atom:
            /* FIX: verify < on strcmp */
            if (strcmp(l.u.atom, r.u.atom) < 0) return (1);
            break;
        case LT_tuple:
            if (l.u.tuple->size < r.u.tuple->size) return (1);
            break;
        case LT_list:
#ifdef _CONS_H
            {
                _Lelement * le = l.u.list;
                _Lelement * re = r.u.list;
                while (le && re) 
                {
                    le = le->next;
                    re = re->next;
                }
                if (le == NULL && re != NULL) return 1;
            }
#else
            {
                int ll = 0, rl = 0;
                if (l.u.list) ll = l.u.list->length;
                if (r.u.list) rl = r.u.list->length;
                if (ll < rl) return (1);
            }
#endif
            break;
        default:
            break;
    }
    return (0);
}

int    Egreaterequal(Val l, Val r)
{
#if 0
    TC(l, LT_number | LT_real | LT_string, 1, "\">=\"");
    TC(r, LT_number | LT_real | LT_string, 2, "\">=\"");
    TCEQ(l,r, "\">=\"");
#endif
    switch(l.tag)
    {
        case LT_number:
            if (l.u.number >= r.u.number) return (1);
            break;
        case LT_real:
            if (l.u.real >= r.u.real) return (1);
            break;
        case LT_string:
            /* FIX: verify >= on strcmp */
            if (strcmp(l.u.string, r.u.string) >= 0) return (1);
            break;
        case LT_atom:
            /* FIX: verify >= on strcmp */
            if (strcmp(l.u.atom, r.u.atom) >= 0) return (1);
            break;
        case LT_tuple:
            if (l.u.tuple->size >= r.u.tuple->size) return (1);
            break;
        case LT_list:
#ifdef _CONS_H
            {
                _Lelement * le = l.u.list;
                _Lelement * re = r.u.list;
                while (le && re) 
                {
                    le = le->next;
                    re = re->next;
                }
                if (re == NULL) return 1;
            }
#else
            {
                int ll = 0, rl = 0;
                if (l.u.list) ll = l.u.list->length;
                if (r.u.list) rl = r.u.list->length;
                if (ll >= rl) return (1);
            }
#endif
            break;
        default:
            break;
    }
    return (0);
}

int    Egreaterequalii(int l, int r)
{
    return (l >= r);
}

int    Egreaterequalvi(Val l, int r)
{
    int ll;
    if (l.tag & LT_number)
        ll = l.u.number;
    else if (l.tag & LT_real)
        ll = (int) l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll >= r);
}

int    Egreaterthan(Val l, Val r)
{
#if 0
    TC(l, LT_number | LT_real | LT_string, 1, "\">\"");
    TC(r, LT_number | LT_real | LT_string, 2, "\">\"");
    TCEQ(l,r, " > ");
#endif

    switch(l.tag)
    {
        case LT_number:
            if (l.u.number > r.u.number) return (1);
            break;
        case LT_real:
            if (l.u.real > r.u.real) return (1);
            break;
        case LT_string:
            /* FIX: verify > on strcmp */
            if (strcmp(l.u.string, r.u.string) > 0) return (1);
            break;
        case LT_atom:
            /* FIX: verify > on strcmp */
            if (strcmp(l.u.atom, r.u.atom) > 0) return (1);
            break;
        case LT_tuple:
            if (l.u.tuple->size > r.u.tuple->size) return (1);
            break;
        case LT_list:
#ifdef _CONS_H
            {
                _Lelement * le = l.u.list;
                _Lelement * re = r.u.list;
                while (le && re) 
                {
                    le = le->next;
                    re = re->next;
                }
                if (re == NULL && le != NULL) return 1;
            }
#else
            {
                int ll = 0, rl = 0;
                if (l.u.list) ll = l.u.list->length;
                if (r.u.list) rl = r.u.list->length;
                if (ll > rl) return (1);
            }
#endif
            break;
        default:
            break;
    }
    return (0);
}

int    Egreaterthanrr(float l, float r)
{
    return (l > r);
}

int    Egreaterthanvi(Val l, int r)
{
    int ll;
    if (l.tag & LT_number)
        ll = l.u.number;
    else if (l.tag & LT_real)
        ll = (int) l.u.real;
    else
    {
        TC(l, LT_number | LT_real, 1, "\"<\"");
        return 0;
    }

    return (ll > r);
}

inline int Egreaterthanii(int l, int r)
{
    return (l > r);
}

/* how are we going to do this? */
inline int Eexactequal(Val l, Val r)
{
    return 0;
}

inline int Etrue(Val b)
{
    return (b.u.number != 0);
}

inline Val    Ebnot(Val l)
{
    //TC(l, LT_number, 1, "\"bnot\"");
    return EValint(~l.u.number);
}

inline Val    Erem(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"rem\"");
    TC(r, LT_number, 2, "\"rem\"");
#endif

    return EValint(l.u.number % r.u.number);
}

inline Val    Eband(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"band\"");
    TC(r, LT_number, 2, "\"band\"");
    TCEQ(l,r, "\"band\"");
#endif

    return EValint(l.u.number & r.u.number);
}

int    Eequal(Val l, Val r)
{
    //printf("Eequal %d,%d\n", l.tag, r.tag);
    if (!(l.tag & r.tag)) return 1;

    switch (l.tag & LT_anyvar)
    {
        case LT_number:
            return (l.u.number == r.u.number);
        case LT_real:
            return (l.u.real == r.u.real);
        case LT_list:
            return listcmp(l.u.list,r.u.list);
        case LT_atom:
            return atomcmp(l,r);
        case LT_tuple:
            return tuplecmp(l,r);
        case LT_string:
            return atomcmp(l,r);
        case LT_binary:
            return binarycmp(l,r);
        case LT_pid:
            // return pidcmp(l,r);
        case LT_function:
            //return functioncmp(l,r);
        case LT_port:
            // return portcmp(l,r);
        case LT_var:
        case LT_token:
        default:
            return 0;
    }
}

inline int    Eequalbi(Val l, int b)
{
    return (l.u.number == b);
}

inline int    Estrcmp(Val l, const char * r)
{
    TC(l, LT_atom, 1, "\"Estrcmp\"");

    return strcmp(l.u.atom, r);
}

inline int    Eequalvi(Val l, int x)
{
#if 0
    if (l.tag != LT_number) 
    {
        return 0;
    }
#endif
    if (l.u.number == x) return 1;
    return 0;
}

inline int Eequalvr(Val l, float r)
{
    if (l.u.real == r) return 1;
    return 0;
}

int    Eequalvs(Val l, const char * s)
{
    /* FIX: typecheck? */
    if ((l.tag & LT_single) != LT_atom) return 0;
    if (strcmp(l.u.atom, s) == 0) return 1;
    return 0;
}

Val    Ebor(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"bor\"");
    TC(r, LT_number, 2, "\"bor\"");
#endif

    return EValint(l.u.number | r.u.number);
}

inline Val    Ebxor(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"bxor\"");
    TC(r, LT_number, 2, "\"bxor\"");
#endif

    return EValint(l.u.number ^ r.u.number);
}

inline Val    Ebsl(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"bsl\"");
    TC(r, LT_number, 2, "\"bsl\"");
#endif

    return EValint(l.u.number << r.u.number);
}

inline Val    Ebsr(Val l, Val r)
{
#if 0
    TC(l, LT_number, 1, "\"bsr\"");
    TC(r, LT_number, 2, "\"bsr\"");
#endif

    return EValint(l.u.number >> r.u.number);
}

inline Val    Epreplus(Val v)
{
    return v;
}

Val    Epreminus(Val v)
{
    if (v.tag & LT_number)
        v.u.number = - v.u.number;
    else if (v.tag & LT_real)
        v.u.real = - v.u.real;
    else
        TC(v, LT_number | LT_real, 1, "\"preminus\"");
    return v;
}


#ifndef _ALLOCS_H
/* 
 * The following three routines are used very 
 * often when library routines return.
 */

Val Eokay()
{
    return EValatom("ok");
}

Val Eok(Val v)
{
    return Emake_tuple(2, EValatom("ok"), v);
}

Val Eerror(const char *emsg)
{
    return Emake_tuple(2, EValatom("error"), EValmstring(emsg));
}
#endif

int atomcmp(Val l, Val r)
{
    return !(strcmp(l.u.atom, r.u.atom));
}

