
%%
%% Erlang Built-in-Functions Module
%% Contains skeleton code for Bifs
%% This is then partially processed into C
%% Further (direct) changes are then made to the C code
%%
%% Reference: Armstrong, Virding, Williams (1993): Concurrent
%%		Programming in ERLANG.
%%
%% Author: Geoff Wong
%%

-module(bif).
-author('geoff@shattered.org').
-export([abs/1,apply/3,apply/2,atom_to_list/1,date/0,element/2,
    erase/0,erase/1,exit/1,exit/2,float/1,record/2,binary/1,get/0]).

pragma
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include "str.h"
end pragma;
.

%% Name: abs/1
%%	takes a number and returns the absolute value
abs(N) ->
	if 
		N < 0 -> -N;
		N >= 0 -> N
	end
    .
	
%%
%%
apply(Module,Function,ArgList) ->
pragma
    return Ecalllist(string_cat(Module.u.atom,".so"), 
            Emangle(Module.u.atom, Function.u.atom, 
                ArgList.u.list->length), ArgList);
end pragma;
	.


%%
%%
apply( {Module,Function} ,ArgList) ->
    apply(Module,Function,ArgList)
	.

%% Name: atom_to_list/1
%%	takes an Atom and converts it to a list of ASCII numbers
atom_to_list(Atom) ->
pragma
#ifdef _ALLOCS_H
    Val _xx = listmalloc(0);
    return EValcopy(EValstring(_xx, Atom.u.atom));
#else
    return EValstring(Atom.u.atom);
#endif
end pragma;
	.	

%% Name: date/0
%% 	returns a tuple containing { year, month, day }
date() ->
pragma
     time_t t;
     struct tm *tm;
     Val dd,mm,yy; 

     t = time(NULL);
     tm = localtime(&t);

     dd = EValint(tm->tm_mday); 
     mm = EValint(tm->tm_mon + 1); 
     yy = EValint(tm->tm_year + 1900); 
#ifdef _ALLOCS_H
     return Emake_tuple(tuplemalloc(3), dd, mm, yy);
#else
     return Emake_tuple(3, dd, mm, yy);
#endif
end pragma;
	.

%% Name: element/2
%% 	return the nth element of a tuple
element(N, Tuple) ->
pragma
    TC(N, LT_number, 1, "bif:element/2");
    TC(Tuple, LT_tuple, 2, "bif:element/2");
    return EValcopy(Eindex_tuple(Tuple, N.u.number-1));
end pragma;
 	.


%% Name: erase/0
%% 	returns the process dictionary and deletes it
erase() ->
    throw('Unimplemented: bif:erase/0')
	.

%% Name: erase/1
%% 	returns the value associated with Key and deletes it from
%%	the process dictionary (can return undefined)
erase(Key) ->
    throw('Unimplemented: bif:erase/1')
	.

%% Name: exit/1
%% 	sends an exit message to process (Pid), returns true
%%	equivalent to Pid ! { 'EXIT', self(), goodbye }
exit(Reason) ->
pragma;
#ifdef _ALLOCS_H
    Ethrow(Emake_tuple(tuplemalloc(2), EValmatom("EXIT",4), Reason));
#else
    Ethrow(Emake_tuple(2, EValmatom("EXIT",4), Reason));
#endif
end pragma;
%%    exit(self(), Reason)
    .

%% Name: exit/2
%% 	sends an exit message to process (Pid), returns true
%%	equivalent to Pid ! { 'EXIT', self(), goodbye }
exit(Pid, Reason) ->
%%	Pid ! { 'EXIT', self(), Reason }
    'error'
    .

%% Name: float/1
%% 	converts a number to a float
float(X) ->
pragma 
    if (X.tag & LT_real) return EValbool(1);
    return EValbool(0);
end pragma;
 	.

%% Name: reference/1
%% Purpose: guard: returns true/false if it's a reference to a function
reference(X) ->
pragma
    if (X.tag & LT_function) return EValbool(1);
    return EValbool(0);
end pragma;
    .

%pid(X) ->
%pragma
%    if (X.tag & LT_pid) return EValbool(1);
%    return EValbool(0);
%end pragma;
%    .

port(X) ->
pragma
    if (X.tag & LT_port) return EValbool(1);
    return EValbool(0);
end pragma;
    .

%% Name: float_to_list/1
%% 	returns a list of ASCII values corresponding to Float
float_to_list(Float) ->
pragma
#define MAX_INT_LENGTH 128
    char buf[MAX_INT_LENGTH];
    sprintf(buf, "%f", Float.u.number);
    return EValmstring(buf);
end pragma;
	.

%% Name: get/0
%%	returns the process dictionary as a list of { Key, Value } tuples
get() ->
    throw('Unimplemented: bif:get/0')
	.

%% Name: get/2
%% 	returns a value association with key in the process dictionary
get(Key) ->
    throw('Unimplemented: bif:get/1')
	.

%% Name: get_keys/1
%%	returns a list of keys which correspond to Value in process dictionary
get_keys(Value) ->
    throw('Unimplemented: bif:get_keys/1')
	.

%% Name: halt/0
%%	halt the system
halt() ->
    throw('Unimplemented: bif:halt/0')
	.

%% Name: hd/1
%% 	return the head of a list
hd(AList) ->
pragma    
#ifdef _CONS_H
    return EValcopy(AList.u.list->e);
#else
    return EValcopy(Eindex_list(AList,0));
#endif
end pragma;
	.

%% Name: integer_to_list/1
%%  returns a list of ASCII values corresponding to Integer ie 77 = [55,55]
integer_to_list(Integer) ->
pragma
#define MAX_INT_LENGTH 128
    char buf[MAX_INT_LENGTH];
    sprintf(buf, "%d", Integer.u.number);
    return EValmstring(buf);
end pragma;
	.

%% Name: length/1
%% 	returns the length of a list
length(AList) ->
pragma
#ifdef _CONS_H
    int count = 0;
    _Lelement * le = AList.u.list;

    while (le) 
    {
        count++;
        le = le->next;
    }
     return EValint(count);
#else
    return EValint(AList.u.list->length); 
#endif
end pragma;
	.

%% Name: link/1
%%	make a link to a process/port of process Pid, returns true
link(Pid) ->
    throw('Unimplemented: bif:link/1')
	.

atom(A) ->
pragma
    if ((A.tag & LT_single) == LT_atom) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

binary(A) ->
pragma
    if ((A.tag & LT_single) == LT_binary) return EValbool(1);
    return EValbool(0);
end pragma;
    .

list(L) ->
pragma
    if ((L.tag & LT_single) == LT_list) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

tuple(L) ->
pragma
    if ((L.tag & LT_single) == LT_tuple) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

record(L,X) ->
pragma
    if ((L.tag & LT_single) == LT_tuple) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

integer(I) ->
pragma
    if ((I.tag & LT_single) == LT_number) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

number(X) ->
pragma
    if (((X.tag & LT_single) == LT_number) ||
        ((X.tag & LT_single) == LT_real)) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

pid(P) ->
pragma
    if ((P.tag & LT_single) == LT_pid) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

function(P) ->
pragma
    if ((P.tag & LT_single) == LT_function) return EValbool(1);
    else return EValbool(0);
end pragma;
    .

%% Name: list_to_atom/1
%%	returns an atom whose textual representation is that of integers
list_to_atom(Ascii) ->
pragma
    Val ret;
    char * buf;
    int x;

    // printf("bif:list_to_atom/1\n");
    TC(Ascii, LT_list, 1, "bif:list_to_atom/1");

    buf = (char *) malloc(Ascii.u.list->length+1);
    for (x = 0; x < Ascii.u.list->length; x++)
    {
        buf[x] = Eindex_list(Ascii,x).u.number;
    }
    buf[x] = '\0';
    ret = EValnatom(buf, Ascii.u.list->length);
    free(buf);
    return ret;
end pragma;
	.

%% Name: list_to_float/1
%%	returns a float whose textual representation is that of inters
list_to_float(AsciiIntegerList) ->
    'error'
	.
	
%% Name: list_to_integer/1
%%	returns an integer whose textual representation is that of inters
list_to_integer(AsciiIntegerList) ->
    'error'
	.

%% Name: list_to_pid/1
%%	returns a pid whose textual representation is that of inters
list_to_pid(AsciiIntegerList) ->
    'error'
	.

%% Name: list_to_tuple/1
%%	returns a tuple whose textual representation is that of inters
list_to_tuple(AsciiIntegerList) ->
    'error'
	.

%% Name: list_to_binary/1
%%
list_to_binary(ByteList) ->
pragma
    return Elist_to_binary(ByteList);
end pragma;
    .

term_to_binary(TermList) ->
    'error'
    .

binary_to_term(Binary) ->
    'error'
    .

split_binary(Bin, Spl) ->
    'error'
    .

binary_to_list(Bin) ->
pragma
    return Ebinary_to_list(Bin);
end pragma;
    .

binary_to_list(Bin, Spl, L) ->
    'error'
    .

%% Name: make_ref/0
%% 	returns a unique reference
make_ref() ->
    throw('Unimplemented: bif:make_ref/0')
	.

%% Name: node/0
%% 	returns an erlang term which represents the current node
node() ->
    throw('Unimplemented: bif:node/0')
	.

%% Name: node/1
%% 	returns an erlang term which represents the node on which Pid
%%	was opened.
node(Pid) ->
    throw('Unimplemented: bif:node/1')
	.

%% Name: disconnect_node/1
%% 
disconnect_node(X) ->
    throw('Unimplemented: bif:disconnect_node/1')
	.

%% Name: now/1
%%  return the current time
now() ->
pragma
    time_t t = time(0);

#ifdef _ALLOCS_H
    return Emake_tuple(tuplemalloc(3), EValint(0), EValint(t), EValint(t * 1000));
#else
    return Emake_tuple(3, EValint(0), EValint(t), EValint(t * 1000));
#endif
end pragma;
    .

%% Name: pid_to_list/1 ->
pid_to_list({_,_,Pid}) ->
pragma
#define MAXBUF 32
    char buf[MAXBUF];

    snprintf(buf, MAXBUF, "%d", Pid.u.number);
    return EValstring(buf);
end pragma;
	.


%% Name: process_flag/2
%% 	set certain flags for the process which calls this function
%%	returns the old value of the flag
process_flag(Flag,Option) ->
    'error'
	.

%% Name: process_info/1
%%	returns a long list containg information about the process Pid
%%	notes: used for debugging only
process_info(Pid) ->
    throw('Unimplemented: bif:process_info/1')
	.

process_info(X,Y) ->
    throw('Unimplemented: bif:process_info/2')
	.

%% Name: processes/0
%% 	returns a list of all process on the current noed
processes() ->
    throw('Unimplemented: bif:processes/1')
	.

put(Key, Value) ->
    throw('Unimplemented: bif:put/2')
	.

register(Name, Pid) ->
    throw('Unimplemented: bif:register/2')
	.

registered() ->
    throw('Unimplemented: bif:registed/0')
	.

round(Number) ->
    throw('Unimplemented: bif:round/0')
	.

self() ->
pragma
    extern Val _Eserverpid;
    return _Eserverpid;
end pragma;
	.

setelement(Index,Tuple,Value) ->
pragma
     Val NewTuple;

     NewTuple = tuplecopy(Tuple);
     Efre(NewTuple.u.tuple->e[Index.u.number - 1]);
     NewTuple.u.tuple->e[Index.u.number - 1] = EValcopy(Value);
     return NewTuple;
end pragma;
.

size(TupleOrBinary) ->
pragma
     switch (TupleOrBinary.tag)
     {
     case LT_tuple:
        return EValint(TupleOrBinary.u.tuple->size);
        break;
     case LT_binary:
        return EValint(TupleOrBinary.u.binary->size);
        break;
     default:
        Ethrowv("size/1: badarg (%d)", TupleOrBinary.tag);
	    break;
     }
end pragma;
	.

%
% Temp random for testing.
%
random(X) ->
pragma
    return EValint(random() % X.u.number);
end pragma;
.

spawn(Module, Function, ArgumentList) ->
pragma
    return Espawn(Module, Function, ArgumentList);
end pragma;
.

spawn(Module, Function, ArgumentList, X) ->
    throw('Unimplemented: bif:spawn/4')
    .

spawn_link(Module, Function, ArgumentList) ->
    throw('Unimplemented: bif:spawn_link/3')
	.

spawn_link(Node, Module, Function, ArgumentList) ->
    throw('Unimplemented: bif:spawn_link/4')
	.

statistics(Type) ->
pragma
#include <sys/time.h>
    static struct timeval mytime;
    struct timeval current;
    long milli;

    gettimeofday(&current, NULL);
    milli = (current.tv_sec - mytime.tv_sec) * 1000 + 
            (current.tv_usec - mytime.tv_usec) / 1000;

    mytime.tv_sec = current.tv_sec;
    mytime.tv_usec = current.tv_usec;

    return Emake_tuple(2, EValint(0), EValint(milli));
end pragma;
	.

term_to_words(Term) ->
    'error'
	.

throw(Any) ->
pragma
    Ethrow(Any);
end pragma;
	.

time() ->
pragma
     time_t t;
     struct tm *tm;
     Val s,m,h; 

     t = time(NULL);
     tm = localtime(&t);

     s = EValint(tm->tm_sec); 
     m = EValint(tm->tm_min); 
     h = EValint(tm->tm_hour); 
#ifdef _ALLOCS_H
     return Emake_tuple(tuplemalloc(3), h, m, s);
#else
     return Emake_tuple(3, h, m, s);
#endif
end pragma;
	.

tl(Mist) ->
pragma

#ifdef _ALLOCS_H
    return Etail_list(_ret, Mist, 1, "bif:tl/1");
#else
    return Etail_list(Mist, 1, "bif:tl/1");
#endif
end pragma;
	.

trunc(Number) ->
pragma
    int x;
    float f;
    if ((Number.tag & LT_normal) == LT_number) return Number;
    x = (int) Number.u.real;  /* truncates it .. */
    return EValint(x);
end pragma;
	.

not(V) ->
pragma
    if (!(V.tag & LT_bool)) Ethrowv("bif:not/1 on a non-boolean\n");
    if (V.u.number) return EValbool(0);
    return EValbool(1);
end pragma;
    .

tuple_to_list(Tuple) ->
pragma
    Val ret;
    TC(Tuple, LT_tuple, 1, "bif:tuple_to_list/1");

    ret = EValcopy(Tuple);
    ret.tag = LT_list;
    return ret;
end pragma;
	.

unlink(Pid) ->
    throw('Unimplemented: bif:unlink/1')
	.

unregister(Name) ->
    throw('Unimplemented: bif:unregister/1')
	.

whereis(Name) ->
    throw('Unimplemented: bif:whereis/1')
	.

words_to_term(Mist) ->
    throw('Unimplemented: bif:word_to_term/1')
	.

group_leader() ->
    throw('Unimplemented: bif:group_leader/0')
    .

group_leader(X,Y) ->
    throw('Unimplemented: bif:group_leader/2')
    .

nodes() ->
    throw('Unimplemented: bif:nodes/0')
    .

constant(X) ->
    throw('Unimplemented: bif:constant/1')
    .

is_alive() ->
    throw('Unimplemented: bif:is_alive/0')
    .

monitor_node(X,Y) ->
    throw('Unimplemented: bif:monitor_node/2')
    .

exit_shell(Reason) ->
    throw('Unimplemented: bif:exit_shell/1')
    .

%% Name: garbage_collect()
%%   Does nothing in Gerl cause of its stack based nature.
garbage_collect() ->
    1
    .

%% FIX:
%%
%% PortName:
%% {spawn, Command} 
%%
%% PortSettings (with spawn):
%% {env, Environment vars}
%% {cd, Dir}
%% 
%%
open_port(PortName, PortSetting) ->
pragma
#include <fcntl.h>
    Val spawn, command, dummy, item;
    int pid, i, count;
    int ifd[2], ofd[2], efd[2];
    int use_stdio = 1, redirerr = 1, usesocket = 0;
    // FIX: should use a sys #define
    int max_files = 1024;
    char ** comargv = NULL;
    char ** comenv = NULL;
    char * atom;

    fflush(stdout);
    if (PortName.tag & LT_tuple)
    {
        // printf("OpenPort PortName type=%d size=%d\n", PortName.tag, PortName.u.tuple->size);
        if (PortName.u.tuple->size == 2)
        {
            // { spawn, Command }, { Settings }
            spawn = PortName.u.tuple->e[0];
            command = PortName.u.tuple->e[1];
            
            // printf("spawn=%s\n", spawn.u.atom);
            // printf("command=%s\n", command.u.atom);

            // spawn.tag & LT_atom 
            if (strncmp(spawn.u.atom, "spawn", 5))
            {
                Ethrowv("Unknown port (%s) opened\n", spawn.u.atom);
                return dummy;
            }

            // FIX: Unimplemented Settings:
            if (PortSetting.tag & LT_list)
            {
                // printf("Found port settings %d\n", PortSetting.u.list->length);
                for (count = 0; count < PortSetting.u.list->length; count++)
                {
                    item = Eindex_list(PortSetting, count);

                    if (item.tag & LT_tuple && item.u.tuple->size > 1)
                    {
                        printf("Found item %d as tuple first element type %d\n", count,item.u.tuple->e[0].tag);

                        if (!(item.u.tuple->e[0].tag & LT_atom)) continue;
                        atom = item.u.tuple->e[0].u.atom;
                        if (!strncmp(atom, "packet",6))
                        {
                            // { packet, N } - message preceeded by length, sent in N bytes
                        }
                        else if (!strncmp(atom, "cd",2))
                        {
                            // { cd , Dir } - cd before executing program
                        }
                        else if (!strncmp(atom, "env",3))
                        {
                            // { env, Environment } - set environment variable
                        }
                        else if (!strncmp(atom, "socket",5))
                        {
                            // NOTE: non-standard extension
                            // { socket, Socket } - use a socket instead of stdio
                            printf("Using TCP/IP socket %d\n",
                                item.u.tuple->e[1].u.number);
                            usesocket = item.u.tuple->e[1].u.number;
                        }
                        else
                        {
                        }
                    }
                    else if (item.tag & LT_atom)
                    {
                        if (!strncmp(item.u.atom, "exit_status",11))
                        {
                            // exit_status
                        }
                        else if (!strncmp(item.u.atom, "use_stdio",9))
                        {
                            // use_stdio - fd 1,2
                            printf("Using stdio\n");
                            use_stdio = 1;
                        }
                        else if (!strncmp(item.u.atom, "nouse_stdio",11))
                        {
                            // nouse_stdio - fd 3,4
                            use_stdio = 0;
                        }
                        else if (!strncmp(item.u.atom, "in",2))
                        {
                            // in - input only
                        }
                        else if (!strncmp(item.u.atom, "out",3))
                        {
                            // out - output only
                        }
                        else if (!strncmp(item.u.atom, "binary",6))
                        {
                            // binary - all comms is binary not lists
                        }
                        else if (!strncmp(item.u.atom, "oef",3))
                        {
                            // eof - exit at eof instead of closing
                        }
                        else
                        {
                            // unknown
                        }
                    }
                }
            }

            // Command.tag & LT_atom / LT_list

            // If usesocket ...
            if (usesocket)
            {
                // hmmm ... might have to dup2()
                // since socket is already connected
                ofd[0] = dup(usesocket);
                ofd[1] = dup(usesocket);
                ifd[0] = dup(usesocket);
                ifd[1] = dup(usesocket);
                efd[0] = dup(usesocket);
                efd[1] = dup(usesocket);

                if (ofd[0] == -1 || ofd[1] == -1 ||
                    ifd[0] == -1 || ifd[1] == -1 ||
                    efd[0] == -1 || efd[1] == -1)
                {
                    Ethrowv("Failed to dup socket.\n"); 
                }
#if 0
                // socketpair is using a socket?
                if (pipe(ifd) < 0)
                {
                    Ethrowv("Failed to create input pipe.\n"); 
                }
                if (pipe(ofd) < 0)
                {
                    // Ethrowv
                    Ethrowv("Failed to create output pipe.\n"); 
                }
                if (pipe(efd) < 0)
                {
                    // Ethrowv
                    Ethrowv("Failed to create error pipe.\n"); 
                }
#endif
            }
            else
            {
                // create 3 new socket(s) for dups?
                if (pipe(ifd) < 0)
                {
                    Ethrowv("Failed to create input pipe.\n"); 
                }
                if (pipe(ofd) < 0)
                {
                    // Ethrowv
                    Ethrowv("Failed to create output pipe.\n"); 
                }
                if (pipe(efd) < 0)
                {
                    // Ethrowv
                    Ethrowv("Failed to create error pipe.\n"); 
                }
            }

            // 0 - stdin
            // 1 - stdout
            // 2 - stderr

            // Close all fds on exec - maybe?
            for (i = 3; i < max_files; i++)
                if (i != usesocket) fcntl(i, F_SETFD, 1);

            // printf("Will fork now (%s)\n", command.u.atom);

            pid = fork();

            if (pid > 0)
            {
                // parent - make different type of pid (4-tuple).
                // printf("Returning 4 tuple %d\n", pid);
                close(ifd[0]);
                close(ofd[1]);
                close(efd[1]);
                return Emake_tuple(4,EValint(pid), EValint(ifd[1]), 
                    EValint(ofd[0]), EValint(efd[0]));
            }

            if (pid == 0)
            {
                // FIX: not right.
                // child
                // 0 - stdin, 1 - stdout
                // bind fds to in/out/err
                if (use_stdio)
                {
                    if (dup2(ofd[1], 1) == -1)
                        perror("dup2 output");
                    if (dup2(ifd[0], 0) == -1)
                        perror("dup2 input");
                    if (redirerr)
                        dup2(ofd[1], 2);
                    else
                        dup2(efd[1], 2);
                }
                else
                {
                    if (dup2(ofd[1], 4) == -1)
                        perror("dup2 output");
                    if (dup2(ifd[0], 3) == -1)
                        perror("dup2 input");
                    if (redirerr)
                        dup2(ofd[1], 5);
                    else
                        dup2(efd[1], 5);
                }

                close(ifd[1]);
                close(ofd[0]);
                close(efd[0]);

                // Exec Command whatever
                // printf("execve %s\n", command.u.atom);
                if (execve(command.u.atom, comargv, comenv) == -1)
                {
                    perror("open_port");
                    Ethrowv("Failed to open port %s\n", command.u.atom);
                    exit(1);
                }
            }

        }
        else if (PortName.u.tuple->size == 3)
        {
            // { fd, In, Out }
            Ethrowv("Unimplemented 3-tuple port directive\n");
        }
        else
        {
            Ethrowv("Failed to open an unknown port with tuple of size %d\n", PortName.u.tuple->size);
        }
    }
    else if (PortName.tag & LT_atom) 
    {
        Ethrowv("Failed to open an unknown port %s\n", PortName.u.atom);
    }
    else
    {
        Ethrowv("Illegal type to arg 1 erlang:open_port/2\n");
    }
    return dummy;
    
end pragma;
    .


