
/*
** $Id: binary.c,v 1.23 2002/07/31 09:24:52 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/binary.c,v $
** $Revision: 1.23 $
** $Date: 2002/07/31 09:24:52 $
** $State: Exp $
**
** Author: tobbe@serc.rmit.edu.au
**
** Modifications: geoff@serc.rmit.edu.au
*/

#include <stdarg.h>
#include <assert.h>
#include <string.h>

#include "erlib.h"
#include "typecheck.h"

#ifndef _ALLOCS_H
EBinary binaryalloc(int size)
{
  EBinary v;
  v = (EBinary) malloc(sizeof(_Binary) + sizeof(char) * (size-1));
  v->size = size;
  return v;
}

Val EValbinary(int size)
{
  Val v;

  v.u.binary = binaryalloc(size);

  return v;
}

void binaryfree(EBinary b)
{
    free(b);
}

#endif

Val Emake_binary(unsigned char * buf, int sz)
{
    Val v;

    v = binarymalloc(sz);
    memcpy(v.u.binary->b, buf, sz);

    return v;
}


EBinary _binarymalloc(int size)
{
  EBinary v;
  v = (EBinary) malloc(sizeof(_Binary) + sizeof(char) * (size-1));
  v->size = size;
  return v;
}

Val binarymalloc(int size)
{
    Val v;
    
    v.tag = LT_binary | LT_on_heap;
    v.u.binary = _binarymalloc(size);
    return v;
}

Val EValmbinary(const void * buf, int size)
{
    Val v;

    v = binarymalloc(size);
    memcpy(v.u.binary->b, buf, size);

    return v;
}


Val Ebinary_to_list(Val b)
{
    Val v;
    int i, size, start = 0;
#ifdef _CONS_H
    _Lelement * lblock, * trav;
#endif

    TC(b, LT_binary, 1, "internal: Ebinary_to_list");
    size = b.u.binary->size;

    v = listmalloc(size);
#ifdef _CONS_H
    lblock = make_block_lelement(size);
    v.u.list = lblock;
    trav = lblock;
    i = 0;
    while (trav)
    {
        trav->e = EValint(b.u.binary->b[i++]);
        if (!start)
        {
            trav->e.tag = trav->e.tag | LT_on_heap;
            start = 1;
        }
        trav = trav->next;
    }
#else
    for (i = 0; i < size; i++) 
    {
        Eindex_list(v, i) = EValint(b.u.binary->b[i]);
    }
#endif

    return v;
}

Val Elist_to_binary(Val l)
{
    Val v, ret;
    int i=0,size;
#ifdef _CONS_H
    _Lelement * le;
#endif    


#ifdef _CONS_H
    le = l.u.list;
    size = 0;
    while (le != NULL)
    {
        size++;
        le = le->next;
    }
    ret = binarymalloc(size);

    le = l.u.list;
    while (le != NULL)
#else
    size = l.u.list->length;
    ret = binarymalloc(size);

    for (i = 0; i < size; i++) 
#endif
    {
#ifdef _CONS_H
        v = le->e;
        le = le->next;
#else
        v = Eindex_list(l, i);
#endif
        TC(v, LT_number, i, "list_to_binary/1: badarg");
        if (v.u.number < 256) 
            ret.u.binary->b[i] = v.u.number;
        else
            Ethrowv("list_to_binary/1: badarg, element %d exceeded 255", i);
#ifdef _CONS_H
        i++;
#endif
    }

    return ret;
}

int binarycmp(Val l, Val r)
{
    return  !(memcmp(l.u.binary->b, r.u.binary->b, l.u.binary->size));
}
