
/*
** $Id: call.c,v 1.37 2004/09/13 14:08:01 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/call.c,v $
** $Revision: 1.37 $
** $Date: 2004/09/13 14:08:01 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
**
**
*/


#include <string.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <signal.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "erlib.h"
#include "code.h"
#include "str.h"
#include "../config.h"

#define MAX_PARAMS 7
#define MAX_FUN_LENGTH 1024

Hash_table call_table = NULL;

void Einitcall()
{
#ifdef GC
    GC_enable_incremental();  
#endif    
    if (call_table == NULL)
    {
        // printf("creating new call table\n");
        call_table = make_hash_table(1001, code_hash, code_key, code_match);
    }
}

void Eclosecall()
{
    int i;
    Hash_element next;
    Code_handle ch;

    for (i = 0; i < call_table->size; i++)
    {
        next = call_table->table[i];
        while (next)
        {
            ch = (Code_handle) next->element; 
            if (ch) dlclose(ch->handle);
            next = next->next;
        }
    }
}

/*
** Number of params is built into the function name
** That way we know we're calling with the right number of parameters
*/
Val (*Esym(const char * module, const char * funstr))()
{
    Code_handle ch;
    void * handle = NULL;
    void * func;
    const char *error;

#if 0
fprintf(stderr, "Esym: %s, %s\n", module, funstr);
#endif
    /* handle should go into a hash table so we don't open  */
    /* it multiple times; that makes it easy to close 		*/
    ch = find_hash(call_table, module);
    if (ch != NULL) handle = ch->handle;

    if (handle == NULL)
    {
        handle = dlopen(module, RTLD_NOW | RTLD_GLOBAL);
        if (handle == NULL) 
        {
            Ethrowv("Failed to load module (%s) : %s\n", module, dlerror());
            return 0;
        }
        else
        {
            add_hash(call_table, make_code_handle(module,handle));
        }
    }

    func = dlsym(handle, funstr);
    if ((error = dlerror()) != NULL)  
    {
        Ethrowv("Unable to find :%s:%s:(%s)\n", module, funstr, error);
        printf("Unable to find :%s:%s:(%s)\n", module, funstr, error);
        return 0;
    }

#if 0
    /* FIX */
    Esendmonitor( '{ function, start, time }' );
#endif

    return func;
}

/*
** What's the cheapest way to pass args without counting the buggers?
*/
Val Ecall(const char * module, const char * funstr, int n, ...)
{
    va_list args;
	Val ret, a;
    int i;

    va_start(args, n);
#ifdef _CONS_H
    a = parammalloc(n);
#else
    a = listmalloc(n);
#endif

    /* make a call */
	for (i = 0; i < n; i++)
	{
#ifdef _CONS_H
		a.u.param->e[i] = va_arg(args, Val);
#else
        Eindex_list(a,i) = va_arg(args, Val);
#endif
	}

    va_end(args);
    ret = Ecalllist(module, funstr, a);
    listfree(a);

    return ret;
}


#define IL(X) Eindex_list(vlist,X)

// does this need to pass a message to the monitoring process?
Val Ecalllist(const char * module, const char * funstr, Val vlist)
{
    Val ret;
	void * func;
	Val (*func0)(), 
        (*func1)(Val), 
        (*func2)(Val,Val),
		(*func3)(Val,Val,Val), 
        (*func4)(Val,Val,Val,Val),
		(*func5)(Val,Val,Val,Val,Val), 
        (*func6)(Val,Val,Val,Val,Val,Val),
		(*func7)(Val,Val,Val,Val,Val,Val,Val), 
        (*func8)(Val,Val,Val,Val,Val,Val,Val,Val);

    func = Esym(module, funstr);

    // printf("#### vlist length=%d\n",vlist.u.list->length);
	switch (vlist.u.list->length)
	{
		case 0:
			func0 = func;
			ret = func0();
			break;
		case 1:
			func1 = func;
			ret = func1(IL(0));
			break; case 2:
			func2 = func;
			ret = func2(IL(0),IL(1));
			break;
		case 3:
			func3 = func;
			ret = func3(IL(0),IL(1),IL(2));
			break;
		case 4:
			func4 = func;
			ret = func4(IL(0),IL(1),IL(2),IL(3));
			break;
		case 5:
			func5 = func;
			ret = func5(IL(0),IL(1),IL(2),IL(3),IL(4));
			break;
		case 6:
			func6 = func;
			ret = func6(IL(0),IL(1),IL(2),IL(3),IL(4),IL(5));
			break;
		case 7:
			func7 = func;
			ret = func7(IL(0),IL(1),IL(2),IL(3),IL(4),IL(5),IL(6));
			break;
		case 8:
			func8 = func;
			ret = func8(IL(0),IL(1),IL(2),IL(3),IL(4),IL(5),IL(6),IL(7));
			break;
		default:
#ifdef _CONS_H
			fprintf(stderr, "Too many parameters (%d)to handle in this brain damaged call.\n", vlist.u.param->size);
#else
			fprintf(stderr, "Too many parameters (%d)to handle in this brain damaged call.\n", vlist.u.list->length);
#endif
			break;
	}
#if 0
//	printf ("%f\n", (*cosine)(2.0));
//	dlclose(handle);
#endif
	return ret;
}


#if 0
//
// Warning: this currently space leaks...
//
#endif
#define MAX_VALUE_LENGTH 128
char *  ESV(Val ret)
{
    char * staticbuf = malloc(MAX_VALUE_LENGTH);
    switch (ret.tag & LT_normal)
    {
    case LT_number:
        sprintf(staticbuf, "%d", ret.u.number);
        break;
    case LT_atom:
        sprintf(staticbuf, "'%s'", ret.u.atom);
        break;
    case LT_real:
        sprintf(staticbuf, "%f", ret.u.real);
        break;
    case LT_tuple:
        sprintf(staticbuf, "Tuple(%d)", ret.u.tuple->size);
        break;
    case LT_list:
        sprintf(staticbuf, "List(%d)", ret.u.list->length);
        break;
    case LT_bool:
        sprintf(staticbuf, "Boolean(%d)", ret.u.number);
        break;
    default:
        sprintf(staticbuf, "Unknown(tag=%d)", ret.tag);
        break;
    }
    return staticbuf;
}

void EPrintvalue(FILE * fp, Val ret)
{
    switch (ret.tag & LT_normal)
    {
    case LT_number:
        fprintf(fp, "%d", ret.u.number);
        break;
    case LT_atom:
        fprintf(fp, "%s", ret.u.atom);
        break;
    case LT_real:
        fprintf(fp, "%f", ret.u.real);
        break;
    case LT_tuple:
        {
            int i;
            fprintf(fp, "{ ");
            for (i = 0; i < ret.u.tuple->size; i++)
            {
                if (i != 0) printf(", ");
                EPrintvalue(fp, ret.u.tuple->e[i]);
            }
            fprintf(fp, " }");
        }
        break;
    case LT_list:
        {
            int i = 0;
#ifdef _CONS_H
            _Lelement * le;
            fprintf(fp, "[ ");
            le = ret.u.list;
            while (le)
            {
                if (i++ != 0) printf(", ");
                EPrintvalue(fp, le->e);
                le = le->next;
            }
            fprintf(fp, " ]");
#else
            fprintf(fp, "[ ");
            for (i = 0; i < ret.u.list->length; i++)
            {
                if (i != 0) fprintf(fp,", ");
                EPrintvalue(fp, Eindex_list(ret,i));
            }

            fprintf(fp, " ]");
#endif
        }
        break;
    case LT_binary:
        fprintf(fp, "Binary return value");
        break;
    case LT_bool:
        if (ret.u.number == 0)
            fprintf(fp, "false");
        else
            fprintf(fp, "true");
        break;
    default:
	    fprintf(fp, "Unimplemented EPrintvalue() return type: %d", ret.tag);
		break;
    }
}

#if 0
#define MAXBUF 2048

char * EValtoString(Val ret)
{
    char buf[MAXBUF];
    
    switch (ret.tag & LT_normal)
    {
    case LT_number:
        fprintf(fp, "%d", ret.u.number);
        break;
    case LT_atom:
        fprintf(fp, "%s", ret.u.atom);
        break;
    case LT_real:
        fprintf(fp, "%f", ret.u.real);
        break;
    case LT_tuple:
        {
            int i;
            fprintf(fp, "{ ");
            for (i = 0; i < ret.u.tuple->size; i++)
            {
                if (i != 0) printf(", ");
                EPrintvalue(fp, ret.u.tuple->e[i]);
            }
            fprintf(fp, " }");
        }
        break;
    case LT_list:
        {
            int i = 0;
#ifdef _CONS_H
            _Lelement * le;
            fprintf(fp, "[ ");
            le = ret.u.list;
            while (le)
            {
                if (i++ != 0) printf(", ");
                EPrintvalue(fp, le->e);
                le = le->next;
            }
            fprintf(fp, " ]");
#else
            fprintf(fp, "[ ");
            for (i = 0; i < ret.u.list->length; i++)
            {
                if (i != 0) fprintf(fp,", ");
                EPrintvalue(fp, Eindex_list(ret,i));
            }

            fprintf(fp, " ]");
#endif
        }
        break;
    case LT_binary:
        fprintf(fp, "Binary return value");
        break;
    case LT_bool:
        if (ret.u.number == 0)
            fprintf(fp, "false");
        else
            fprintf(fp, "true");
        break;
    default:
	    fprintf(fp, "Unimplemented EPrintvalue() return type: %d", ret.tag);
		break;
    }

    return string_copy(buf);
}
#endif

Val EParam(const char * str)
{
    int x;
    if (str[0] >= '0' && str[0] <= '9') 
    {
        x = atoi(str); 
        return EValint(x);
    }
    else
    {
        return EValmatom(str,strlen(str));
    }
}

char * Emangle(const char * mod, const char * fun, int params)
{
    char buf[MAX_FUN_LENGTH+1];
    char numbuf[12];

    strncpy(buf, mod, MAX_FUN_LENGTH);
    strncat(buf, "_", MAX_FUN_LENGTH);
    strncat(buf, fun, MAX_FUN_LENGTH);
    strncat(buf, "_", MAX_FUN_LENGTH);
    sprintf(numbuf, "%d", params);
    strncat(buf, numbuf, MAX_FUN_LENGTH);
    buf[MAX_FUN_LENGTH] = '\0';
    
    return string_copy(buf);
}


#define MAX_NODE 1024
Val Espawn(Val Module, Val Function, Val ArgumentList)
{
    extern Val _Eserverpid;
    Val ret;
    unsigned int addr = 0, port = 0;

    int pid = fork();

    srandom(time(0));

    /* check for child */
    if (pid == 0)
    {
        /* register ourselves with our home node */
        // MAYBE Einitserver(addr, port);

        Emonitorcall(Module.u.atom, Function.u.atom, ArgumentList);
        return EValint(0);
    }
    else if (pid > 0)
    {
        //while (waitpid(0, 0, WNOHANG) > 0);
#ifdef _ALLOCS_H
        return Emake_tuple(tuplemalloc(3), Eindex_tuple(_Eserverpid, 0), Eindex_tuple(_Eserverpid, 1), EValint(pid));
#else
        return Emake_tuple(3, Eindex_tuple(_Eserverpid, 0), Eindex_tuple(_Eserverpid, 1), EValint(pid));
#endif
    }
    Ethrowv("Failed to spawn process: bif:spawn/3\n");
    return ret;
}

Val Emonitorcall(const char * Module, const char * Function, Val ArgumentList)
{
    struct stat fbuf;
    Val ret;
    unsigned int magic_no;
    unsigned int addr = 0, port = 0;
    char * module, * function;
    int pid;

    extern Val _Eserverpid;

    (void) signal(SIGUSR1, SIG_IGN);
    magic_no = random();
        
    // Spawn monitoring process.
    // (should also be an option driven thing)
    module = string_cat(Module,"_mon.so");
    function = string_cat(Module,"_mon");

    // Check there is a monitoring module 
    stat(module, &fbuf);

    // printf("fbuf.st_mode=%d\n", fbuf.st_mode);
    if ((fbuf.st_mode & S_IRUSR) && (fbuf.st_size > 0) && MonitoringOn())
    {
        //printf("Spawning monitor\n");
        pid = fork();
        if (pid > 0)
        {
            //printf("  setting monitor %d\n", pid);
            Esetmonitor(pid);
        }
        else if (pid == 0)
        {
            // MAYBE Einitserver(addr, port);
            // FIX: we should get the pid of the process were being spawned
            // from instead of an empty list
            Ecalllist(module, Emangle(function, "start", 1), Emake_list(1, EValcopy(_Eserverpid)));
            free(module);
            free(function);
            //printf("finished monitor\n");
            // while (waitpid(0, 0, WNOHANG) > 0);
            Ecloseserver();
            exit(0);
        }
        else
        {
            perror("fork monitor");
        }
    }

    free(module);
    free(function);


    // Startup the spawned process
    // fun = Ereceivemagic((unsigned int) magic_no);
    module = string_cat(Module,".so");
    //printf("Running main function (spawned %d)\n", pid);
    Ecalllist(module,
            Emangle(Module, Function, 
            ArgumentList.u.list->length), EValcopy(ArgumentList));
    free(module);

    // are we now finished if we are a child?
    // printf("finished main function\n");
    // while (waitpid(0, 0, WNOHANG) > 0);
    Ecloseserver();
    exit(0);
    return ret;
}
