
/*
** $Id: clist.c,v 1.19 2002/11/06 12:47:30 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/clist.c,v $
** $Revision: 1.19 $
** $Date: 2002/11/06 12:47:30 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright (C) 2001 
** geoff@cs.rmit.edu.au
*/

/* Erlang List Support */
/* NOTES:  needs to be made more efficient */
/* 	should have shared/unshared and allocated blocks */

#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include "erlib.h"
#include "typecheck.h"

#define LIST_BLOCK 32

int list_inited = 0;
Val empty_list;


List listalloc()
{
    // big number of refs so it doesn't get freed 
    // (it shoudln't anyway)..
    List newl;
    newl = (List) malloc(sizeof(_List));
    newl->ref = 1;
    newl->e = NULL;
    newl->length = 0;
    newl->start = 0;

    return newl;
}


/* actually creates a list */
Val EValmstring(const char * atom)
{
	Val v;
    int n, i, count;

    n = strlen(atom);
	v = listmalloc(n);
    count = v.u.list->start;
	for (i = 0; i < n; i++)
	{
		v.u.list->e->mem[count++] = EValint(atom[i]);
	}

	return v;
}

Val EValstring(const char * atom)
{
    Val V = EValmstring(atom);
    return V;
}

/* Take a chunk of bytes and turn them into an Erlang list */
Val Ebytes_to_list(const char *bytes, int size)
{
  Val v;
  int i;

  v = listmalloc(size);
  for (i = 0; i < size; i++) 
  {
    v.u.list->e->mem[v.u.list->start+i] = EValint(bytes[i]); 
  }

  return v;
}

void init_empty()
{
    empty_list.tag = LT_list;
        // "not" on heap should not get freed!
    empty_list.u.list = listalloc();
    empty_list.u.list->ref = 1 << 8;
        // a big # just in case ..
    list_inited = 1;
}

inline Val listmalloc(unsigned int sz)
{
	Val v;
    unsigned int allocsz;

    // assert(sz >= 0);

    /* special case - shared don't allocate memory */
    if (sz == 0)
    {
        if (list_inited == 0) init_empty();
        return empty_list;
    }

    /* twice the size up to next 32 size boundary up */
    allocsz = ((sz<<1)+(LIST_BLOCK-1)) & (~(LIST_BLOCK));

	v.tag = LT_list | LT_on_heap;
	v.u.list = (List) malloc(sizeof(_List));
    /* memset(v->u.list, 0, sizeof(_List)); */

	v.u.list->ref = 1;
	v.u.list->length = sz;
    if (allocsz) 
    {
        v.u.list->e = (ListMember)malloc(sizeof(_List_member) 
                        + (allocsz - 1) * sizeof(Val));
        memset(v.u.list->e->mem, 0, allocsz * sizeof(Val));
	    v.u.list->start = allocsz/2;
	    v.u.list->e->size = allocsz;
	    v.u.list->e->first = allocsz/2;
        v.u.list->e->ref = 1;
    }
    else
    {
        v.u.list->e = NULL;
        v.u.list->start = 0;
    }

	return v;
}

#if 0
void listrealloc(List l, int sz)
{
    int allocsz;

    allocsz = ((sz<<1)+(LIST_BLOCK-1)) & (~(LIST_BLOCK));

	l->size = allocsz;
	l->length = sz;

    l->start = (Val *) malloc(allocsz * sizeof(Val));
    memset(l->start, 0, allocsz * sizeof(Val));
#if 0
    memcpy(l.u.list->start,
	v.u.list->e = allocsz/2;
#endif
}
#endif


inline Val Econcatvl(Val item, Val l)
{
    Val v;

    TC(l, LT_list, 2, "Econcatvl");

    if ((l.u.list->length > 0 )
            && (l.u.list->e->first > 0)
            && (l.u.list->start == l.u.list->e->first))
    {
        v.tag = LT_list | LT_on_heap;
        v.u.list = listalloc();
        memcpy(v.u.list, l.u.list, sizeof(_List)); 
        v.u.list->ref = 1;
        v.u.list->length++;
        v.u.list->start = l.u.list->start - 1;
	    v.u.list->e->mem[v.u.list->start] = (item);
        v.u.list->e->first--;
        v.u.list->e->ref++;
        return v;
    }
    else
    {
        // printf("concatvl length=%d\n", l.u.list->length+1);
        v = listmalloc(l.u.list->length + 1);

	    v.u.list->e->mem[v.u.list->start] = (item);
        if (l.u.list->length)
        {
            int i, c, d;

            c = v.u.list->start + 1;
            d = l.u.list->start;
            for (i = 0; i < l.u.list->length; i++)
            {
                v.u.list->e->mem[c++] = EValcopy(l.u.list->e->mem[d++]);
            }
        }

	    return v;
    }
}

Val Econcat(Val h, Val t)
{
    Val v;
	int sz;

//  should already be type checked before this is called (effectively)
    TC(h, LT_list, 1, "Econcat");
    TC(t, LT_list, 2, "Econcat");

    if (t.u.list->length == 0)
    {
        if (h.u.list->length == 0) 
        {                
            if (list_inited == 0) init_empty();
            return empty_list;
        }

        v.tag = LT_list | LT_on_heap;
        v.u.list = listalloc();
        v.u.list->length = h.u.list->length;
        v.u.list->start = h.u.list->start;
        v.u.list->e = h.u.list->e;
        if (v.u.list->e) v.u.list->e->ref++;
        return v;
    }

    if (h.u.list->length == 0)
    {
        v.tag = LT_list | LT_on_heap;
        v.u.list = listalloc();
        v.u.list->length = t.u.list->length;
        v.u.list->start = t.u.list->start;
        v.u.list->e = t.u.list->e;
        if (v.u.list->e) v.u.list->e->ref++;
        return v;
    }
    

	sz = h.u.list->length + t.u.list->length;
	
    // FIX: what about the case of adding t onto end of h?
    //      perhaps need to track "last" as well as first (sigh)
    if ((h.u.list->length < t.u.list->e->first)
                    && (t.u.list->start == t.u.list->e->first))
    {
        int i, c, d;

        c = t.u.list->start - h.u.list->length;
        d = h.u.list->start;
        for (i = 0; i < h.u.list->length; i++)
        {
            t.u.list->e->mem[c++] = EValcopy(h.u.list->e->mem[d++]);
        }

        v.tag = LT_list | LT_on_heap;
        v.u.list = listalloc();
        v.u.list->length = sz;
        v.u.list->start = t.u.list->start - h.u.list->length;
        v.u.list->e = t.u.list->e;
        v.u.list->e->ref++;
        v.u.list->e->first = t.u.list->start - h.u.list->length;
    }
    else
    {
        int i, c, d;

        v = listmalloc(sz);

        c = v.u.list->start;
        d = h.u.list->start;
        for (i = 0; i < h.u.list->length; i++)
        {
            v.u.list->e->mem[c++] = EValcopy(h.u.list->e->mem[d++]);
        }

        c = v.u.list->start + h.u.list->length;
        d = t.u.list->start;
        for (i = 0; i < t.u.list->length; i++)
        {
            v.u.list->e->mem[c++] = EValcopy(t.u.list->e->mem[d++]);
        }
    }

	return v;
}

//
// FIX: broken.
//
// The - - operator produces a list which is a copy of the first argument, 
// subjected to the following procedure: for each element in the second 
// argument, its first occurrence in the first argument is removed.
//      X = [1,2,3,2,1,2],
//      Y = [2,1,2],
//      X -- Y.
// results in [3,1,2]. 
//
Val Elistsub(Val h, Val t)
{
    int i,j,hc,nc,from;
    Val nv;

    // it'll probably be shorter than h ...
    // just allocate anyway.
    nv = listmalloc(h.u.list->length);

    nc = nv.u.list->start;
    hc = h.u.list->start;

#if 0
    // FIX:
    for (j = 0; j < h.u.list->length; j++)
    {
        if (Elistfinditem(t, h.u.list->e->mem[hc], from) == -1)
            nv.u.list->e->mem[nc++] = EValcopy(h.u.list->e->mem[hc++]);
        else
            from++;
    }
#endif

    return nv;
}

Val Eappendv(Val h, Val t)
{
    assert(1==0);
}

inline Val	Emake_list(int n, ...)
{
	Val v;
	int i, count;
	va_list args;
	
    if (n == 0) 
    {
        if (list_inited == 0) init_empty();
        return empty_list;
    }

	v = listmalloc(n);
	va_start(args, n);
    count = v.u.list->start;
	for (i = 0; i < n; i++)
	{
		v.u.list->e->mem[count++] = EValcopy(va_arg(args, Val));
	}
	va_end(args);

	return v;
}

#if 0
// destructive tail_list 
inline Val Etail_listd(Val l, int arg, const char * from)
{
    if (l.tag & LT_on_heap) 
    {
        Val r;
        r.tag = LT_list;
        r.u.list = listalloc();
        memcpy(r.u.list, l.u.list, sizeof(_List));
        r.u.list->length--;
        r.u.list->start++;
        return r;
    }
    l.u.list->length--;
    l.u.list->start++;
    // l.u.list->e->ref++;
    return l;
}
#endif

inline void Efred(Val l)
{
    if (l.tag & LT_on_heap) return Efre(l);
    if (l.tag & LT_list) 
    {
        if (!l.u.list->length) return;
        //l.u.list->ref--;
        //if (l.u.list->ref == 0) 
        //free(l.u.list);
    }
}

/* returns the tail of a list (a list itself) */
inline Val	Etail_list(Val l, int arg, const char * from)
{
    Val ret;

	// TC(l, LT_list, arg, from);
    // assert(ret.u.list->size >= 0);
    // assert(l.u.list->size >= 0);
    //
    if (!l.u.list->length) 
    {
        if (list_inited == 0) init_empty();
        return empty_list;
    }
    ret.tag = LT_list | LT_on_heap;
    ret.u.list = listalloc();

    memcpy(ret.u.list, l.u.list, sizeof(_List));
    ret.u.list->length--;
    ret.u.list->start++;
    ret.u.list->ref = 1;

    // FIX: should check against MAXINT
    ret.u.list->e->ref++;

    return ret;
}

Val	Entail_list(Val l, int tailx, const char * from)
{
	// TC(l, LT_list, arg, from);

    // assert(ret.u.list->size >= 0);
    // assert(l.u.list->size >= 0);

    if (l.u.list->length > tailx)
    {
        Val ret;
        ret.u.list = listalloc();
        memcpy(ret.u.list, l.u.list, sizeof(_List));
        ret.tag = LT_list | LT_on_heap;
        ret.u.list->length = ret.u.list->length - tailx;
        ret.u.list->start = l.u.list->start + tailx;
        ret.u.list->ref = 1;
        ret.u.list->e->ref++;
        return ret;
    }

    if (list_inited == 0) init_empty();
    return empty_list;
}

inline int	Eis_empty_list(Val l)
{
	/* TC(l, LT_list, 1, "is_empty_list"); */
    if ((l.tag & LT_list) == 0) return 0;

	if ((l.u.list == NULL) || (l.u.list->length == 0)) return 1;
	return 0;
}

inline int	Eis_full_list(Val l)
{
	/* TC(l, LT_list, 1, "is_full_list"); */
    if ((l.tag & LT_list) == 0) return 0;

	if (l.u.list && (l.u.list->length > 0)) return 1;
	return 0;
}


/* Transforms an "Erlang string" into a "C string" */

char *Elist_to_string(Val l, int arg, const char * from)
{
  int i,size;
  char *cp=NULL;
  Val v;

  TC(l, LT_list, arg, from);
  size = l.u.list->length;
  cp = (char *) malloc(size + 1);
  if (cp == NULL) Ethrowv("Out of memory (Elist_to_string)\n");

  for (i = 0; i < size; i++) 
  {
    v = l.u.list->e->mem[l.u.list->start+i];
    cp[i] = v.u.number;
  }
  cp[i] = '\0';

  return cp;
}

Val parammalloc(int sz)
{
    Val v;
    assert(sz >= 0);

    v.tag = LT_list | LT_on_heap;
//  v.tag = LT_list;
    v.u.param = (ParamList)malloc(sizeof(_ParamList));
    v.u.param->size =sz;
    if (sz) v.u.param->e = (Val *)malloc(sizeof(Val)*sz);
    else v.u.param->e = NULL;
    return v;
}

//
// FIX: may recurse into lists/tuples infinitly?
//
int listcmp(List l, List r)
{
    int i;
    //printf("clist listcmp (%d,%d)\n", l->length, r->length);

    if (l->length != r->length) return 0;
    if ((int)(&(l->e->mem[l->start])) ==
        (int)(&(r->e->mem[r->start]))) return 1;

    for (i = 0; i < l->length; i++)
    {
        if (!(Eequal(l->e->mem[l->start+i],
                    r->e->mem[r->start+i]))) return 0;
    }

    return 1;
}


inline void listfree(Val v)
{
    if (v.u.list)
    {
        v.u.list->ref--;
        if (v.u.list->ref == 0)
        {
            if (v.u.list->e && v.u.list->e->ref)
            {
                v.u.list->e->ref--;
                if (v.u.list->e->ref == 0) 
                {
                    int i;
                    for (i = v.u.list->e->first; (i < v.u.list->e->size); i++)
                    {
                        if (v.u.list->e->mem[i].tag != LT_none)
                            Efre(v.u.list->e->mem[i]);
                    }
                    free(v.u.list->e);
                }
                v.u.list->e = NULL;
            }
            free(v.u.list);
            v.u.list = NULL;
        }
    }
}

#if 0
int Enotequalvl(Val v, List l)
{
    if (!(v.tag & LT_list)) return 0;
    return listcmp(v.u.list, l);
}
#endif
