
/*
** $Id: cons.c,v 1.19 1999/01/13 00:54:20 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/cons.c,v $
** $Revision: 1.19 $
** $Date: 1999/01/13 00:54:20 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
*/

#include <stdarg.h>
#include <assert.h>
#include "erlib.h"
#include "typecheck.h"
#include "cons.h"

/* allocation stuff */

_Lelement * make_lelement(Val x)
{
    _Lelement * ret;

    ret = (_Lelement *) malloc(sizeof(struct _lelement));
    assert(ret != NULL);
    ret->next = NULL;
    ret->e = EValcopy(x);
    ret->e.tag = ret->e.tag | LT_on_heap; 
    return ret;
}

_Lelement * make_lelementvl(Val x, _Lelement * nl)
{
    _Lelement * ret;

    ret = (_Lelement *) malloc(sizeof(struct _lelement));
    assert(ret != NULL);
    ret->next = nl;
    ret->e = EValcopy(x);
    ret->e.tag = ret->e.tag | LT_on_heap; 
    return ret;
}

_Lelement * make_block_lelement(int n)
{
    _Lelement * ret, * tel;
    int i;

#if 0
    /* ideally we could do this for entire block deallocation .. */
    ret = (_Lelement *) malloc(sizeof(struct _lelement) * n);
#endif

    ret = NULL;
    for (i = 0; i < n; i++)
    {
        tel = (_Lelement *) malloc(sizeof(struct _lelement));
        tel->next = ret;
        ret = tel;
    }

    return ret;
}

/* actually creates a list */
Val EValmstring(const char * atom)
{
    Val v;
    int n, i, start = 0;
    _Lelement * lblock, * trav;

    n = strlen(atom);
    v.tag = LT_list | LT_on_heap;

    if (n)
    {
        lblock = make_block_lelement(n);
        v.u.list = lblock;
        trav = lblock;
        i = 0;
        while (trav)
        {
            trav->e = EValint(atom[i++]);
            if (!start)
            {
                trav->e.tag |= LT_on_heap;
                start = 1;
            }
            trav = trav->next;
        }
    }
    else
    {
        v.u.list = NULL;
    }
    return v;
}

Val listmalloc(int sz)
{
    Val v;

    assert(sz >= 0);

    v.tag = LT_list | LT_on_heap;
    v.u.list = NULL;
    return v;
}

Val Econcatvl(Val v, Val item, Val l)
{
    _Lelement * le;
    
    if (l.u.list)
    {
        le = make_lelementvl(item, l.u.list);
        l.u.list->e.tag = l.u.list->e.tag & ~LT_on_heap;
    }
    else
    {
        le = make_lelement(item);
    }
    v.tag = LT_list | LT_on_heap;
    v.u.list = le;

    return v;
}

Val Econcathl(Val v, _Lelement * le, Val l)
{
    if (l.u.list != NULL)
        l.u.list->e.tag = l.u.list->e.tag & ~LT_on_heap;
    le->next = l.u.list;
    v.tag = LT_list | LT_on_heap;
    v.u.list = le;
    v.u.list->e.tag = v.u.list->e.tag | LT_on_heap; 

    return v;
}

Val Econcat(Val v, Val h, Val t)
{
    _Lelement * le;

/*  should already be type checked before this is called (effectively) */
    TC(h, LT_list, 1, "Econcat arg 1");
    TC(t, LT_list, 1, "Econcat arg 2");

    le = h.u.list;
    if (le == NULL) 
    {
        return t;
    }

    while (le->next != NULL) 
    {
        le = le->next;
    }

    if (t.u.list != NULL)
    {
        le->next = t.u.list;
        t.u.list->e.tag = t.u.list->e.tag & ~LT_on_heap;
    }

    return h;
}

Val Eappendv(Val h, Val t)
{
    assert(1 == 0);
}

Val    Emake_list(int n, ...)
{
    Val v;
    int i, start = 0;
    _Lelement * lblock;
    _Lelement * trav;
    va_list args;
    va_start(args, n);

    v.tag = LT_list | LT_on_heap;

    if (n != 0)
    {
        lblock = make_block_lelement(n);
        v.u.list = lblock;
        trav = lblock;
        i = 0;
        while (trav)
        {
            trav->e = EValcopy(va_arg(args, Val));
            if (start) trav->e.tag = trav->e.tag & ~LT_on_heap;
            else trav->e.tag = trav->e.tag | LT_on_heap;
            start = 1;
            trav = trav->next;
        }
    }
    else
    {
        v.u.list = NULL;
    }
    va_end(args);
    return v;
}
    

/* returns the tail of a list (a list itself) */
Val    Etail_list(Val ret, Val l, int arg, const char * from)
{
#if 0
    // TC(l, LT_list, arg, from);
    // assert(ret.u.list->size >= 0);
    // assert(l.u.list->size >= 0);
#endif

    ret.tag = LT_list;
    ret.u.list = l.u.list->next;
    return ret;
}

Val    Entail_list(Val ret, _Lelement * le, int tailx, const char * from)
{

#if 0
    int count = 0;
    _Lelement * le = l.u.list;
    while(count < tailx && le)
    {
        le = le->next;
    }
#endif

    ret.tag = LT_list;
    ret.u.list = le;

    return ret;
}

#if 0
int    Eis_empty_list(Val l)
{
    /* TC(l, LT_list, 1, "is_empty_list"); */
    if ((l.tag & LT_list) == 0) return 0;

    if (l.u.list->size == 0) 
    {
        return 1;
    }
    return 0;
}

int    Eis_full_list(Val l)
{
    /* TC(l, LT_list, 1, "is_full_list"); */
    if ((l.tag & LT_list) == 0) return 0;

    if (l.u.list->size > 0)
    {
        return 1;
    }
    return 0;
}
#endif


/* Transforms an "Erlang string" into a "C string" */

char *Elist_to_string(Val l, int arg, const char * from)
{
  int i,size;
  char *cp=NULL;
  _Lelement * le;

  TC(l, LT_list, arg, from);
  le = l.u.list;
  size = 0;
  while (le != NULL)
  {
    size++;
    le = le->next;
  }
  cp = (char *) malloc(size + 1);
  if (cp == NULL) Ethrowv("Out of memory (Elist_to_string)\n");

  le = l.u.list;
  for (i = 0; i < size; i++) 
  {
    cp[i] = le->e.u.number;
    le = le->next;
  }
  cp[i] = '\0';

  return cp;
}

Val parammalloc(int sz)
{
    Val v;
    assert(sz >= 0);

    v.tag = LT_list | LT_on_heap;
    v.u.param = (ParamList)malloc(sizeof(_ParamList));
    v.u.param->size =sz;
    if (sz) v.u.param->e = (Val *)malloc(sizeof(Val)*sz);
    else v.u.param->e = NULL;
    return v;
}

Val EValstring(Val V,const char * _atom) 
{
    _Lelement * le;
    int count = 0; 
    V.tag = LT_list | LT_on_heap; 
    V.u.list = make_block_lelement(strlen(_atom)); 
    le = V.u.list; 
    while (le) 
    { 
        le->e = EValint(_atom[count]); 
        le = le->next; 
        count++; 
    } 
    return V; 
}

