
/*
** $Id: ehash.h,v 1.1 1998/05/08 05:49:26 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/ehash.h,v $
** $Revision: 1.1 $
** $Date: 1998/05/08 05:49:26 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
*/

#ifndef _EHASH_H
#define _EHASH_H

/*
** Structs for an external hash table
** First the bucket for holding elements and then the table
** definition itself.
*/

typedef struct HASH_ELEMENT 
{
    void * element;
    struct HASH_ELEMENT * next;
} * Hash_element;


typedef struct HASH_TABLE 
{
    int entries;
    int size;
    int (*hash_element)(struct HASH_TABLE *, void * element);
    int (*hash_key)(struct HASH_TABLE *, void * key);
    int (*match)(struct HASH_TABLE *, void * element, void * key);
    Hash_element * table;
} * Hash_table;

extern Hash_table make_hash_table(int size, int (*hashelement)(), 
                    int (*hashkey)(), int (*match)());
    /* hash table constructor */

extern void   free_hash_table(Hash_table HT);
    /* hash table destructor */

extern int    remove_hash(Hash_table HT, void * key);
    /* remove a key from the hash table */

extern int    add_hash(Hash_table HT, void * element);
    /* add a key to the hash table */

extern void * find_hash(Hash_table HT, void * key);
    /* find a key in the hash table */

extern int    hash_upsize(Hash_table HT);
    /* increase the size of the hash table */

extern unsigned int hashstr(char *s, int hashsize);
    /* return a raw hash value (of the given string) */

#endif
