

/*
** $Id: erlib.h,v 1.95 2002/12/09 13:12:09 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/erlib.h,v $
** $Revision: 1.95 $
** $Date: 2002/12/09 13:12:09 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1997, 1998
** geoff@serc.rmit.edu.au
**
** Erlang Support Library Code
**
**  This defines the key elements of a shared library on target system.
**
*/

#ifndef _ERLIB_H
#define _ERLIB_H
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <setjmp.h>
#include "malloc.h"
#include "stack.h"

#ifndef _GNU_C_
#define inline
#define __inline__
#endif

#if defined(__cplusplus) || defined(c_plusplus)
extern "C" {
#endif

#include "etype.h"

struct _list;
struct _tuple;
struct _binary;
struct _paramlist;

typedef union 
{
    int number;
    float real;
    char * atom;    
    char * string;
    char * var;
    struct _tuple * tuple;
#if 0
    struct _lelement * list;
#else
    struct _list * list;
#endif
    struct _paramlist * param;
    struct _binary * binary;
    void * any;
} Many;

typedef struct value 
{
    Etype tag;
    Many u;
} Val;

/* Some built in guard tests for speed */
#define E_float_1(X)    ((X).tag & LT_real)
#define E_number_1(X)   ((X).tag & LT_number)
#define E_integer_1(X)  ((X).tag & LT_number)
#define E_atom_1(X)     ((X).tag & LT_atom)
#define E_tuple_1(X)    ((X).tag & LT_tuple)
#define E_list_1(X)     ((X).tag & LT_list)
#define E_binary_1(X)   ((X).tag & LT_binary)

#if 0
#include "cons.h"
#include "allocs.h"
typedef struct _tuple 
{
    int size;
    Val e[1];
} _Tuple;
typedef struct _tuple * ETuple;
#else

#include "clist.h"
#include "tuple.h"
#endif


// #include "allocs.h"

typedef struct _binary 
{
    int size;
    unsigned char b[1];
} _Binary;
typedef struct _binary * EBinary;

/* Definition for imports/exports */
typedef struct symbol
{
    char * module;
    int parameters;
} _Symbol;


/* Support (allocation) stuff */
Val EValint(int),
#ifndef _ALLOCS_H
    EValbinary(int),
    EValatom(const char *),
    EValnatom(const char *, int len),
    EValstring(const char *),
#else
    EValstring(Val, const char *),
#endif
    EValmstring(const char *),
    EValmbinary(const void *, int),
    EValmatom(const char *, int len),
    EValbool(int),
    EValreal(float),
    EValfunc(void *),
    EValcopy(Val),
    EValcopyin(Val, Val)
    ;

/* Support for deallocation */
void Efre(Val),
    ENomonitor(),
    Esetmonitor(int),
    EPrintvalue(FILE *, Val)
    ;

/* Basic Operator Support */
Val Emultiply(Val, Val),
    Emultiplyvi(Val, int),
    Emultiplyiv(int, Val),
    Emultiplyrv(float, Val),
    Emultiplyvr(Val, float),
    Eminus(Val, Val),
    Eminusvi(Val, int),
    Eminusiv(int, Val),
    Eminusir(int, float),
    Eminusri(float, int),
    Eminusvr(Val, float),
    Eminusrv(float, Val),
    Edivide(Val, Val),
    Edivideir(int, float),
    Edivideri(float, int),
    Edividerv(float, Val),
    Edividevr(Val, float),
    Edividevi(Val, int),
    Ediv(Val, Val),
    Eplus(Val, Val),
    Eplusvi(Val, int),
    Eplusiv(int, Val),
    Eplusrv(float, Val),
    Eplusvr(Val, float),
    Eplusrr(float, float),
    Eplusii(int, int),
    Erem(Val, Val),
    Eband(Val, Val),
    Ebor(Val, Val),
    Ebxor(Val, Val),
    Ebsl(Val, Val),
    Ebsr(Val, Val),
    Ebnot(Val),
    Epreplus(Val),
    Epreminus(Val),
    Emultiplyii(int, int),
    Eminusrr(float, float),
    Eminusii(int, int)
    ;


/* Boolean Pattern Matching Support */
int
    MonitoringOn(),
    Enotequal(Val, Val),
    Enotequalvi(Val, int),
    Enotequalva(Val, char *),
    Enotequalvl(Val, List),
    Elessequal(Val, Val),
    Elessequalii(int, int),
    Elessequalvi(Val, int),
    Elessequalrr(float, float),
    Elessthan(Val, Val),
    Elessthanvi(Val, int),
    Elessequalvr(Val, float),
    Egreaterequal(Val, Val),
    Egreaterequalii(int, int),
    Egreaterequalvi(Val, int),
    Egreaterthan(Val, Val),
    Egreaterthanrr(float, float),
    Egreaterthanvi(Val, int),
    Egreaterthanii(int, int),
    Eequal(Val, Val),
    Eequalvi(Val, int),
    Eequalvs(Val, const char *),
    Eequalbi(Val, int),
    Eequalii(int, int),
    Eexactequal(Val, Val),
#ifndef _CONS_H
    Eis_empty_list(Val),
    Eis_full_list(Val),
    Eis_size_list(Val, int),
#endif
    Eis_tuple(int, Val),
    Etrue(Val),
    Estrcmp(Val, const char *),
    tuplecmp(Val, Val),
    atomcmp(Val, Val),
    binarycmp(Val, Val)
    ;

/* List Support */
Val 
    Econcat(Val, Val),
    Espawn(Val, Val, Val),
    Emonitorcall(const char *, const char *, Val),
    Elistsub(Val, Val),
    Emake_list(int, ...),
#ifdef _ALLOCS_H
#ifdef _CONS_H
    Econcathl(Val,_Lelement *, Val),
    Entail_list(Val, _Lelement *, int tailx, const char * from),
#endif
    Etail_list(Val, Val, int arg, const char * from),
    Entail_list(Val ret, Val l, int tailx, const char * from),
    Ebytes_to_list(const char *, int),
#else
    Etail_list(Val, int arg, const char * from),
    Entail_list(Val l, int tailx, const char * from),
    Ebytes_to_list(const char *, int),
    Econcatvl(Val, Val),
    // Etail_list(Val, int arg, const char * from),
    // Entail_list(Val, int arg, const char * from),
    // Eindex_list(Val, int),
#endif
    Emake_binary(unsigned char *, int sz),
    Eappendv(Val, Val)
    ;

char * Elist_to_string(Val, int arg, const char * from),
     * Emangle(const char * mod, const char * fun, int params),
     * ESV(Val)
     ;

/* Tuple Support */
Val 
#ifdef _ALLOCS_H
    Emake_tuple(Val n, ...)
#else
    Emake_tuple(int n, ...)
    // Eindex_tuple(Val, int)
#endif
    ;

/* Binary Support */
Val Ebinary_to_list(Val),
    Elist_to_binary(Val)
    ;

/* Send/Receive Support */
Val Esend(Val pid, Val message),
    Esendmonitor(Val mes),
    Ereceive(Val waittime),
    Ereceivemagic(int)
    ;

extern Val zero;

/* Only called on startup via `gerl' which spawns processes */
void Einitserver(unsigned int addr, unsigned int port),
     Ecloseserver(),
     Eclosecall(),
     Einitmonitor(unsigned int addr, unsigned int port, unsigned int pid);

/* Error Support */
Val Ethrowv(const char *, ...),
    Ethrow(Val)
    ;

Val Emake_catch(),
    Epop_catch(Val)
#ifndef _ALLOCS_H
    ,
    Eokay(),
    Eok(Val),
    Eerror(const char *)
#endif
    ;

/* Static startup support - converts a string to a Val */
Val EParam(const char *);

jmp_buf * Epush_catch()
    ;


/* External Module Support */
Val Ecall(const char *  module, const char *  func, int n, ...),
    Ecalllist(const char *  module, const char *  func, Val vlist),
    (*Esym(const char * module, const char * func))()
    ;


/* Private Stuff */
void * Eopen_module(const char *)
    ;

/* Allocation of Vals */
Val valalloc(Etype),
    listmalloc(unsigned int),
    binarymalloc(int),
    tuplemalloc(unsigned int)
#ifndef _ALLOCS_H
    ,
    tuplealloc(int)
#endif
    ;

#ifndef _ALLOCS_H
EBinary binaryalloc(int size);
void    binaryfree(EBinary b);
#endif
EBinary _binarymalloc(int size);

#include "typecheck.h"

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#endif /* _ERLIB_H */

