

/*
** $Id: error.c,v 1.15 2002/12/11 12:32:01 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/error.c,v $
** $Revision: 1.15 $
** $Date: 2002/12/11 12:32:01 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
**
** Error support and handling stuff
** Does each process need it's own error stack? (yes)
*/

#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include "erlib.h"
#include "monitor.h"

#define MAX_BUF 4096
#define MAX_STACK 256

/* 
 * MAX_STACK limit could be a problem in recursive functions
 * with a catch(). They should clear this as part of LCO.
 * alternatively the jmp_bufs should be on the C stack somehow..
 */

jmp_buf * Ejmp_stack[MAX_STACK];
Val       Eerror_stack[MAX_STACK];

int Ejmp_ptr = 0;
int Eerr_ptr = 0;

Val Ethrow(Val v)
{
    assert(Ejmp_ptr > 0);
    assert(Eerr_ptr < MAX_STACK - 1);

    // printf("Ethrow: %s\n", v.u.atom);
    Eerror_stack[Eerr_ptr++] = v;
    // Send a duplicate to the monitoring node 
    Esendmonitor(v);

    // Do the longjump
    longjmp(*(Ejmp_stack[--Ejmp_ptr]), 1);
    return v;
}

Val Ethrowv(const char * fmt, ...)
{
    Val v;
    va_list args;
    char buf[MAX_BUF];

    assert(Ejmp_ptr > 0);
    assert(Eerr_ptr < MAX_STACK - 1);

    va_start(args, fmt);
    (void) vsprintf(buf, fmt, args);
    va_end(args);

    // printf("Ethrowv: %s\n", buf);
    v = EValmatom(buf, strlen(buf));

    Eerror_stack[Eerr_ptr++] = v;
    // Send a duplicate to the monitoring node 
    Esendmonitor(v);

    // Do the longjump
    longjmp(*(Ejmp_stack[--Ejmp_ptr]), 1);
    return v;
}

jmp_buf * Epush_catch()
{
    /* create a new jmp_buf and push on stack */
    jmp_buf * tmp;

    assert(Ejmp_ptr < MAX_STACK - 1);

    tmp = (jmp_buf *)malloc(sizeof(jmp_buf));
    assert(Ejmp_ptr < MAX_STACK-1);
    Ejmp_stack[Ejmp_ptr++] = tmp;

    return tmp;
}

Val Epop_catch(Val x)
{
    /* pop stack */
    assert(Ejmp_ptr > 0);

    free(Ejmp_stack[--Ejmp_ptr]);
    return x;
}

Val Emake_catch()
{
    /* pop stack */
    Val ret;
    assert(Eerr_ptr > 0);
    --Eerr_ptr;
    ret = Eerror_stack[Eerr_ptr];

    // FIX: Send a duplicate to the monitoring node 
    // Esendmonitor(v);

    return ret;
}

