/*
** $Id: esocket.c,v 1.15 2002/12/11 05:09:28 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/esocket.c,v $
** $Revision: 1.15 $
** $Date: 2002/12/11 05:09:28 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
*/

#include <assert.h>

#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <strings.h>
#ifdef SunOS
#include <inet/tcp.h>
#include <netinet/in.h>
#else
#include <netinet/tcp.h>
#endif

#include "esocket.h"
#include "message.h"
#include "str.h"

 
Socket Ecreate_socket(int domain, int tipe, int prot)
{
    Socket ret;

    ret = (Socket) malloc(sizeof(struct _socket));

    ret->socket_no = socket(domain, tipe, prot);
    // printf("%d creating socket %d\n", getpid(), ret->socket_no);
    ret->family = domain;
    ret->socket_type = tipe;
    ret->protocol = prot;
    ret->int_address = 0;
    ret->port = 0;

    ret->hostname = NULL;
    ret->error = NULL;

    return ret;
}
 
/* NOTE: should we should TCP_NODELAY for TCP connections in here? */
Socket Eaccept(Socket S)
{
    struct sockaddr_in addr;
    int             length, new_socket;
    Socket o;
 
    /* make sure we're bound to a socket */
    if (S->socket_no == 0) return 0;
 
    /* can only accept on stream sockets */
    if (S->socket_type != SOCK_STREAM) return 0;
 
    /* accept connections */
    length = sizeof addr;
    new_socket = accept(S->socket_no,(struct sockaddr *) &addr, &length);

    if (new_socket > 0) 
    {
        struct protoent * prot;

        printf("%d accepting socket %d\n", getpid(), new_socket);
        /* get a null socket */
        o = Ecreate_socket(0,0,0);     
        if (o->socket_no) close(o->socket_no);
        o->socket_no = new_socket;
        o->int_address = addr.sin_addr.s_addr;        /* set remote address */
        o->port = addr.sin_port;
        /*      printf("finally got an accept\n"); */

        prot = getprotobyname("TCP");
#if 0
        Setsockopt(o, prot->p_proto, TCP_NODELAY, 0);
        o->Setsockopt(SOL_SOCKET, SO_REUSEADDR, 0))
        { perror("setsockopt"); //    exit(1); }
#endif
        return o;
    }
    return 0;
}
 
/*
 * Bind to an address on the local machine
 * Internet bind
 */
int Ebind(Socket S, int port_number)
{
    struct sockaddr_in sin;        

#if 0
    if (!Setsockopt(SOL_SOCKET, SO_REUSEADDR, 0))
    {
        perror("setsockopt");
        exit(1);
    }
#endif

    memset(&sin, 0, sizeof sin);

    sin.sin_port = htons(port_number);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
 
    if (bind(S->socket_no, (struct sockaddr *) &sin, sizeof sin) == -1)
    {
        Ethrowv("bind() failed in Ebind()\n");
/*         fprintf(stderr, "ERR: %d\n", errno); */
        return 0;
    }
    return 1;
}
 
/*
** Bind to a local socket
*/
int Bind_str(Socket S, const char * name)
{
    struct sockaddr_un sin;        
    int len;
 
    memset((char *) &sin, '\0', sizeof sin);
    sin.sun_family = AF_UNIX;
    strcpy(sin.sun_path, name);
    len = sizeof(sin) - sizeof(sin.sun_path) + strlen(sin.sun_path);
 
    if (bind(S->socket_no, (struct sockaddr *) &sin, len) == -1)
    {
        Ethrowv("bind() failed in Ebindstr()\n");

        return 0;
    }
    return 1;
}

int Eclose(Socket S)
{
        /* if it's a unix domain socket we should unlink too.. */
        // printf("%d closing socket %d\n", getpid(), S->socket_no);
        return close(S->socket_no);
}
 
/*
 * Connect to a remote address
 */
int Econnect_str(Socket S, const char * address)
{
    int len;
        /* set up identification for server's inwards socket */
    if (S->family == AF_UNIX) 
    {
        struct sockaddr_un server;

        server.sun_family = AF_UNIX;
        strcpy(server.sun_path, address);
        len = sizeof(server) - sizeof(server.sun_path) + strlen(server.sun_path);
        /* block waiting for a connection */
        if (connect(S->socket_no, (struct sockaddr *)&server, len) == -1) 
        {
/*            perror("client connect()"); */
               return 0;
        }
        return 1;
    }
    else     
    {
		/* inet stuff */
        unsigned int port;
        char * loc;

        loc = index(address, ':');
        if (loc == NULL) return 0;
        port = atoi(&(loc[1]));    
        return Econnect_addr_port(S, address, port);
    }
}

/*
** Connect to a remote inet address
*/
int Econnect_str_port(Socket S, const char * address, unsigned int port)
{
    if (S->family == AF_INET)
    {
        struct sockaddr_in sin;
 
        sin.sin_port = htons(port);
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = inet_addr(address);
        if (connect(S->socket_no, (struct sockaddr *) &sin, sizeof sin) == -1) 
        {
/*            perror("connect"); */
            return 0;
        }
        return 1;
    }
    return 0;
}

/*
** Connect to a remote inet address
*/
int Econnect_addr_port(Socket S, unsigned long addr, unsigned int port)
{
    if (S->family == AF_INET)
    {
        struct sockaddr_in sin;
 
		memset(&sin, '\0', sizeof sin);	
        S->port = port;
        sin.sin_port = htons(port);
        S->family = sin.sin_family = AF_INET;
        S->int_address = sin.sin_addr.s_addr = ntohl(addr);
        if (connect(S->socket_no, (struct sockaddr *) &sin, sizeof sin) == -1) 
        {
            perror("Econnect");
            return 0;
        }
        return 1;
    }
    return 0;
}

/* 
** Connect assume we have the address in our buffer
** from a recvfrom
*/
int Econnect(Socket S, Val Pid)
{
    return Econnect_addr_port(S, Eindex_tuple(Pid, Pid_ipaddr).u.number, 
                    Eindex_tuple(Pid, Pid_port).u.number);
}

/*
** Set up values for Sendto
*/
void Econnectto(Socket S, const char * address, unsigned int prt)
{
    assert(address != NULL);

    S->int_address = inet_addr(address);
    S->port = prt;
    return;
}
 
int Fileno(Socket S)
{
    return 0;
}
 
char * Getpeername(Socket S)
{
    return 0;
}
 
char * Getsockopt(Socket S, int level, int optname, int buflen)
{
#if 0
    S->level = 0;
    S->optname = 0;
    S->buflen = 0;
#endif
    return 0;
}

int Efcntl(Socket S, int command, void * arg)
{
    return fcntl(S->socket_no, command, arg);
}

int Eioctl(Socket S, int command, void * arg)
{
    return ioctl(S->socket_no, command, arg);
}
 
int Elisten(Socket S,int backlog)
{
    return listen(S->socket_no, backlog);
}
 
EBinary Erecv(Socket S, int bufsize)
{
    int x;
    EBinary ret;
    ret = binaryalloc((bufsize+1)*sizeof(unsigned char));
    //fprintf(stderr, "Erecv on %d\n", S->socket_no);
    x = recv(S->socket_no, ret->b, bufsize, 0);
    if (x == -1) 
    {
        binaryfree(ret);
        return NULL;
    }
    ret->size = x;
    return ret;
}

EBinary Erecvfrom(Socket S, int bufsize, Socket sock)
{
        int x;
        EBinary ret;      
        struct sockaddr_in from;
        int fromlen = sizeof(struct sockaddr_in);

        memset(&from, 0, fromlen);

        ret = binaryalloc(bufsize+1);

        x = recvfrom(S->socket_no, ret->b, bufsize, 0, (struct sockaddr *)&from, &fromlen);
        if (x == -1) 
        {
            binaryfree(ret);
            return NULL;
        }
        ret->size = x;

        sock->family = from.sin_family;
        sock->int_address = from.sin_addr.s_addr;
        sock->port = from.sin_port;

        return ret;
}
 
/*
** return # of characters sent (or -1 on failure)
** /int send(char * what, int flags) - for out of band data?
*/
int Esendbinary(Socket S, EBinary bin)
{
    int x;
        
    assert(bin != NULL);
    if (bin->size == 0) return 0;

    x = send(S->socket_no, bin->b, bin->size, 0);
    // fprintf(stderr, "Esendbinary (%d)(%s)=%d\n", bin->size, &(bin->b[7]), x);
    return x;
}
 
/* send to last place we knew.. */
int Esendto(Socket S, const char * what)
{
    int x;
    struct sockaddr_in sin;
    int len = sizeof(struct sockaddr_in);

    if (what == NULL) return 0;

    sin.sin_family = S->family;
    sin.sin_addr.s_addr = S->int_address;
    sin.sin_port = htons(S->port);

    x = sendto(S->socket_no, what, strlen(what), 0, (struct sockaddr *)&sin, len);
    return x;
}
 
int Setblocking(Socket S, int flags)
{
    if (flags == 0) 
    {
#if 0
        if (ioctl(socket_no, FIONBIO, &tmp) == -1) {
/*                perror("ioctl socket FIONBIO"); */
                return 0;
        }
#else                           /* sun */
        if (fcntl(S->socket_no, F_SETFL, O_NDELAY) == -1) {
/*                perror("fcntl socket FIONBIO"); */
                return 0;
        }
#endif                          /* sun */
        return 1;
    }
    else  
    {
                return 0;
    }
}
 
int Setsockopt(Socket S, int level, int optname, char * value)
{
    int tmp = 1;
    value = 0;
    if (setsockopt(S->socket_no, level, optname, (char *)&tmp, sizeof(tmp))==-1) 
    {
        Ethrowv("setsockopt() failed in Esetsockopt()\n");
        return 0;
    }
    return 1;
}
 
int Eshutdown(Socket S, int how)
{
        return shutdown(S->socket_no, how);
}
 
 
char * Eip_number(Socket S)
{
    struct in_addr xx;
    xx.s_addr = S->int_address;
    if (S->socket_no == 0) 
        return string_copy("");
    return string_copy(inet_ntoa(xx));
}
 
 
#if 0
/*
** Note: if the address char * is "a.b.c.d" the address number is a * 256^3 +
** b * 256^2 + c * 256 + d
** 
*/
 
char * ip_number()
{
    int         a, b, c, d;
    char *      addr;   
 
    if (!socket_no) return "";
 
    d = int_address & 0xff;
    c = (int_address & 0xff00) >> 8;
    b = (int_address & 0xff0000) >> 16;
    a = (int_address & 0xff000000) >> 24;
    addr = a+"."+b+"."+c+"."+d;
    return addr;
}
#endif                          /* INET_NTOA_OK */
 
#define HOST_BUFF 1024
char * Ehostname()
{
    static char     name[HOST_BUFF];
 
    gethostname(name, sizeof name);
    name[sizeof name - 1] = '\0';       /* Just to make sure */
    return string_copy(name);
}
