
/*
** $Id: list.c,v 1.49 2002/08/03 14:31:42 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/list.c,v $
** $Revision: 1.49 $
** $Date: 2002/08/03 14:31:42 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright (C) 1998 
** geoff@cs.rmit.edu.au
*/

/* Erlang List Support */
/* NOTES:  needs to be made more efficient */
/* 	should have shared/unshared and allocated blocks */

#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include "erlib.h"
#include "typecheck.h"
#include "blist.h"

#define LIST_BLOCK 32

Val listalloc(int sz)
{
    return listmalloc(sz);
#if 0
	Val v;
    int allocsz;

    assert(sz >= 0);

    /* twice the size up to next 32 size boundary up */
    allocsz = ((sz<<1)+(LIST_BLOCK-1)) & (~(LIST_BLOCK));

	v.tag = LT_list | LT_on_heap;
    if (sz) v.u.list = (List)malloc(sizeof(_List) + sizeof(Val)*(allocsz-1));
    else v.u.list = NULL;
	v.u.list->size = allocsz;
	v.u.list->length = sz;
	v.u.list->start =  &(v.u.list->e[allocsz/4]);
	return v;
#endif
}


/* actually creates a list */
Val EValmstring(const char * atom)
{
	Val v;
    int n, i, count;

    n = strlen(atom);
	v = listmalloc(n);
    count = v.u.list->e;
	for (i = 0; i < n; i++)
	{
		v.u.list->start[count++] = EValint(atom[i]);
	}

	return v;
}

Val EValstring(Val V, const char * atom)
{
    V = EValmstring(atom);
    return V;
}

/* Take a chunk of bytes and turn them into an Erlang list */
Val Ebytes_to_list(const char *bytes, int size)
{
  Val v;
  int i;

  v = listalloc(size);
  for (i = 0; i < size; i++) 
  {
    v.u.list->e->mem[v.u.list->start+i] = EValint(bytes[i]); 
  }

  return v;
}

Val listmalloc(int sz)
{
	Val v;
    int allocsz;

    assert(sz >= 0);

    /* special case - don't allocate memory */
    if (sz == 0)
    {
        allocsz = 0;
    }
    else
    {
        /* twice the size up to next 32 size boundary up */
        allocsz = ((sz<<1)+(LIST_BLOCK-1)) & (~(LIST_BLOCK));
    }

	v.tag = LT_list | LT_on_heap;
	v.u.list = (List) malloc(sizeof(_List));
    /* memset(v->u.list, 0, sizeof(_List)); */

	v.u.list->size = allocsz;
	v.u.list->length = sz;
    v.u.list->ref = 1;
    if (allocsz) 
    {
        v.u.list->start = (Val *) malloc(allocsz * sizeof(Val));
        memset(v.u.list->start, 0, allocsz * sizeof(Val));
	    v.u.list->e = allocsz/2;
    }
    else
    {
        v.u.list->start = NULL;
        v.u.list->e = 0;
    }

    /* I am the memory holder .. */
    v.u.list->memholder = v.u.list;

	return v;
}

void listrealloc(List l, int sz)
{
    int allocsz;

    allocsz = ((sz<<1)+(LIST_BLOCK-1)) & (~(LIST_BLOCK));

	l->size = allocsz;
	l->length = sz;

    l->start = (Val *) malloc(allocsz * sizeof(Val));
    memset(l->start, 0, allocsz * sizeof(Val));
#if 0
    memcpy(l.u.list->start,
	v.u.list->e = allocsz/2;
#endif
}


Val Econcatvl(Val v, Val item, Val l)
{
    if (l.u.list->e != 0 && 
        (l.u.list->memholder->start[l.u.list->e-1].tag == LT_none))
    {
        v = listmalloc(0);
        memcpy(v.u.list, l.u.list, sizeof(struct _list)); 
        v.u.list->length++;
        v.u.list->e = l.u.list->e - 1;
        if (v.u.list->start != NULL)
        {
            v.u.list->memholder = l.u.list->memholder;
            v.u.list->start = NULL;
        }
	    v.u.list->memholder->start[v.u.list->e] = item;
        return v;
    }
    else
    {
        v = listmalloc(l.u.list->length + 1);

	    v.u.list->memholder->start[v.u.list->e] = item;
        if (l.u.list->length)
            memcpy(&(v.u.list->memholder->start[v.u.list->e+1]),
            &(l.u.list->memholder->start[l.u.list->e]), 
            sizeof(Val) * l.u.list->length);
#if 0
        if (l.u.list->start != NULL)
        {
            free(l.u.list->start);
            l.u.list->start = NULL;
        }
        l.u.list->e = v.u.list->e + 1;
        l.u.list->memholder = v.u.list;
#endif
	    return v;
    }
}

Val Econcat(Val h, Val t)
{
    Val v;
	int sz, i, count = 0;

//  should already be type checked before this is called (effectively)
//	TC(h, LT_list, 1, "Econcat arg 1");
//	TC(t, LT_list, 1, "Econcat arg 2");

    if (t.u.list->length == 0)
    {
        v = listmalloc(0);
        v.u.list->length = h.u.list->length;
        v.u.list->e = h.u.list->e;
        v.u.list->start = NULL;
        v.u.list->memholder = h.u.list->memholder;
	    return v;
    }

	sz = h.u.list->length + t.u.list->length;
        // + (int) (h.u.list->start - h.u.list->e);
	
    if (h.u.list->size < t.u.list->e &&
        (t.u.list->memholder->start[t.u.list->e-1].tag == LT_none))
    {
        memcpy(&(t.u.list->memholder->start[t.u.list->e - h.u.list->size]),
                &(h.u.list->memholder->start[h.u.list->e]), 
                sizeof(Val) * h.u.list->size);
        // if (h.u.list->start != NULL) free(h.u.list->start);
        v = listmalloc(0);
        v.u.list->length = sz;
        v.u.list->e = t.u.list->e - h.u.list->size;
        v.u.list->start = NULL;
        v.u.list->memholder = t.u.list->memholder;
#if 0
        h.u.list->start = NULL;
        h.u.list->memholder = t.u.list;
#endif
    }
    else
    {
        v = listmalloc(sz);
        memcpy(&(v.u.list->start[v.u.list->e]), 
            &(h.u.list->memholder->start[h.u.list->e]), 
                sizeof(Val) * h.u.list->length);
        memcpy(&(v.u.list->start[v.u.list->e+h.u.list->length]),
                &(t.u.list->memholder->start[t.u.list->e]), 
                sizeof(Val) * t.u.list->length);
#if 0
        if (t.u.list->start != NULL) free(t.u.list->start);
        if (h.u.list->start != NULL) free(h.u.list->start);
        h.u.list->e = v.u.list->e;
        h.u.list->start = NULL;
        h.u.list->memholder = v.u.list;
        t.u.list->e = v.u.list->e + h.u.list->size;
        t.u.list->start = NULL;
        t.u.list->memholder = v.u.list;
#endif
    }

	return v;
}

Val Eappendv(Val h, Val t)
{
    assert(1==0);
}

Val	Emake_list(int n, ...)
{
	Val v;
	int i, count;
	va_list args;
	va_start(args, n);
	
	v = listmalloc(n);
    count = v.u.list->e;
	for (i = 0; i < n; i++)
	{
		v.u.list->start[count++] = va_arg(args, Val);
	}

	va_end(args);
	return v;
}

/* returns the tail of a list (a list itself) */
Val	Etail_list(Val ret, Val l, int arg, const char * from)
{
	// TC(l, LT_list, arg, from);

    // assert(ret.u.list->size >= 0);
    // assert(l.u.list->size >= 0);

    ret = listmalloc(0);
    memcpy(ret.u.list, l.u.list, sizeof(struct _list));
    ret.u.list->length--;
    ret.u.list->e = l.u.list->e + 1;
    ret.u.list->memholder = l.u.list->memholder;
    ret.u.list->start = NULL;
    return ret;
}

Val	Entail_list(Val ret, Val l, int tailx, const char * from)
{
	// TC(l, LT_list, arg, from);

    // assert(ret.u.list->size >= 0);
    // assert(l.u.list->size >= 0);

    if (l.u.list->length > tailx)
    {
        ret.tag = LT_list;
        ret.u.list = (List) malloc(sizeof(struct _list));
        memcpy(ret.u.list, l.u.list, sizeof(struct _list));
        ret.u.list->length = ret.u.list->length - tailx;
        ret.u.list->e = l.u.list->e + tailx;
        ret.u.list->memholder = l.u.list;
        ret.u.list->start = NULL;
        return ret;
    }

    // FIX: check this
    ret.u.list = (List) malloc(sizeof(struct _list));
    ret.tag = LT_list;
    ret.u.list->memholder = ret.u.list;
    ret.u.list->start = NULL;
    ret.u.list->e = 0;

    return ret;
}

int	Eis_empty_list(Val l)
{
	/* TC(l, LT_list, 1, "is_empty_list"); */
    if ((l.tag & LT_list) == 0) return 0;

	if ((l.u.list == NULL) || (l.u.list->length == 0)) return 1;
	return 0;
}

int	Eis_full_list(Val l)
{
	/* TC(l, LT_list, 1, "is_full_list"); */
    if ((l.tag & LT_list) == 0) return 0;

	if (l.u.list && (l.u.list->length > 0)) return 1;
	return 0;
}


/* Transforms an "Erlang string" into a "C string" */

char *Elist_to_string(Val l, int arg, const char * from)
{
  int i,size;
  char *cp=NULL;
  Val v;

  TC(l, LT_list, arg, from);
  size = l.u.list->length;
  cp = (char *) malloc(size + 1);
  if (cp == NULL) Ethrowv("Out of memory (Elist_to_string)\n");

  for (i = 0; i < size; i++) 
  {
    v = l.u.list->memholder->start[l.u.list->e+i];
    cp[i] = v.u.number;
  }
  cp[i] = '\0';

  return cp;
}

Val parammalloc(int sz)
{
    Val v;
    assert(sz >= 0);

    v.tag = LT_list | LT_on_heap;
//  v.tag = LT_list;
    v.u.param = (ParamList)malloc(sizeof(_ParamList));
    v.u.param->size =sz;
    if (sz) v.u.param->e = (Val *)malloc(sizeof(Val)*sz);
    else v.u.param->e = NULL;
    return v;
}

//
// FIX: may recurse into lists/tuples infinitly?
//
int listcmp(Val l, Val r)
{
    int i;

    if (l.u.list == r.u.list) return 1;
    if (l.u.list->length != r.u.list->length) return 0;
    if ((int)(&(l.u.list->memholder->start[l.u.list->e])) ==
        (int)(&(r.u.list->memholder->start[r.u.list->e]))) return 1;

    for (i = 0; i < l.u.list->length; i++)
    {
        if (!(Eequal(l.u.list->memholder->start[l.u.list->e+i],
                    r.u.list->memholder->start[r.u.list->e+i]))) return 0;
    }

    return 1;
}

