
/*
** $Id: tuple.c,v 1.25 2002/08/03 14:31:42 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/tuple.c,v $
** $Revision: 1.25 $
** $Date: 2002/08/03 14:31:42 $
** $State: Exp $
**
** Author: Geoff Wong
*/

/* Tuple Support */
/* NOTES:  needs to be made more efficient */
/* should have shared/unshared and allocated blocks */

#include <stdarg.h>
#include <assert.h>
#include "erlib.h"
#include "typecheck.h"

inline void tuplefree(Val v)
{
    v.u.tuple->ref--;

    if (v.u.tuple->ref == 0)
    {
        int i;
        for (i = 0; i < v.u.tuple->size; i++)
        {
            Efre(v.u.tuple->e[i]);
        }
        free(v.u.tuple);
    }
}

Val tuplealloc(int sz)
{
	Val v;
	v.tag = LT_tuple;
	v.u.tuple = (ETuple)malloc(sizeof(_Tuple) + sizeof(Val)*(sz-1));
	v.u.tuple->size = sz;
	v.u.tuple->ref = 1;
	return v;
}

inline Val tuplemalloc(unsigned int sz)
{
	Val v;
	v.tag = LT_tuple | LT_on_heap;
	v.u.tuple = (ETuple)malloc(sizeof(_Tuple) + sizeof(Val)*(sz-1));
	v.u.tuple->size = sz;
	v.u.tuple->ref = 1;
	return v;
}

Val tuplecopy(Val v)
{
	Val ret;
    int i;

    ret = tuplemalloc(v.u.tuple->size);
    for (i = 0; i < v.u.tuple->size; i++)
        ret.u.tuple->e[i] = EValcopy(v.u.tuple->e[i]);

	return ret;
}

#ifdef _ALLOCS_H
Val	Emake_tuple(Val n, ...)
{
	int i, count = 0;
	va_list args;
	va_start(args, n);
	
	for (i = 0; i < n.u.tuple->size; i++)
	{
		n.u.tuple->e[count++] = EValcopy(va_arg(args, Val));
	}

	va_end(args);
	return n;
}

#else
Val	Emake_tuple(int n, ...)
{
	Val v;
	int i, count = 0;
	va_list args;
	va_start(args, n);
	
	v = tuplemalloc(n);
	for (i = 0; i < n; i++)
	{
		// v.u.tuple->e[count++] = EValcopy(va_arg(args, Val));
		v.u.tuple->e[count++] = va_arg(args, Val);
	}

	va_end(args);
	return v;
}

#endif

inline int	Eis_empty_tuple(Val l)
{
	/* TC(l, LT_tuple, 1, "Eis_empty_tuple"); */
    if (l.tag != LT_tuple) return 0;
	if (l.u.tuple->size == 0) return 1;
	return 0;
}

inline int	Eis_tuple(int x, Val l)
{
	/* TC(l, LT_tuple, 1, "Eis_tuple"); */
    if (!(l.tag & LT_tuple)) return 0;
	if (l.u.tuple->size == x) return 1;
	return 0;
}

int tuplecmp(Val l, Val r)
{
    int i;

    if (l.u.tuple == r.u.tuple) return 1;
    if (l.u.tuple->size != r.u.tuple->size) return 0;

    for (i = 0; i < l.u.tuple->size; i++)
    {
        if (!(Eequal(l.u.tuple->e[i],
                    r.u.tuple->e[i]))) return 0;
    }

    return 1;
}

