/*
** Server: handle multiple incoming messages and buffer them
**      notify clients of their arrival and deliver when
**      the client is ready to receive them
**
** Todo: how much of this could be built in Erlang?
**      Also the connect structure (C) should be a list
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#define IP_PROT 0

#define MAX_CONNECT 64

main(int argc, char * argv[])
{
    struct sockaddr_in sin, addr;
    int S, max = 0;
    int C[MAX_CONNECT], connect = 0, T;
    fd_set readfds, totalfds;
    int length = sizeof addr;
    int i;

    if (argc < 2)
    {
        fprintf(stderr, "server <port>\n");
        exit(1);
    }
    
    /* initialise structures */
    for (i = 0; i < MAX_CONNECT; i++) C[i] = 0;
    FD_ZERO(&readfds);
    FD_ZERO(&totalfds);

    /* create, bind and listen to a port */
    S = socket(AF_INET, SOCK_STREAM, IP_PROT);
    memset((char *) &sin, '\0', sizeof sin);
    sin.sin_family = AF_INET;
    sin.sin_port = htons(atoi(argv[1]));
    sin.sin_addr.s_addr = INADDR_ANY;
    if (bind(S, &sin, sizeof sin) == -1)
    {
        fprintf(stderr, "Server unable to bind to port.\n");
        exit(1);
    }
    listen(S,5);

    printf("Bound to port\n");

    /* non blocking on accepts */
    if (fcntl(S, F_SETFL, O_NDELAY) == -1) 
    {
        fprintf(stderr, "Server unable to set port into nonblocking mode.\n");
        exit(1);
    }

    /* main listen loop */
    while (1)
    {
        struct timeval timeout;
        int res, x, i;
        char buf[65536];
        
        T = accept(S,&addr,&length);
        if (T == -1 && errno != EAGAIN)
        {
            fprintf(stderr, "Accept error %d.\n", errno);
            exit(1);
        }
        if (T > 0)
        {
            fcntl(T, F_SETFL, O_NDELAY);
            printf("Accepted connection (%d).\n", T);
            C[connect++] = T;
            FD_SET(T, &totalfds);
            if (T >= max) max = T+1;
        }

        readfds = totalfds;

        timeout.tv_sec = 0;
        timeout.tv_usec = 10000;
        res = select(max, &readfds, 0, 0, &timeout);
        if (res == -1)
        {
            fprintf(stderr, "Select error.\n");
            exit(1);
        } 
        if (res > 0)
        {
            printf("select result = %d\n", res);
            for (i = 0; i < connect; i++)
            {
                if (FD_ISSET(C[i], &readfds))
                {
                    printf("FD_ISSET true\n");
                    res = recvfrom(C[i], buf, 65536, 0, 0, 0);
                    if (res > 0)
                    {
                        buf[res] = 0;
                        printf("received (%d on %d) = %s\n", i, C[i], buf);
                    }
                    else if (res == 0)
                    {
                        FD_CLR(C[i], &totalfds);
                        C[i] = 0;
                    }
                }
            }

        }
    }
}
