

//
// $Id: basic.cc,v 1.36 2002/08/04 14:45:28 geoff Exp $
//
// $Source: /home/cvs/gerl/src/basic.cc,v $
// $Revision: 1.36 $
// $Date: 2002/08/04 14:45:28 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "str.h"
#include "basic.h"

extern int debug;

// #define DEBUG 1

// Class Atom
Atom::Atom(const string& tmp, const string * at)
{
    storage = tmp;
	value = *at;
    _type = LT_atom;
}

list<string> Atom::emit_code()
{
#ifdef _ALLOCS_H
    return string2list("EValatom(" + storage + ",\""+value+"\")");
#else
    return string2list("EValatom(\""+value+"\")");
#endif
}

string Atom::print()
{
	return value;
	// return value;
	// shouldn't be here !?
	//return string2list("EValatom(\""+value+"\")");
}

string Atom::emit_pattern(string p)
{
    string ret;

    if (value == "_")
    {
        //ret = "(" + p + ")";
        ret = "(1)";
    }
    else
    {
        ret = "Eequalvs(" + p + ", \"" + value + "\")";
    }
    return ret;
}                            

// Class Number
Number::Number(int num)
{
	value = num;
    _type = LT_number;
}

string Number::print()
{
	return itoa(value);
}

list<string> Number::emit_code()
{
	return string2list("EValint("+itoa(value)+")");
}

string Number::emit_pattern(string p)
{
    string ret;

    ret = ret + "(Eequalvi(" + p + ", " + itoa(value) + "))";
    return ret;                   
}

// Class Real
Real::Real(string& num)
{
	value = num;
    _type = LT_real;
}

string Real::print()
{
	return value;
}

list<string> Real::emit_code()
{
	return string2list("EValreal("+value+")");
}

string Real::emit_pattern(string p)
{
    string ret;

    ret = ret + "(Eequalvi(" + p + ", " + value + "))";
    return ret;                   
}

// Erlang string Class
Estring::Estring(const string& tmp, const string * str)
{
    storage = tmp;
	value = *str;
    _type = LT_string;
}

string Estring::print()
{
	return value;
}

list<string> Estring::emit_code()
{
#ifdef _ALLOCS_H
	return string2list("EValstring(" + storage + "," + value + ")");
#else
	return string2list("EValstring(" + value + ")");
#endif
}

string Estring::emit_pattern(string p)
{
    string ret;

    ret = ret + "(Estrcmp(" + p + "," + value + ") == 0)";
    return ret;
}                

Variable::Variable()
{
}

// Variable Class
Variable::Variable(const string * v)
{
	value = *v;
}

string Variable::print()
{
    return value;
}

list<string> Variable::emit_code()
{
    if (debug) fprintf(stderr, "Variable::emit_code(): %s(%d)(%d)\n", 
            value.c_str(), type(), (int)(this));
	return string2list(value);
}

string Variable::emit_pattern(string p)
{
    string ret = "(1)";
    SymbolTable * st;
    Symbol * sym;

    st = CurrentTable;

#ifdef DEBUG 
    fprintf(stderr, "emitparameter: %s = %s\n", value.c_str(), p.c_str());
#endif

    if ((sym = st->find(value, S_Variable)))
    {
        if ((sym->symbol() == value) && (sym->parameter().empty() == false))
        {
            ret = "(Eequal(" + sym->parameter() + ", " + p + "))";
        }
        else
        {
            if (debug) fprintf(stderr,"Variable::emit_pattern(): %s = %s (%d)(%d)\n",
            sym->symbol().c_str(), p.c_str(), sym->type(), (int)(this));
            sym->setparameter(p);
            type(sym->type());
            // LIST HACK:
            if (extra() == "destructive") sym->destructive(true);
        }
    }
    else
    {
        fprintf(stderr,"Can't find Variable: %s\n", value.c_str());
    }
    return ret;

}

void Variable::add_symbol(SymbolTable * st, int df)
{
    Symbol * s;
    string ts = "";

    if (debug) 
        fprintf(stderr, "Var.add_symbol(): %s (%d) \n", value.c_str(), df);
    if (st != NULL)
    {
        s = new Symbol(value, S_Variable, ts);
        s->associated(this);
        if (st->add(s) == false)
        {
            delete s;
            s = st->find(value, S_Variable);
            s->type(s->type() | df);
        }
        else s->type(df);
    }

    assert(s != NULL);

    type(s->type());
}

int Variable::type()
{
    if (debug) fprintf(stderr,"Var.type():%s(%d)\n", value.c_str(), _type);
    return _type;
}

void Variable::type(int t)
{
    if (debug) fprintf(stderr,"Var::type(%d) for %s\n", t, value.c_str());
    // Node::type(t);
    _type = t;
}

// Boolean Class
Boolean::Boolean(int val)
{
	if (val == 0) value = false;
	else value = true;
    _type = LT_bool;
}

list<string> Boolean::emit_code()
{
	if (value == false) 
		return string2list("EValbool(0)");
	return string2list("EValbool(1)");
}

string Boolean::print()
{
	if (value == false) 
		return string("false");
	return string("true");
}

string Boolean::emit_pattern(string p)
{
    return "Eequalbi(" + p + "," + itoa(value) + ")";
}

