
//
// $Id: binary.cc,v 1.12 2002/11/06 12:47:30 geoff Exp $
//
// $Source: /home/cvs/gerl/src/binary.cc,v $
// $Revision: 1.12 $
// $Date: 2002/11/06 12:47:30 $
// $State: Exp $
//
// Copyright (C) Geoff Wong (1996, 1997, 1998)
// geoff@cs.rmit.edu.au
//
// This file is distributed under the terms of the
// GNU Public License, details of which can be found
// in the file "Copying" which accompanies this code.        
//
// Note: code emission herein assumes constant folding has been done
//

#include "binary.h"

// Class Binary
Binary::Binary(string * str, Node * l, Node * r)
{
	LHS = l; RHS = r;
	operate = *str;
}

Binary::Binary(const string& op, Node * l, Node * r)
{
	LHS = l; RHS = r;
	operate = op;
}

string Binary::type_code(Node * node)
{
    // int cnt = node->type() & LT_constant;

    switch (node->type() & LT_single)
    {
    case LT_none:
        return "v";
    case LT_real:
        return "r";
    case LT_number:
        return "i";
    case LT_atom:
        return "a";
    case LT_tuple:
        return "t";
    case LT_list:
        return "l";
    default:
        return "v";
    }
}

int Binary::tag_code(Node * node, string& res, string& app)
{
    string cut;

    res = node->expr_code();
    switch (node->type() & LT_single)
    {
    case LT_none:
        app = res;
        return LT_none;
    case LT_real:
        if (node->type() & LT_constant)
        {
            app = node->print();
        }
        else
        {
            cut = "EValreal";
            if (res.substr(0,cut.length()) == cut)
                app = res.substr(cut.length());
            else
                app = res + ".u.real";
        }
        return LT_real;
    case LT_number:
        if (node->type() & LT_constant)
        {
            app = node->print();
        }
        else
        {
            res = node->expr_code();
            cut = "EValint";
            if (res.substr(0,cut.length()) == cut)
                app = res.substr(cut.length());
            else
                app = res + ".u.number";
        }
        return LT_number;
    case LT_atom:
        app = res + ".u.atom";
        return LT_atom;
    case LT_tuple:
        app = res + ".u.tuple";
        return LT_tuple;
    case LT_list:
        app = res + ".u.list";
        return LT_list;
    default:
        app = res; // + ".u.any";
        return LT_none;
    }
  
}

list<string> Binary::emit_code()
{
    add_pre_code(LHS->pre_code());
    add_pre_code(RHS->pre_code());

	return string2list(operate 
        + "(" + LHS->expr_code() + ", " + RHS->expr_code() + ")");
}

string Binary::print()
{
	return LHS->print() + operate + RHS->print();
}

PlusExpr::PlusExpr(Node * l, Node * r) :
        Binary("Eplus", l, r)
{
}

list<string> PlusExpr::emit_code()
{
    string alhs, arhs, lhs, rhs, res, ret;
    int flag = 0, et;

    add_pre_code(LHS->pre_code());
    add_pre_code(RHS->pre_code());

    res = "EValint";
    // type(LT_number);

    et = tag_code(LHS, lhs, alhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
            type(LT_real);
        }
        flag++;
    }
    et = tag_code(RHS, rhs, arhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
            type(LT_real);
        }
        flag++;
    }
    switch (flag)
    {
        case 1:
            return string2list(operate + type_code(LHS) + type_code(RHS) 
             + "(" + alhs + ", " + arhs + ")");
        case 2:
            return string2list(res + "(" + alhs + " + " + arhs + ")");
        default:
            return string2list(operate + "(" + lhs + ", " + rhs + ")");
    }
}

MinusExpr::MinusExpr(Node * l, Node * r) :
    Binary("Eminus", l, r)
{
}

list<string> MinusExpr::emit_code()
{
    string alhs, arhs, lhs, rhs, res, ret;
    int flag = 0, et;
    res = "EValint";

    add_pre_code(LHS->pre_code());
    add_pre_code(RHS->pre_code());

    // type(LT_number);
    et = tag_code(LHS, lhs, alhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
            type(LT_real);
        }
        flag++;
    }
    et = tag_code(RHS, rhs, arhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
            type(LT_real);
        }
        flag++;
    }
    switch(flag)
    {
        case 1:
            return string2list(operate + type_code(LHS) + type_code(RHS) 
             + "(" + alhs + ", " + arhs + ")");
        case 2:
            return string2list(res + "(" + alhs + " - " + arhs + ")");
        default:
            return string2list(operate + "(" + lhs + ", " + rhs + ")");
    }
}

MultiplyExpr::MultiplyExpr(Node * l, Node * r) :
    Binary("Emultiply", l, r)
{
}

list<string> MultiplyExpr::emit_code()
{
    string alhs, arhs, lhs, rhs, res, ret;
    int flag = 0, et, xt;

    add_pre_code(LHS->pre_code());
    add_pre_code(RHS->pre_code());

    res = "EValreal";
    et = tag_code(LHS, lhs, alhs);
    if (et) flag++;
    xt = tag_code(RHS, rhs, arhs);
    if (xt) flag++;
    if ( (et == LT_number) && (xt == LT_number))
    {
        res = "EValint";
        type(LT_number);
    }
    switch(flag)
    {
        case 1:
            return string2list(operate + type_code(LHS) + type_code(RHS) 
             + "(" + alhs + ", " + arhs + ")");
        case 2:
            return string2list(res + "(" + alhs + " * " + arhs + ")");
        default:
            return string2list(operate + "(" + lhs + ", " + rhs + ")");
    }
}

DivideExpr::DivideExpr(Node * l, Node * r) :
    Binary("Edivide", l, r)
{
}

list<string> DivideExpr::emit_code()
{
    string alhs, arhs, lhs, rhs, res, ret;
    int flag = 0, et;
    res = "EValint";

    add_pre_code(LHS->pre_code());
    add_pre_code(RHS->pre_code());

    // type(LT_real);
    et = tag_code(LHS, lhs, alhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
        }
        flag++;
    }
    et = tag_code(RHS, rhs, arhs);
    if (et)
    {
        if (et == LT_real) 
        {
            res = "EValreal";
        }
        flag++;
    }
    switch(flag)
    {
        case 1:
            return string2list(operate + type_code(LHS) + type_code(RHS) 
             + "(" + alhs + ", " + arhs + ")");
        case 2:
            return string2list(res + "(" + alhs + " / " + arhs + ")");
        default:
            return string2list(operate + "(" + lhs + ", " + rhs + ")");
    }
}

BorExpr::BorExpr(Node * l, Node * r) :
    Binary("Ebor", l, r)
{
}

BxorExpr::BxorExpr(Node * l, Node * r) :
    Binary("Ebxor", l, r)
{
}

BslExpr::BslExpr(Node * l, Node * r) :
    Binary("Ebsl", l, r)
{
}

BsrExpr::BsrExpr(Node * l, Node * r) :
    Binary("Ebsr", l, r)
{
}

BandExpr::BandExpr(Node * l, Node * r) :
    Binary("Eband", l, r)
{
}

DivExpr::DivExpr(Node * l, Node * r) :
    Binary("Ediv", l, r)
{
}

RemExpr::RemExpr(Node * l, Node * r) :
    Binary("Erem", l, r)
{
}

ConcatExpr::ConcatExpr(Node * l, Node * r) :
    Binary("Econcat", l, r)
{
}


ListSubtract::ListSubtract(Node * l, Node * r) :
    Binary("Elistsub", l, r)
{
}


                                  
