

//
// $Id: case.cc,v 1.29 2002/07/24 14:01:38 geoff Exp $
//
// $Source: /home/cvs/gerl/src/case.cc,v $
// $Revision: 1.29 $
// $Date: 2002/07/24 14:01:38 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "case.h"
#include "str.h"
#include "program.h"
#include "error.h"

// #define DEBUG


CaseList::CaseList(Node * n)
{
    nodes.push_front(n);
    stmt_type(ST_stmt);
}

string CaseList::emit_code(string p)
{
    list<Node *>::iterator i;
    string ret;

    if (nodes.size() == 0)
    {
        ret = "1";
        return ret;
    }

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        ret = ret + (*i)->emit_pattern(p);
        if (i != --(nodes.end())) ret = ret + "else ";
    }
    ret = ret + "\telse\n\t{\n\t" + error.emit_exit("case_clause") + ";\n\t}\n\t"; 
    return ret; 
}

void CaseList::set_parent(Case * x)
{
    list<Node *>::iterator i;
    CrClause * pack;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        pack = (CrClause *)(*i);
        if (pack != NULL)
        {
            pack->head->parent = x;
        }
    }
}

int Case::_labelcount = 0;

//	Node * expr;
//   NodeList * clauses;

Case::Case(Node * e, CaseList * cs)
{
	_label = "c" + itoa(_labelcount);
	_labelcount++;

	expr = e;
	clauses = cs;	
    stmt_type(ST_stmt);
}

list<string> Case::emit_code()
{
    Symbol * s;
    string ret, RHS, pre;
	string pam = _label;
    list<string> lret;

    add_pre_code(expr->pre_code());

    RHS = expr->expr_code();
    
	ret = "\t" + pre + pam + " = " +  RHS;
    ret = ret + ";\n\t";
    lret.push_back(ret);
    ret = clauses->emit_code(pam);
    lret.push_back(ret);

    return lret;
}

string Case::print()
{
	return "case " + expr->print() + "\n" + clauses->print() + "\n";
}

string Case::label()
{
	return _label;
}

void Case::set_depth(int lvl)
{
    depth = lvl;
    expr->set_depth(lvl);
    clauses->set_depth(lvl);
}

//	Node * guard, * body;
CrClause::CrClause(Node * p, Node * g, Node * b)
{
	head = (CrHead *)p;
	guard = g;
	body = (ClauseBody *) b;
    stmt_type(ST_stmt);
}

list<string> CrClause::emit_code()
{
	assert(head != NULL);
	assert(body != NULL);
	// guard may be NULL!

	string ret;

    // emit the "if"
    // emit the pattern matching assignments
    // emit the rest of the code


	ret = "\tif (cX == " + head->expr_code() + ")";

	if (guard != NULL)
	{
		ret = ret + guard->expr_code();
	}
    ret = ret + " ? ";

    // body->delimiter(", ");
    ret = ret + "(" + body->expr_code() + "):\n\t";
	
	// ret = ret + "\n\t{\n\t" + body->emit_code() + "\t}\n";

	return string2list(ret);
}

string CrClause::emit_pattern(string p)
{
	string ret, res, prestr;
    list<string> pre;
    list<string>::iterator i;
    // int assignflag = 0;

    p = p;
	assert(head != NULL);
	assert(body != NULL);

    SymbolTable * st = head->table;

    assert(st != NULL);

    TableStack.push_front(st);
    CurrentTable = TableStack.front();

	res =  "if ( " + head->emit_pattern(head->parent->label()); 
        // + head->parent->label() + "L == " 

    // should each case entity have it's own symbol table?
    // before emitting the body of the code we should emit
    // variable assignment (based on the pattern match

	if (guard != NULL)
	{
        ret = res + " && " + guard->expr_code() 
                + ")\n\t{\n\t";
    }
    else
    {
        ret = res + ")\n\t{\n\t";
    }

    // Emit the code in the body 
    
    // clear any assignable?
    body->assign(NULL);

    pre = body->pre_code();
    for (i = pre.begin(); i != pre.end(); i++)
        prestr = prestr + (*i);
    
    ret = ret + prestr + body->expr_code() + "\n\t}\n\t";


    TableStack.pop_front();
    CurrentTable = TableStack.front();

	return ret;
}

string CrClause::print()
{
	assert(head != NULL);
	assert(body != NULL);

	string ret;

	ret = head->print();
	if (guard) ret = ret + guard->print();
	ret = ret + body->print();

	return ret;
}

void CrClause::set_depth(int lvl)
{
    depth = lvl;
    head->set_depth(lvl);
    if (guard != NULL) guard->set_depth(lvl);
    body->set_depth(lvl);
}

CrHead::CrHead(string i, SymbolTable * st, Node * pat)
{
    id = i;
    table = st;
    pattern = pat;
}

list<string> CrHead::emit_code()
{
    return pattern->emit_code();
}

string CrHead::emit_pattern(string p)
{
    return pattern->emit_pattern(p);
}

string CrHead::print()
{
    return pattern->print();
}

void CrHead::add_symbol(SymbolTable * st, int df)
{
#ifdef DEBUG
    fprintf(stderr, "CrHead add_symbol %s\n", st->symbol().c_str());
#endif
    pattern->add_symbol(st, df);
}

