

//
// $Id: case.h,v 1.13 2001/06/18 15:32:09 geoff Exp $
//
// $Source: /home/cvs/gerl/src/case.h,v $
// $Revision: 1.13 $
// $Date: 2001/06/18 15:32:09 $
// $State: Exp $
//
// Author: Geoff Wong
//

#ifndef _CASE_H
#define _CASE_H

//
// Looks of "friend" ugliness herein.
//

#include "node.h"
#include "function.h"
#include "symbol.h"


class Case;

// a slightly modified NodeList
class CaseList : public NodeList
{
    friend class CrHead;
    friend class CrClause;
public:
    CaseList(Node *);
    string emit_code(string);
    void set_parent(Case * x);
        // PURPOSE: set the parent in each CrHead (pointing to the Case it's in)
};

// Encapsulates the Case statement
class Case : public Node
{
protected:
    static int _labelcount;

    string _label;
    Node * expr;
    CaseList * clauses;
public:
    Case(Node *, CaseList *);
    list<string> emit_code();
    string print();
    string label();
    void set_depth(int);
};

class CrHead : public Node
{
    friend class CrClause;
    friend class CaseList;

    Case * parent;
    Node * pattern;
    SymbolTable * table;
    string id;

public:
    CrHead(string, SymbolTable *, Node *);

    list<string> emit_code();

    string emit_pattern(string p);

    string print();

    void add_symbol(SymbolTable * st, int dt = LT_none);
        // add the symbols in the pattern to the symbol table
};

class CrClause : public Node
{
    friend class CrHead;
    friend class CaseList;

    CrHead * head;
    Node * guard;
    ClauseBody * body;
public:
    CrClause(Node *, Node *, Node *);
    list<string> emit_code();

    string emit_pattern(string p);
        // Obviously a Case Clause can't be a parameter
        // instead I use it to identify the case label to be checked
        // against

    string print();

    void set_depth(int lvl);
        // Set depth of all members
};


#endif
