

//
// $Id: elist.cc,v 1.38 2002/08/03 06:17:27 geoff Exp $
//
// $Source: /home/cvs/gerl/src/elist.cc,v $
// $Revision: 1.38 $
// $Date: 2002/08/03 06:17:27 $
// $State: Exp $
//
// Author: Geoff Wong
//
// This is for an Expression List in list form [ ... ]
//

#include "elist.h"
#include "str.h"
#include "error.h"
#include "erlib.h"

extern int debug;

EList::EList(const string& tmp)
{
    storage = tmp;
	tail = 0;
    // post = new Symbol();
}

EList::EList(const string& tmp, const string& tmp1, const string& tmp2, 
             const string& tmp3, Node * n)
{
    storage = tmp;
    tmpstore1 = tmp1;
    tmpstore2 = tmp2;
    tmpstore3 = tmp3;
	tail = n;
//    post = new Symbol();
}

EList::~EList()
{
    delete post;
}

void EList::add_symbol(SymbolTable * st, int df)
{
    if (st == NULL) return;

    NodeList::add_symbol(st, df);
    // NodeList::add_symbol(st, LT_list);
    if (tail != NULL)
    {
        tail->add_symbol(st, df);
    }
}

bool EList::typecheck(Etype check, Etype restrict)
{
    bool res = true;
    list<Node *>::iterator i;

    if (tail != NULL) res = res && tail->typecheck(LT_list);

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        res = res && (*i)->typecheck(LT_all, restrict);
    }
    return res;
}

void EList::type(int t)
{
    if (debug) fprintf(stderr, "EList::type(%d)\n", t);
    Node::type(t);
    if (tail != NULL) tail->type(t);
}

list<string> EList::emit_code()
{
    int x, type;
    bool special = false;
    string res;
    string ret;
    string str;
    string pre;
    list<Node *>::iterator i;


    if (tail != 0 && nodes.size() == 1)
    {
        res = "(";
        special = true;
    }
    else if (nodes.size())
	{
		res = "(Emake_list(" + itoa(nodes.size());
	}
	else
	{	
		res = "(Emake_list(" + itoa(nodes.size());
	}

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        Symbol * s;
        string RHS;
        list<string> tpl;

        if (special != true) 
        {
            // delimit them...
            res = res + ",";
        }

        // Ensure the values put into these lists aren't freed 
        // (note: should this only be the case for the return value?)
        //(*i)->freesym(NULL);

        // FIX: should check for assignable code first
        // and emit the pre list building.
        type = (*i)->type();
        tpl = (*i)->pre_code();
        //printf("EList:-##%s##-\n", join(tpl, "\n").c_str());
        add_pre_code(tpl);
        RHS = (*i)->expr_code();
        //x = RHS.find(';'); if (x != -1) RHS[x] = ',';

        res = res + RHS;
    }

    i = nodes.end();
    i--;
    if (i != nodes.end()) (*i)->freesym(NULL);

    res = res + ")";
    if (special != true) res = res + ")";

    // deal with a tail concatentation 
	if (tail != 0)
	{
        string l,r;

        // Ensure the values put into these lists aren't freed 
        // (note: should this only be the case for the return value?)
        // tail->freesym(NULL);
        add_pre_code(tail->pre_code());
        r = tmpstore2 + "=" + tail->expr_code();
        tail->type(tail->type() & ~LT_on_heap);
        if (special) 
        {
#ifdef _CONS_H
            ret = "("+r+",";
            ret = ret + "Econcatvl("+tmpstore1+","+res+","+tmpstore2+"))";
#else
            ret = "("+r+",";
#ifdef _ALLOCS_H
            ret = ret + "Econcatvl("+tmpstore1+","+res+","+tmpstore2+"))";
#else
            if (! (((type & LT_on_heap))
                 || ((type & LT_anyvar) < LT_list
                 && (type & LT_anyvar) > LT_none)))
            {
                res = "EValcopy"+res;
            }

            ret = ret + "Econcatvl("+res+","+tmpstore2+"))";
#endif
#endif
        }
        else 
        {
            l = tmpstore3 + "=" + res;  
#ifdef _CONS_H
            ret = "("+l+"," +r+",";
            ret = ret + "Econcat("+tmpstore1+","+tmpstore3+","+tmpstore2+"))";
#else
            ret = "("+l+"," +r+",";
            ret = ret + "Econcat("+tmpstore3+","+tmpstore2+"))";
#if 0
            ret = "("+l+"," +r+",listalloc("+tmpstore1+",Lsize("+tmpstore3+")+"
                    +"Lsize("+tmpstore2+")"+"),";
            ret = ret + "Econcat("+tmpstore1+","+tmpstore3+","+tmpstore2+"))";
#endif
#endif
        }

        // FIX: this is "post" (after) code
        // shouldn't really be stuck in a symbol ...
        // post->symbol(res);
        //add_pre_code(string2list(pre));
	}
    else 
    {
        ret = res;
        if (pre != "") 
        {
            // FIX: this is "post" (after) code
            // shouldn't really be stuck in a symbol ...
            // post->symbol(ret);
            add_pre_code(string2list(pre));
        }
    }

	return string2list(ret);
}

string EList::print()
{
    string ret;
    list<Node *>::iterator i;

    ret = "[";
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
		if (i != nodes.begin())
		{
			// shouldn't do this if it's "tail" item..
			// where should this be FIXED?
			ret = ret + ",";
		}
        ret = ret + (*i)->print();
    }
    ret = ret + "]";
    return ret;
}

string EList::emit_pattern(string p)
{
    list<Node *>::iterator i;
    string ret, nxt;
    int count = 0;

    if (nodes.size() == 0)
    {
        ret = p;
        return ret;
    }

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
#ifdef _CONS_H
        ret = ret + (*i)->emit_pattern("("+p+".u.list"+nxt+"->e)");
        nxt = nxt + "->next";
        count++;
#else
        ret = ret + (*i)->emit_pattern("Eindex_list("+p+","+itoa(count++)+")");
#endif
    }

    // FIX: use an allocated tmp variable instead of __vvv
    if (tail)
    {
#ifdef _CONS_H
        ret = ret + tail->emit_pattern("(Entail_list("
        +p+".u.list"+nxt+","+itoa(count)+",\""+ error.function() + "\"))");
#else
#ifdef _ALLOCS_H
        ret = ret + tail->emit_pattern("(Entail_list(__vvv,"
#else
        ret = ret + tail->emit_pattern("(Entail_list("
#endif
        +p+","+itoa(count)+",\""+ error.function() + "\"))");
#endif
    }

    return ret;
}

void EList::set_depth(int x)
{
    depth = x;
}

#if 0
Symbol * EList::assign()
{
    return post;
}


ETail::ETail()
{
}

ETail::ETail(Node * n)
{
	value = n;
}

list<string> ETail::emit_code()
{
	return string("");
}

string ETail::print()
{
	return "|" + value->print();
}

#endif
