%{
//
// $Id: erlang.l,v 1.43 2002/11/06 12:39:13 geoff Exp $
//
// $Source: /home/cvs/gerl/src/erlang.l,v $
// $Revision: 1.43 $
// $Date: 2002/11/06 12:39:13 $
// $State: Exp $
//
// Author: Geoff Wong
//
//
// Temporary Lexical Analyser for Erlang
// Copyright (C) 1995,1996,1997,1998 -- Geoff Wong.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "define.h"
#include "nodes.h"
#include "case.h"
#include "y.tab.h"
#include "illegal.h"
#include "str.h"
#include "dictionary.h"
#include "yy.h"
#define yywrap() 1

extern YYSTYPE yylval;
extern int debug;

#define DEFAULT_LIB_PATH "/usr/local/lib/erlang/lib"

/* defines hash table */
/* handle -define, -ifdef .. else  etc */
#define MAX_DEFINE 256
#define MAX_RECORD 256
#define MAX_DEFINE_PARAM 16
Dictionary<DefClause *> definitions;
int br_depth = 0;

/* include handler stack */
#define MAX_INCLUDE_DEPTH 16
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
const char *    include_file[MAX_INCLUDE_DEPTH];
int             include_line[MAX_INCLUDE_DEPTH];

int include_stack_ptr = 0;

int un_include();
int do_include(char *);
int do_includelib(char * text);
int perform_include(char * fname);

char * eat_define();

/* keep count of where we're up to (in each file)? */
int current_line = 1;
const char * current_file = NULL;

int number(char * s),
    scinum(char * s),
    realnum(char * s),
    hex(char * s),
    octal(char * s),
    uglynumber(char * s),
    atom(char * s),
    qatom(char * s),
    estring(char * s)
    ;

string * 
    legal_ident(const char * str)
    ;

void 
    add_define(char * s, void *, char *),
    sub_define(char * s),
    sub_record_access(char * s)
    ;

#if 0
 T_bor T_bxor T_bsl T_bsr T_equal T_notequal T_le T_lt
 T_ge T_gt T_exactequal T_div T_rem T_band T_bnot T_atom
 T_number T_string T_var T_bool T_begin T_end T_catch 
 T_case T_of T_if T_receive T_after T_end T_when T_arrow 
realnumber ([0-9]+'.'[0-9]+'e'?'-'?[0-9]*)
realnumber ([0-9]+'.'[0-9]+'e'?'-'?[0-9]*)
#endif

#if 0
{defident}  { 
        /* insert identifier stuff into parse stream */
        }
#endif

%}

ident	([A-Z][a-zA-Z_0-9]*)
defident (\?[A-Za-z][a-zA-Z_0-9]*)
recaccess ([^ \t\n].*#[a-z_][a-zA-Z_0-9]*\.[a-z_][a-zA-Z_0-9]*)
number  ([1-9][0-9]*|0)
octal (\$\\[0-9][0-9][0-9])
hex     (16#[0-9a-fA-F][0-9a-fA-F]*)
base8     (8#[0-9][0-9]*)
uglynumber  (\$\\.|\$.|\$\\^.)
realnumber ([0-9]+\.[0-9]+)
scinumber ([0-9]\.[0-9]+[eE][+-][0-9]+)
qatom	(\'([^\n\']|\\\')*\')
atom 	([a-z_][a-zA-Z_0-9]*)
anyident 	([a-zA-Z_][a-zA-Z_0-9]*)
quotedstring (\"([^\"]|\\\")*\")
operator        ([\+\*=\-/@^:{}&\~\^\()\[\],;\.\|])
defoperator ([,])
comment  (%[^\n]*)
whitespace ([ \t]*)

%x pragma
%x incl
%x inclib
%x define

%%

"-include" { BEGIN(incl); }
"-include_lib" { BEGIN(inclib); }
"-define"  {
        BEGIN(define); 
        return T_define;
    } 
{defident} { 
        sub_define(&(yytext[1]));
        break;
    }
"-record"   { return T_record; }
"."         { return T_dot; }
"#"         { return T_rechash; }
"!"		    { return T_send; }
"++"        { return T_concat; }
"--"        { return T_listsub; }
"->"		{ return T_arrow; }
"<-"		{ return T_generator; }
"=="		{ return T_equal; }
"=/="		{ return T_notequal; }
"/="		{ return T_notequal; }
"=<"		{ return T_le; }
"<"		    { return T_lt; }
">="		{ return T_ge; }
">"		    { return T_gt; }
"=:="		{ return T_exactequal; }
"||"		{ return T_such; }
{operator}	{ return yytext[0]; }
"after"		{ return T_after; }
"band"		{ return T_band; }
"bor"		{ return T_bor; }
"bsl"		{ return T_bsl; }
"bsr"		{ return T_bsr; }
"bxor" 		{ return T_bxor; }
"begin"		{ return T_begin; }
"bnot"		{ return T_bnot; }
"case"		{ return T_case; }
"catch"		{ return T_catch; }
"div"		{ return T_div; }
"end"		{ return T_end; }
"if"		{ return T_if; }
"of"		{ return T_of; }
"pragma"    { BEGIN(pragma); return T_pragma; }
"receive"	{ return T_receive; }
"rem"		{ return T_rem; }
"true"		{ 
			yylval.truth = 1;
			return T_bool; 
		}
"false"		{	
			yylval.truth = 0;
			return T_bool;
		}
"fun"       { return T_fun; }
"when"		{ return T_when; }
{realnumber} { return realnum(yytext); }
{scinumber} { return scinum(yytext); }
{qatom}		{ return qatom(yytext); }
{atom}		{ return atom(yytext); }
{hex}       { return hex(yytext); }
{base8}       { return hex(yytext); }
{octal}     { return octal(yytext); }
{number}	{ return number(yytext); }
{uglynumber} { return uglynumber(yytext); }
{ident} 	{ 
		yylval.stringp = legal_ident(yytext);
		return T_var; 
		}
{comment} { break; }
{quotedstring} { return estring(yytext);				}
""	{ break; }
"\n"	{ current_line++; break; }
[ \t]	{ break; }
<<EOF>> { 
    if ( --include_stack_ptr < 0 ) 
    {
        if (debug) fprintf(stderr, "yyterminate()\n");
        yyterminate();
    }
    else    
    {
        if (debug) fprintf(stderr, "un_include()\n");
        un_include(); 
    }
    }
.	{
		yyerrorf("Unknown character (%c)",yytext[0]);
	}
<pragma>"end pragma;" { 
//    fprintf(stderr, "END PRAGMA\n"); 
    BEGIN(INITIAL); 
    return T_endpragma; 
    }
<pragma>"\n"         { 
//    fprintf(stderr, "PRAGMA current_line++\n"); 
    current_line++; break; 
    }
<pragma>.*       { 
//    fprintf(stderr, "PRAGMA %s\n", yytext); 
    yylval.stringp = new string(yytext);
    return T_text; 
    }
<incl>{whitespace}    { break; }
<incl>(\([^ \t\n]+\)\.)  { do_include(yytext); break; }
<inclib>{whitespace}    { break; }
<inclib>(\([^ \t\n]+\)\.)  { do_includelib(yytext); break; }
<define>{whitespace}  { break; }
<define>{anyident} { return atom(yytext); }
<define>"("     { br_depth++; return yytext[0]; }
<define>")"     { br_depth--; return yytext[0]; }
<define>","     {
    if (br_depth == 1)
    {
        yylval.stringp = new string(eat_define());
        BEGIN(INITIAL);
        return T_defbody;
    }
    return yytext[0];
}
<define>.	{
		yyerrorf("Unknown character (%c)", yytext[0]);
	}

%%

void start_new_file(const char * name, FILE *f)
{
    current_file = name;
    yy_init = 1;
    yyin = f;
}

int number(char * s)
{
	yylval.number = atoi(s);
	return T_number;
}

int realnum(char * s)
{
    yylval.stringp = new string(s);
    return T_real;
}

int scinum(char * s)
{
    yylval.stringp = new string(s);
    return T_real;
}

//
// FIX: needs error checking on length of string (etc) 
//
int uglynumber(char * s)
{
	if (s[1] == '\\')
	{
		switch (s[2])
		{
			case 'n':
				yylval.number = '\n';
				break;
			case 't':
				yylval.number = '\t';
				break;
			case 'b':
				yylval.number = '\b';
				break;
			case '\\':
				yylval.number = s[2];
				break;
            case '^':
                yylval.number = s[3] - 64;
                break;
			default:
				yylval.number = s[2];
				break;
		}
	}
	else
	{
		yylval.number = s[1];
	}
	return T_number;
}

//
// FIX: needs to actually work...
//
int octal(char * s)
{
    yylval.number = strtol(&(s[2]), NULL, 8);
    return T_number;
}

// FIX: needs to actually work...
int hex(char * s)
{
    int base;
    char * ind;

    base = atoi(s);
    ind = index(s, '#');
    yylval.number = strtol(&(ind[1]), NULL, base);
    return T_number;
}

#define MAX_QUOTES 256

int qatom(char * s)
{
    int x;
    string tmp, ret;

	assert(strlen(s) >= 2);

    tmp = string(&s[1], strlen(s)-2);
    x = tmp.find('"');
    while (x != -1)
    {
        if (x != 0)
            ret = ret + tmp.substr(0, x-1);
        ret = ret + "\\\"";
        tmp = tmp.substr(x+1, tmp.length()-1);
        x = tmp.find('"');
    }
    ret = ret + tmp;
	yylval.stringp = new string(ret);
	return T_atom;
}

int atom(char * s)
{
	yylval.stringp = new string(s);
	return T_atom;
}

int estring(char * s)
{
    char * buf = new char[strlen(s) * 2 +1];
    int i = 0, count = 0;

    while (s[i])
    {
        buf[count++] = s[i];
        if (s[i] == 92 && 
            s[i+1] != '"' &&
            s[i+1] != 'r' &&
            s[i+1] != 't' &&
            s[i+1] != 'n'
            ) 
            buf[count++] = 92;
        i++;
    }
    buf[count] = '\0';
	yylval.stringp = new string(buf);
    delete [] buf;
	return T_string;
}

#define IDENTMAX 1024
string * legal_ident(const char * str)
{
	char buf[IDENTMAX];	
	int loc = 0;
	int i;

//	printf("legal_ident (%s)\n", str);
//	printf("sizeof illegal (%d)\n", sizeof(illegal));
	assert(yytext != NULL);
	for (i = 0; i < (int)(sizeof(illegal) / sizeof(char *)); i++)
	{
		if (strcmp(illegal[i], yytext) == 0)
		{
			loc = 2;
			break;
		}
	}
	
	if (loc != 0)
	{
		buf[0] = '_'; buf[1] = '_';
		strncpy(&buf[loc], yytext, IDENTMAX-2);
		buf[strlen(yytext)+2] = '\0';
		return new string(buf);
	}
	else
	{
		return new string(str);
	}
}

/*
 * Preprocessor support stuff is below
 * Note: probably should be split out.
 */

int do_includelib(char * text)
{
    char fullname[MAX_DEFINE+1];
    char * libpath;
    char * fname;

    libpath = getenv("ERLANG_LIB_PATH");
    if (libpath == NULL)
    {
        strncpy(fullname, DEFAULT_LIB_PATH, MAX_DEFINE);
    }
    else
    {
        strncpy(fullname, libpath, MAX_DEFINE);
    }
    
    // FIX: Need to search through all libs in ERLANG_LIB_PATH
    // and find the .hrl file in question (if it exists)
    // then run -include_lib
    if (fullname[strlen(fullname)-2] != '/')
        strcat(fullname, "/");

    strncat(fullname, &(text[2]), MAX_DEFINE);
    fullname[strlen(fullname)-3] = '\0';

    return perform_include(string_copy(fullname));
}

int do_include(char * text)
{
    char * fname;

    // FIX: Need to search through all include paths
    // and find the .hrl file in question
    
    fname = string_copy(&(text[2]));
    fname[strlen(fname)-3] = '\0';

    return perform_include(fname);
}


int perform_include(char * fname)
{
    FILE * yynew;

    if (debug)
        fprintf(stderr, "do_include (%s)\n", fname);

    /* extract file name */
    /* got the include file name */
    if (include_stack_ptr >= MAX_INCLUDE_DEPTH)
    {
        fprintf(stderr, "-include() nested too deeply\n");
        exit(1);
    }

    yynew = fopen(fname, "r");
    if (yynew == NULL)
    {
        yyerrorf("Unable to -include(\"%s\").\n", fname);
        // FIX: causes a seg fault in later processing
        //      not a good recovery from a bad include(lib)
        BEGIN(INITIAL);
        return 0;
        // exit(1);
    }

    yyin = yynew;
    include_file[include_stack_ptr] = current_file;
    include_line[include_stack_ptr] = current_line;
    include_stack[include_stack_ptr++] = YY_CURRENT_BUFFER;

    current_file = fname;
    current_line = 1;

    yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
    BEGIN(INITIAL);
    return 1;
}

int un_include()
{
    if (debug) fprintf(stderr, "un_include\n");

    yy_delete_buffer( YY_CURRENT_BUFFER );
    current_line = include_line[include_stack_ptr];
    current_file = include_file[include_stack_ptr];
    yy_switch_to_buffer(include_stack[include_stack_ptr]);

    return 0;
}

/*
 * Name: add_define
 * Purpose: actually subtitute build up a def
 */

void add_define(DefClause * defclause)
{
    definitions.add_element(defclause->defstring().c_str(), defclause);
}

/*
 * Name: sub_define
 * Purpose: actually substitute in a ?define
 * Note: use  yy_buffers? or squirt it back into the current buffer?
 */
void sub_define(char * def)
{
    DefClause * dc;
    list<Node *>::iterator ni;
    NodeList * nl = NULL;
    string body;
    char c = EOF;

    char buf[MAX_DEFINE+1]; 
    string * params[MAX_DEFINE_PARAM];
    int depth = 0, count = 0, noparams = 0;
    int i;

    dc = definitions[def];

    if (debug) fprintf(stderr, "sub_define_params(%s)\n", def);

    // special cases
    if (strncmp(def, "MODULE", 6) == 0)
    {
        // clip ".erl" from current_file name.
        if (strlen(current_file) > 4)
        {
            body = current_file;
            body = body.substr(0,body.length()-4);
        }
        else
        {
            body = current_file;
        }
    }
    else if (dc == NULL)
    {
        yyerrorf("Unable to substitute ?%s\n", def);
        return;
    }
    else
    {
        body = dc->body();
    }
    c = yyinput();

    //  consume defined params into an array of strings for use
    if (c == '(') while (c != EOF)
    {
        if (c == ')') 
        {
            depth--;
            if (depth < 0)
            {
                yyerrorf("Illegal bracketing of definiton ?%s\n", def);
                return;
            }
        }
        else if (c == '(') 
        {
            // should  we ignore whitespace before first '('?
            if (depth != 0) buf[count++] = c;
            depth++;
            c = yyinput();
            continue;
        }

        if (depth == 0)
        {
            buf[count] = '\0';
            if (debug) fprintf(stderr, "final buffer=%s\n", buf);
            count = 0;
            params[noparams] = new string(buf);
            noparams++;
            br_depth = 0;
            break;
        }

        // split it into an array 
        if (depth == 1 && c == ',')
        {
            buf[count] = '\0';
            count = 0;
            params[noparams] = new string(buf);
            noparams++;

        }
        else
        {
            buf[count++] = c;
        }
        c = yyinput();
    }
    else
    {
        unput(c);
    }

    if (dc != NULL) nl = dc->plist();

    if ((!nl && noparams) || (nl && (noparams != nl->size())))
    {
        if (debug && nl)
        {
            fprintf(stderr, "list size = %d, noparams = %d\n", nl->size(), noparams);
        }
        else if (debug)
        {
            fprintf(stderr, "no param llist, noparams = %d\n", noparams);
        }
        yyerrorf("Parameters don't match in substitution of definiton ?%s\n", def);
        return;
    }

    // ok - substitute them 
    if (noparams > 0)
    {
        noparams = 0;
        for (ni = nl->nodes.begin(); ni != nl->nodes.end(); ni++)
        {
            // substitute params into body 
            string tmp;
            tmp = ((*ni))->expr_code();
            if (debug) fprintf(stderr, "body=%s\nsubstituting %s for %s\n", body.c_str(), tmp.c_str(), params[noparams]->c_str());
            gsub(body, tmp, *(params[noparams]));
            noparams++;
        }
    }

    if (debug) fprintf(stderr, "unputting %s\n",body.c_str());
    // put it back on the lexer's stream
    for (i = body.length() - 1; i >= 0; i--)
    {
        unput(body[i]);
    }
}


/*
 * Name: eat_define
 * Purpose: actually suck in a define
 * Returns: an allocated string which is the define
 */
char * eat_define()
{
    char def[MAX_DEFINE+1]; 
    int c, count = 0;
    int depth = 1;

    while ((c = yyinput()) != EOF)
    {
        if (c == ')') 
        {
            depth--;
        }
        else if (c == '(') depth++;
        if (depth == 0)
        {
            unput(c);
            def[count] = '\0';
            br_depth = 0;
            return string_copy(def);
        }
        def[count++] = c;
    }
    yyerror("EOF in -define()");
    exit(1);
    return NULL;
}

void sub_record_access(char * def)
{
    string body, record, field, variable;
    char buf[MAX_RECORD];
    char rec[MAX_RECORD];
    char fld[MAX_RECORD];
    int dex, count, i;
    char * c;

    if (debug) fprintf(stderr,"sub_record_access(%s)\n", def);

    /* Ok - lets work out the record access */
    /* 1st - get the var                    */
    count = 0;
    c = def;
    while (*c != '\0' && *c != '#')
    {
        buf[count++] = *c;
        c++;
    }
    if (*c == '\0') 
    {
        yyerror("EOF(#) in record access");
        exit(1);
    }
    buf[count] = '\0';

    /* eat the '#' */
    c++;
    /* 2nd - get the record */
    count = 0;
    while (*c != '\0' && *c != '.')
    {
        rec[count++] = *c;
        c++;
    }
    if (*c == '\0') 
    {
        yyerror("EOF(.) in record access");
        exit(1);
    }
    rec[count] = '\0';

    /* eat the '.' */
    c++;
    /* 3rd - get the field */
    count = 0;
    while (*c != '\0' && isalnum(*c))
    {
        fld[count++] = *c;
        c++;
    }
    fld[count] = '\0';

    /* find the index */
    variable=buf;
    record = rec;
    field = fld;
    dex = yy_program->find_record(record, field);
    if (dex == -1)
    {
        yyerror("Unknown field accessed in record");
        exit(1);
    }

    /* set the code we want to put back on */
    body = body + "element(" + variable + "," + itoa(dex) + ")";
    if (debug) fprintf(stderr, "unputting %s\n",body.c_str());

    // put it back on the lexer's stream
    for (i = body.length() - 1; i >= 0; i--)
    {
        unput(body[i]);
    }

    return;
}
