

//
// $Id: erlangshell.cc,v 1.26 2002/12/09 13:12:09 geoff Exp $
//
// $Source: /home/cvs/gerl/src/erlangshell.cc,v $
// $Revision: 1.26 $
// $Date: 2002/12/09 13:12:09 $
// $State: Exp $
//
// Author: Geoff Wong
//

/*
 * A simple Erlang shell
 * 
 * Current supports:
 *         module.func(x,y,z)
 * So we can start programs (dynamically).
 * 
 * Should be extended to interpret arbitrary Erlang expressions
 * Also should be runnable as !#/usr/local/bin/erlsh
 *
 * Add support for doing shell type I/O?
 */


#include "erlangshell.h"
#include "erlib.h"

const int MAX_LINE=1024;
const int MAX_PARAMS=256;

ErlangShell * ErlangShell::_instance = 0;

ErlangShell::ErlangShell()
{
}

ErlangShell * ErlangShell::Instance()
{
    if (_instance)
    {
        return _instance;
    }
    _instance = new ErlangShell;
	return _instance;
}

#define MAXROOT 1024
void ErlangShell::Shell()
{
	string sin;
	char * tmproot;
	char buf[MAX_LINE];

	tmproot = getcwd(NULL, MAXROOT);
	_root = tmproot;
	_root = _root + "/";
	free(tmproot);

   // main input loop..
    while (1)
    {
		// do the prompt
		printf("gerl>> ");		
		fflush(stdout);

		if (fgets(buf, MAX_LINE, stdin) == 0) break;
		sin = buf;
        if (sin != "" && sin != "\n")
    		Interpret(sin);
    }
}

#if 0
void ErlangShell::Printvalue(Val ret)
{
    switch (ret.tag & LT_normal)
    {
    case LT_number:
        printf("%d", ret.u.number);
        break;
    case LT_atom:
        printf("%s", ret.u.atom);
        break;
    case LT_real:
        printf("%f", ret.u.real);
        break;
    case LT_tuple:
        {
            int i;
            printf("{ ");
            for (i = 0; i < ret.u.tuple->size; i++)
            {
                if (i != 0) printf(", ");
                Printvalue(ret.u.tuple->e[i]);
            }
            printf(" }");
        }
        break;
    case LT_list:
        {
            int i;
            _Lelement * le;
            printf("[ ");
            le = ret.u.list;
            while (le)
            {
                if (i != 0) printf(", ");
                Printvalue(le->e);
                le = le->next;
            }
            printf(" ]");
        }
        break;
    case LT_bool:
        if (ret.u.number == 0)
            printf("false");
        else
            printf("true");
        break;
    default:
	    printf("Unimplemented return type: %d", ret.tag);
		break;
    }
}
#endif

void ErlangShell::Interpret(string& str)
{
   // interpret a string (in a very limited fashion!)
   // extracts parameters does a Call if required
	Mangle m;
	Val ret;

	m.demangle(str);
	if (m.mangle() != "")
	{
		ret = Call(m);

        EPrintvalue(stdout,ret);
        printf("\n");
	}
}

Val ErlangShell::Call(Mangle& what)
{
	// do the function call and do it..
#if 0
	void * func;
	Val (*func0)(), (*func1)(Val), (*func2)(Val,Val),
		(*func3)(Val,Val,Val), (*func4)(Val,Val,Val,Val),
		(*func5)(Val,Val,Val,Val,Val), (*func6)(Val,Val,Val,Val,Val,Val);
	const char *error;
#endif
	Val ret;
	string module;

	module = what.module(); // + ".so";

    Val a = listmalloc(what._parameters.size());
	int count = 0;
	deque<Val>::iterator i;
	for (i = what._parameters.begin(); i != what._parameters.end(); i++)
	{
        Eindex_list(a, count) = (*i);
		count++;
	}

    // printf("calling %s:%s\n", module.c_str(), what.mangle().c_str());
    // FIX: should spwan monitor!
    //      call bif:spawn/3?
    //ret = Ecalllist(module.c_str(), what.mangle().c_str(), a);
    // ret = Emonitorcall(module.c_str(), what.mangle().c_str(), a);
    ret = Emonitorcall(module.c_str(), what.name().c_str(), a);
	
#if 0
    /* ugliness */
	switch(what.params())
	{
		case 0:
			func0 = func;
			ret = func0();
			break;
		case 1:
			func1 = func;
			ret = func1(a[0]);
			break;
		case 2:
			func2 = func;
			ret = func2(a[0], a[1]);
			break;
		case 3:
			func3 = func;
			ret = func3(a[0], a[1], a[2]);
			break;
		case 4:
			func4 = func;
			ret = func4(a[0], a[1], a[2], a[3]);
			break;
		case 5:
			func5 = func;
			ret = func5(a[0], a[1], a[2], a[3], a[4]);
			break;
		case 6:
			func6 = func;
			ret = func6(a[0], a[1], a[2], a[3], a[4], a[5]);
			break;
		case 7:
			fprintf(stderr, "Too many parameters to handle in this brain damaged shell\n");
			break;
	}
#endif
    listfree(a);

	return ret;
}
