

//
// $Id: expr.cc,v 1.66 2002/10/06 14:12:01 geoff Exp $
//
// $Source: /home/cvs/gerl/src/expr.cc,v $
// $Revision: 1.66 $
// $Date: 2002/10/06 14:12:01 $
// $State: Exp $
//
// Author: Geoff Wong
//
// ExprList - a list of expressions (like a list of statements
// in block-structured language).
//


#include "expr.h"
#include "str.h"
#include "program.h"
#include "symbol.h"

extern int debug;

void dumm() { };

ExprList::ExprList()
{
    // do nothing
    _delimiter = ";\n";
    recur_params = 0;
}

ExprList::ExprList(Node * n)
{
    nodes.push_front(n);    
    _delimiter = ";\n";
    recur_params = 0;
}

list<string> ExprList::emit_code()
{
    list<string> lcode;
    list<Node *>::iterator i;
    string ret, RHS, ex, pre, res, single;
    string childex;
    Symbol * s;

    if (debug) printf("    // exprlist: depth = %d\n", depth);
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        int rt = (*i)->type();

        if (i == --(nodes.end())) (*i)->set_depth(2);

        RHS = (*i)->expr_code();
        lcode = (*i)->pre_code();

        // Hmm?

        if (depth < 1 || flat_code()) pre = join(lcode, "");
        else add_pre_code(lcode);

        //printf("ExprList: +##%s##+\n", join(lcode, "\n").c_str());

        // FIX: null symbol hack - bad abuse of symbol
        // if (RHS == "") RHS = s->symbol();

        if (i == --(nodes.end()))
        {
            (*i)->set_depth(1);
            (*i)->freesym(NULL);

            // the following is way ugly and should really checks nodes..
            // should really always be true ..
            if ((RHS.find("return ")) == RHS.npos && (RHS.find("goto ")) == RHS.npos)
            {
                //Symbol * alt = assign();
                Symbol * alt = _assign;
                string retsym;

                dumm();
                if (alt != NULL) retsym = alt->symbol();
                else retsym = "_ret";
                if (debug) fprintf(stderr,"EL:ret_type = %d\n",  rt);

                res = "\t" + retsym + "=";

                if ( ((rt & LT_single) >= LT_list && (rt < LT_on_heap)) 
                                        || (rt & LT_param) ) 
                {
                    if ((*i)->stmt_type() != ST_goto) RHS = "EValcopy("+RHS+")";
                }

                // extra() stuff gets emitted as next full stmt
                if (depth < 1)
                {
                    // need to check for "goto" for extra stuff...
                    if ((*i)->stmt_type() == ST_goto)
                    {
                        res = "";
                        // ex = "cut_stack(sp);\n\trecur = 1;\n\tgoto " + extra() + ";";
                        ex = "recur = 1; continue;";
                    }
                    else
                    {
                        int loop;

                        // if we're part of a lco block emit parameter frees
                        if (recur_params != 0) 
                        {
                            for (loop = 0; loop < recur_params; loop++)
                            {
                                ex = ex + "Efred(p" + itoa(loop) + ");"; 
                            }
                            ex = "if (recur) {\n\t" + ex + "\n\t}\n";
                        }

                        ex = ex + "\treturn "+retsym+";";
                    }
                    extra(ex);
               }
            }
            else
            {

            }

        }

        // delimiter hack ..
        if (RHS.length() != 0)
        {
            single = pre + res + RHS;
            if (single.empty() == false) 
            {
                ret = ret + single + _delimiter;
            }
        }

#if 0
            ret = "#line " + itoa((*i)->line()) + " \"" + yy_program->Module() + ".erl\"\n" + ret;
#endif

    }
    return string2list(ret);

}

void ExprList::returned(SymbolTable * s)
{
    list<Node *>::iterator i;

    if (nodes.size())
    {
        i = --(nodes.end());
        // Need to get the correct current table
        (*i)->add_symbol(s, LT_returned);
    }
}

void ExprList::add_symbol(SymbolTable * st, int df)
{
    list<Node *>::iterator i;

    if (st == NULL) return;

    // NodeList::add_symbol(st, df);
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        if (i != --(nodes.end()))
        {
            (*i)->add_symbol(st, df);
        }
        else
        {
            if (debug) fprintf(stderr,"ExprList:: add_symbol LT_returned\n");
            (*i)->add_symbol(st, df | LT_returned);
        }
    }
}

string ExprList::print()
{
    string ret;
    list<Node *>::iterator i;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        if (i != nodes.begin())
        {
            ret = ret + ", ";
        }
        ret = ret + (*i)->print();
    }
    return ret;
}


string ExprList::emit_pattern()
{
    list<Node *>::iterator i;
    string ret;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        if (i == --(nodes.end())) 
        {
            ret = ret + (*i)->expr_code();
        }
        else
        {
            ret = ret + (*i)->expr_code() + ", ";
        }
    }
    return ret;
}


string ExprList::emit_local(SymbolTable * scope, const string& lab)
{
    list<Node *>::iterator i;
    string ret, param, pre, fre, code, freret;
    int count = 0, ttype;
    Symbol * fs;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        param = "p" + itoa(count++);
        pre = "";
        fre = "";
        code = (*i)->expr_code(); 
        ttype = (*i)->type();

        // Check if code is a variable in current symbol table 
        // if it is then don't free it ..
        // ie. fun ( [ A | B ] ) -> B.
        fs = scope->find(code, S_Variable);
        if (fs) 
        {
            if (debug) fprintf(stderr, "Found a param assignment variable %s\n", code.c_str());
            fs->type(fs->type() & ~LT_on_heap);
            if (!(ttype & LT_single)) ttype = fs->type();
        }

        if (ttype & LT_on_heap)
        {
            // if it's already on the heap we need to free it ...
            // pre = "o"+param+"="+param+";\n\t";
            fre = "Efre(o"+param+");";
            fs = (*i)->freesym();
            if (fs)
            {
                if (debug) fprintf(stderr, "Found a freesym() - NULLifying\n");
                fs->type(fs->type() & ~LT_on_heap);
            }
        }
        else if ( (!(ttype & LT_on_heap) && ((ttype & LT_single) >= LT_list))
                        || ((ttype & LT_param)))
        {
            // if not - copy it and free it 
            code = "EValcopy("  + code + ")";
            // pre = "o"+param+"="+param+";\n\t";
            fre = "Efre(o"+param+");";
        }

        freret = freret + fre;
        ret = ret + pre + param + " = " + code + ";\n\t";
    }

    if (freret.empty() == false)
    {
        freret = "if (recur) {\n\t" + freret + "\n\t}\n\t";
        ret =  ret + freret;
    }

    extra(lab);

    // ret = ret + "\tgoto " + lab;
    return ret;
}


string ExprList::emit_nonlocal(SymbolTable * scope)
{
    list<Node *>::iterator i;
    string pre, RHS, ret, param;
    int count = val_start;
    Symbol * s, * fs;

    // count = Program::val_count;
    // val_start = Program::val_count;
    // fprintf(stderr,"EL:val_start = %d\n", val_start);
    // Program::val_count += nodes.size();
    // FIX: should this should symbols & stuff before emitting?
    
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        param = "_vx" + itoa(count);
        s = scope->find(param, S_Variable);
        add_pre_code((*i)->pre_code());
        RHS = (*i)->expr_code();

        s->type(s->type() | (*i)->type());
        fs = (*i)->freesym();
        if (!fs) fs = scope->find(RHS, S_Variable);
        if (fs) 
        {
            if (!(fs->type() & LT_on_heap)) RHS = "EValcopy(" + RHS + ")";
            fs->type(fs->type() & ~LT_on_heap);
        }

        ret = ret + param + " = " + RHS + ",";

        count++;
    }
    return ret;
}

string ExprList::emit_nonlocal_list()
{
    list<Node *>::iterator i;
    string ret, param;
    int count = val_start;

    // fprintf(stderr,"ELenll:count = %d\n", val_start);
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        ret = ret +  "_vx" + itoa(count++);
        if ( i != --(nodes.end()) ) ret = ret + ",";
    }
    return ret;
}

#if 0
void ExprList::delimiter(const string& delim)
{
    _delimiter = delim;
}
#endif

#if 0
void ExprList::extra(const string& str)
{
    _extra = str;
}

string ExprList::extra()
{
    // fprintf(stderr, "extra=%s\n", _extra.c_str());
    return _extra;
}
#endif

#if 0
void ExprList::type(int t)
{
    list<Node *>::iterator i;

    // fprintf(stderr, "ExprList::type(%d)\n", t);
    Node::type(t);
#if 0
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        (*i)->type(t);
    }
#endif
}

int ExprList::type()
{
    return Node::type();
}
#endif

int ExprList::allocate_locals(int start)
{
    val_start = start;    
    return nodes.size();
}


#if 0
void ExprList::parent(ExprList *e)
{
        _parent = e;
}
#endif

