
//
// $Id: function.h,v 1.36 2002/08/02 15:25:49 geoff Exp $
//
// $Source: /home/cvs/gerl/src/function.h,v $
// $Revision: 1.36 $
// $Date: 2002/08/02 15:25:49 $
// $State: Exp $
//
// Copyright (C) Geoff Wong (1996, 1997, 1998)
// geoff@cs.rmit.edu.au
//
// This file is distributed under the terms of the
// GNU Public License, details of which can be found
// in the file "Copying" which accompanies this code.        
//

#ifndef _FUNCTIONCALL_H
#define _FUNCTIONCALL_H

#include "node.h"
#include "pattern.h"
#include "program.h"
#include "expr.h"
#include "namer.h"

//
// Purpose: Function support
//     A function consists of a number of function clauses
//
class FunctionList : public NodeList
{
protected:
    Program * program;

    string doparams(int);
        // actually construct a textual parameter list

    bool _lco;

public:
    FunctionList();

    FunctionList(Node *);

    string print();

    void set_program(Program * p);
        // allow the program to tell us where we live 
    string mangle(string& name, int param);
//    string demangle();

    list<string> emit_code();
    int verify(const string& name, int num);
        // verifies all the function clauses are legal

    bool lco();
    void lco(bool);
        // set/get the status of whether this function needs
        // to emit stack handling code for LCO jumps.
};

class AnonFunctionList : public FunctionList
{
public:
    list<string> emit_code();
};

enum Lastcall 
{
    LCO_non = 0,
    LCO_local = 1,
    LCO_nonlocal = 2,
    LCO_external = 3,
    LCO_localfun = 4
};

class FunctionCall: public Node
{
    Node * dmodule;
        // dynamic resolution?
    string module;
        // otherwise static mangling
    string name;

    ExprList * value;
    Lastcall _LCO;
        // are we last call optimisable?

    SymbolTable * _scope;

public:
    FunctionCall(const string *, const string *, ExprList *, Symbol * );
    FunctionCall(Node *, const string *, ExprList *, Symbol * );
    list<string> emit_code();
    string print();
    void LCO(Lastcall);
    void freesym(Symbol *);
        // this call should get a temp value so it can be freed.
    bool typecheck(Etype check, Etype restrict = LT_all);

    void scope(SymbolTable *);
        // set the symbol table we're using

};

class RemoteCall : public Node
{
public:
    RemoteCall(const string *, const string *, Node *);
    list<string> emit_code();
    string print();
};

class Atom;

class ClauseHead : public Node
{
    string name;
    string clausename;
    NodeList * expr;    
public:
    ClauseHead(string& internalname, string& clausename, NodeList *);
        // PURPOSE: initialise with an Atom (name) and list of expressions

    string param_match();
        // PURPOSE: emit that pattern matching part of the function

    list<string> emit_code();
        // PURPOSE: emit the main code body (variables assignment to params)

    string print();
        // PURPOSE: pretty print

    int num_params();
        // PURPOSE: return the number of parameters

    friend class FunctionClause;
};

class ClauseBody : public Node
{
    ExprList * value;

    SymbolTable  * _scope;
        // local symbol table ..

    bool _emitted;

    int _lco_params;
        
public:
        
    ClauseBody(ExprList * n);

    list<string> emit_code();

    string print();

    void set_depth(int);

    void delimiter(const string &str);

    void scope(SymbolTable * s);
        // Set our local variable scope

    SymbolTable * scope();
        // return our local symboltable

    string emit_locals();
        // build up the local declarations (& inits?)

    string emit_frees(int);
        // emit the required (local) frees at the end of the clause

    void add_final(SymbolTable * st);
        // forward onto the exprlist ..

    void assign(Symbol * x);
        // what we're actually assigned to 

    void set_emitted(bool v);
        // (set) Have we already emitted these symbols?

    void lco(int x) { _lco_params = x; }

    void set_flat(bool f) { _flat = f; value->set_flat(f); };

};

class When : public Node
{
    Node * value;
public:
    When(Node * n);
    list<string> emit_code();
    string print();
};

//
// Purpose: A Function Clause (according to the book)
// Essentially a wrapper for the other bits of a function
//
class FunctionClause : public Node
{
    ClauseHead    * head;
        // clause head (parameters etc)
    When        * guard;
        // `when' guard
    ClauseBody    * body;
        // the actual expressions

    bool _lco;
        // part of a last call optimisation?
        
public:
    FunctionClause(Node * h, Node * g, Node * b);
    list<string> emit_code();
    string print();

    int num_params();
        // number of parameters
    string name();

    string clausename();

    bool lco() { return _lco; };

    void lco(bool x) { _lco = x; };
        // set/get the status of whether this function needs
        // to emit stack handling code for LCO jumps.
};

class AnonFunction : public Node
{
    FunctionClause  * func;

public:
    AnonFunction(FunctionClause * f);

    list<string> emit_code();

    string print();

    int num_params();
        // number of parameters

    string name();

    string clausename();

};

#endif
