

//
// $Id: gerl.cc,v 1.31 2002/12/12 00:12:04 geoff Exp $
//
// $Source: /home/cvs/gerl/src/gerl.cc,v $
// $Revision: 1.31 $
// $Date: 2002/12/12 00:12:04 $
// $State: Exp $
//
// Author: Geoff Wong
//

// @@DOC
// <h2> Gerl </h2>
// Gerl is a dynamic loader for erlang modules, essentially is
// is a very simple interface that allows a process to be started.
// It accepts very simple Erlang expressions such as function calls
// and basic values which allow the user to begin a process.
//
// @@ENDDOC
 

#include <string.h>
#include <stdio.h>
#include <dlfcn.h>
#include <setjmp.h>
#include <signal.h>
#include <wait.h>

#include "erlangshell.h"
#include "erlib.h"
#include "mangle.h"

const string DefaultRoot = "/u/geoff/erlang/erlib/";

//
// Name: handle_child
// Purpose: SIGCHILD signal handler
//  Needs to be extended to handle the standard Erlang
//  signalling mechanism between processes.
//  Probably all nodes should be started from the "enode" process
//  which could then handle signals properly
//  The wait is here to ensure the child process isn't left as
//  a zombie.
//

void handle_child(int x)
{
    int status, retval;
    // printf("child signal %d received\n", x);
    while ((retval = waitpid(-1, &status, WNOHANG)) > 0);
    // printf("waitpid status=%d (%d)\n", retval, status);
}

//
// Name: main
// Purpose: deal with command line args and start it up.
//

int main(int argc, char *argv[])
{
	ErlangShell * e;
	string _libroot;
    int server_init = 0;
    const char * path;

    (void) signal(SIGUSR1, SIG_IGN);
    (void) signal(SIGCHLD, handle_child);
	
    // Some initialisations
    Eopen_module("liberl.so");
    Eopen_module("bif.so");
    Eopen_module("erlang.so");

	e = ErlangShell::Instance();

#if 1
    for (int i = 1; i < argc; i++)
    {
        fprintf(stderr, "argv[%d] = %s\n", i, argv[i]);
    }

    path = getenv("LD_LIBRARY_PATH");
    if (path != NULL)
        fprintf(stderr, "LD_LIBRARY_PATH %s\n", path);
    else
        fprintf(stderr, "LD_LIBRARY_PATH NOT SET\n");

#endif

restart:
    // Setup some base error handling
    if (setjmp(*(Epush_catch())))
    {
        Val v = Emake_catch();
        //Esendmonitor(v);
        Esendmonitor(EValatom("stop"));
        printf("shell (%d) caught error: ", getpid());
        EPrintvalue(stdout, v);
        printf("\n");
        // interactive shell?
        if (argc > 1) 
        {
            // need to wait for stop to be sent?
            usleep(100);
            Ecloseserver();
            return 0;
        }
        else goto restart;
    }

    // Run a command line version
    // Should have a flag to indicate `fetch call' from <pid>
    // which tells it which functions to run
    // this gets around the problem limited argv[] lengths
    if (argc > 1)
    {
        string all;
        int i;

        for (i = 1; i < argc; i++)
        {
            int do_exit = 0;
            if (argv[i][0] != '-') 
            {
                break;
            }
            if (strlen(argv[i]) < 2) continue;

            switch (argv[i][1])
            {
                case 'm':       /* monitor defintion */
                {
                    if (i + 3 >= argc)
                    {
                        fprintf(stderr, "gerl -monitor <addr> <port> <id#> -node <addr> <port> <magic_no>\n");
                        fprintf(stderr, "     note: indended for internal spawning support.\n");
                        return 0;
                    }
                    Einitmonitor((unsigned int) atoi(argv[i+1]), 
                        atoi(argv[i+2]), atoi(argv[i+3]));
                    i = i + 3;
                    break;
                }
                case 'n':       /* node */
                {
                    Val fun;
                    Mangle manger;
                    string mod, modso, funstr;

                    fprintf(stderr, "processing a node\n");
                    if (i + 3 >= argc)
                    {
                        fprintf(stderr, "gerl -node <addr> <port> <magic_no>\n");
                        fprintf(stderr, "     note: indended for internal spawning support.\n");
                        return 0;
                    }

                    /* initialise our home node */
                    Einitserver((unsigned int) atoi(argv[i+1]), atoi(argv[i+2]));
                    server_init = 1;

                    fun = Ereceivemagic((unsigned int) atoi(argv[i+3]));

                    mod = Eindex_tuple(fun,0).u.atom;
                    modso = mod + ".so";
                    funstr = Eindex_tuple(fun,1).u.atom;

                    Ecalllist( modso.c_str(),
                     manger.mangle(mod, funstr, 
                     // Eindex_tuple(fun,2).u.param->size).c_str(),
                     Eindex_tuple(fun,2).u.list->length).c_str(),
                     Eindex_tuple(fun,2));
                    return 0;
                }
                case 's':   /* set this process's home server */
                    if (i + 2 >= argc)
                    {
                        fprintf(stderr, "gerl -server <addr> <port>\n");
                        fprintf(stderr, "     note: indended for internal spawning support.\n");
                        return 0;
                    }

                    /* initialise our home node */
                    Einitserver((unsigned int) atoi(argv[i+1]), atoi(argv[i+2]));
                    server_init = 1;

                    i = i + 2;
                    do_exit = 1;
                    break;
                case 'z':   /* turn off monitoring */
                    ENomonitor();
                    break;
                case 'h':   /* help */
                    fprintf(stderr, "gerl - no args simply starts up a command line processor.\n");
                    fprintf(stderr, "gerl -node <addr> <port> <magic_no>\n");
                    fprintf(stderr, "     note: indended for internal spawning support.\n");
                    fprintf(stderr, "gerl -server <addr> <port>\n");
                    fprintf(stderr, "     note: indended for internal spawning support.\n");
                    fprintf(stderr, "Extra arguments on the command line are interpreted as a function call.\n");
                    exit(0);
            }
            if (do_exit == 1) break;
        }

        for (i = i; i < argc; i++)
        {
            all = all + argv[i];
        }

    // fprintf(stderr, "Interpreting: %s\n", all.c_str());
        if (server_init == 0) Einitserver(0, 0);
	    e->Interpret(all);
        return 0;
    }

    if (server_init == 0) Einitserver(0, 0);

	// Run it
	e->Shell();
	return 0;
}


