
//
// $Id: main.cc,v 1.23 2002/07/29 08:32:12 geoff Exp $
//
// $Source: /home/cvs/gerl/src/main.cc,v $
// $Revision: 1.23 $
// $Date: 2002/07/29 08:32:12 $
// $State: Exp $
//
// Author: Geoff Wong
// Geoff Wong (C) Copyright 1996, 1997, 1998.
// C/O RMIT University
// Swanston St. Melbourne 3001.
// Australia
//
// email: geoff@serc.rmit.edu.au
//
// This file is distributed under the terms of the 
// GNU Public License, details of which can be found
// in the file "Copying" which accompanies this code.
//


#include <stdio.h>
/* #include "emit.h" */
#include "dictionary.h"
#include "element.h"
#include "program.h"
#include "str.h"
#include "define.h"
#include "yy.h"
#include "function.h"
#include "form.h"

// @@DOC
// <h3> Compilation Options </h3>
// A number of different compilation options exists

enum Mode { 
		Mnormal, 		//
		Mprint, 		// just pretty print the file
		Mcompile,		// compile with dynamic PIC code
		Mshared,		// compile with dynamic PIC code, shared libraries
                        // direct calls to functions (causes uncaught errors?)
                        // auto-loading of modules that are depended on
		Mstatic 		// compile for static linking
		};

// When compiling normally (with dynamic code) then an <tt> ERLANG_LIB </tt>
// environment variable should be specified which indicates
// the path(s) to search for already compiled modules

#define MAX_PATHS 256

// @@ENDDOC

int debug = 0;
extern Program * yy_program;

Dictionary<Element *> mainST;

// Local to the AST walker and code emitter
SymbolTable  * CurrentTable = 0;
list<SymbolTable *> TableStack;      

main(int argc, char *argv[])
{
	FILE *f;
	int wf = 1;
	Mode mode = Mcompile;
    string statix;

    if (argc < 2) {
            printf("backend <filename>\n");
            printf("  -d debug\n");
            printf("  -p pretty print\n");
            printf("  -s <fun/x> static compilation\n");
            printf("  -h shared libraries\n");
            printf("  -c compile to C code (default)\n");
            return 0;
    }

    while ((wf < argc) && (*(argv[wf]) == '-')) 
	{
			if (strlen(argv[wf]) > 1)
			{
				switch (argv[wf][1])
				{
					case 'd':
            			debug = 1;
						break;
					case 'h':
						mode = Mshared;
						break;
					case 'p':
						mode = Mprint;
						break;
					case 's':
						mode = Mstatic;
                        statix = argv[wf+1]; 
                        wf++;
						break;
					case 'c':
					default:
						mode = Mcompile;
						break;
				}
			}
            wf++;
    }

    /* open the file */
    f = fopen(argv[wf], "r");

    if (!f) 
	{
            fprintf(stderr, "%s <filename>\n", argv[0]);
            return 0;
    }

	/* get some environment variables */
	list<string> paths;
	char * env = getenv("ERLANG_LIB");

	if (env)
	{
		string fields[MAX_PATHS], sep = ":";
		string Senv = getenv("ERLANG_LIB");
		int i, n;

		n = split(Senv, fields, MAX_PATHS, sep);
		
		for (i = 0; i < n; i++)
		{
			paths.push_back(fields[i]);
		}
	}

    /* initialise lexer */
    start_new_file(argv[wf], f);

	yy_program = new Program;
	yy_program->set_paths(paths);
	yy_program->initialise();

    // PARSE the program
    yyparse();

	if (!yy_program) 
	{
		fprintf(stderr, "Bogus abstract syntax tree\n");
		return 0;
	}

	if (yy_program->errors > 0)
	{
		fprintf(stderr, "Compilation failed\n");
		return 0;
	}


	// Walk the AST and emit the code
    TableStack.push_front(&yy_program->Table);
    CurrentTable = TableStack.front();

	switch (mode)
	{
		case Mstatic:
        {
			yy_program->ast->typecheck(LT_all);
			printf("%s\n", yy_program->EmitHead().c_str());
			printf("%s\n", yy_program->local->expr_code().c_str());
			printf("%s\n", yy_program->ast->expr_code().c_str());
			printf("%s\n", yy_program->EmitStatic(statix).c_str());
			break;
        }
		case Mnormal:
		case Mcompile:
		case Mshared:
		default:
			yy_program->ast->typecheck(LT_all);
			printf("%s\n", yy_program->EmitHead().c_str());
			printf("%s\n", yy_program->local->expr_code().c_str());
			printf("%s\n", yy_program->ast->expr_code().c_str());
			// yy_program->ast->expr_code().c_str();
			break;
		case Mprint:
			printf("%s\n", yy_program->ast->print().c_str());
			break;
	}

    // END
}


