
//
// $Id: match.cc,v 1.31 2002/08/03 14:31:42 geoff Exp $
//
// $Source: /home/cvs/gerl/src/match.cc,v $
// $Revision: 1.31 $
// $Date: 2002/08/03 14:31:42 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "match.h"
#include "symbol.h"
#include "str.h"
#include "program.h"
#include "typestr.h"

// #define DEBUG 1
extern int debug;

int Match::_labelcount = 0;

Match::Match(Node * l, Node * r)
{
	left = l;
	right = r;

    _label = "cM" + itoa(_labelcount);
    _labelcount++;
}

list<string> Match::emit_code()
{
    Symbol * s;
	string instr, pre, RHS;
    int count = 1, max;
    list<string> pc, lret;
    bool flat;

    pc = right->pre_code();
    add_pre_code(pc);
    RHS = right->expr_code();

    if (depth == 0 || (_parent && _parent->flat_code())) flat = true;
    else flat = false;

	// assign RHS to a local variable
	instr = pre + "\t" + _label + " = " + RHS;

#if 1
    // FIX: this is buggy x = a,b,c; results in x == c 
    // instead of x == a as it should here...
    if (flat) instr = instr + ";\n\t";
    else instr = instr + ", ";
#else
     instr = instr + ";\n\t";
#endif

    // Emit typecheck on returned value (for whatever LHS expects)
    Etype x = (Etype) left->type();
    string xx;
    
    // FIX: hack? problem with type() returns?
    if (x != LT_none)
        xx = Typestr::etype2c(left->type());
    else
        x = LT_anyvar;

    if (x != LT_anyvar && xx != "LT_anyvar" && xx != "LT_none")
    {
        instr = instr + "TCM(" + _label + "," + xx + ", __FILE__, __LINE__)";
#if 1
        // FIX: this is buggy x = a,b,c; results in x == c 
        // instead of x == a as it should here...
        if (flat) instr = instr + ";\n\t";
        else instr = instr + ", ";
#else
        instr = instr + ";\n\t";
#endif
    }

    // build the pattern into the "local" symboltable
    // the result is add (automatically) into the local
    // symbol table for later emission
    TableStack.push_front(&Table);
    CurrentTable = TableStack.front();
    left->emit_pattern(_label);	

    // set up a 'pointer' to the relevant freesym..
    // left->freesym(right->freesym());
    TableStack.pop_front();
    CurrentTable = TableStack.front();

    // emit the local symbol table as instructions..
    Dictionary <Symbol *>::iterator j;                                    
    max = Table.Table.num_elements();
    count = 1;
    for (j = Table.Table.begin(); j != Table.Table.end(); j++)
    {
        Symbol * up;
        string sim;
        if (debug)
        {
            fprintf(stderr, "Match::(%s:%s)\n", 
                (*j)->symbol().c_str(), (*j)->parameter().c_str());
            fprintf(stderr, "Match depth = %d\n", depth);
        }
        sim = (*j)->symbol();
        up = Table.chain()->find(sim, S_Variable);
        // now emit a parameter match if appropriate
        if (up == NULL) up = (*j);      // shouldn't be like this?
        if (up->parameter().length() != 0)
        {
            // up->type((*j)->type() & ~LT_on_heap);
            up->type(LT_normal);
            instr = instr + up->symbol() + " = " + up->parameter(); 
            if (flat) 
            {
                instr = instr + ";\n\t";
            }
            else 
            {
                if (count < max) instr = instr + ",";
            }
        }
        count++;
    }

    if (!flat && pc.empty()) 
            instr = "(" + instr + ")";
    // FIX: above - use lret
    
	return string2list(instr);
}

string Match::print()
{
	return left->print() + " = " + right->print();
}


string Match::label()
{
    return _label;
}

Symbol * Match::freesym()
{
    return right->freesym();    
}

void Match::freesym(Symbol * s)
{
    return right->freesym(s);    
}

void Match::add_symbol(SymbolTable * s, int df)
{
    if (debug)
        fprintf(stderr, "Match::add_symbol (%d)\n", df);

    left->add_symbol(&Table, df);
    // left->add_symbol(s, df);      // should do it because its chained?
}


#if 0
bool Match::typecheck(Etype check, Etype restrict)
{
    
}
#endif

