

//
// $Id: node.cc,v 1.48 2002/01/21 12:22:23 geoff Exp $
//
// $Source: /home/cvs/gerl/src/node.cc,v $
// $Revision: 1.48 $
// $Date: 2002/01/21 12:22:23 $
// $State: Exp $
//
// Geoff Wong (C) Copyright 1995,1996,1997,1998.
//	This file is distributed under the terms of the 
//	GNU Public License, details of which can be found
//	in the file "Copying" which accompanies this code.
//

#include "nodes.h"
#include "str.h"

extern int debug;

// Node tab initialisation
int Node::tablevel = 0;

Node::Node()
{
    depth = 0;
    _line = 0;
    _type = LT_none;
    _assign = NULL;
    _freesym = NULL;
    _stmt = ST_expr;
}

Node::~Node()
{
}

string Node::tabs()
{
	string ret;
	string tab = "\t";
	for (int i = 0; i < tablevel; i++)
	{
		ret = ret + tab;	
	}
	return ret;
}

// a do nothing emit_pattern
string Node::emit_pattern(string p)
{
	string ret;
	return ret;
}

// PURPOSE: return emit_code() list if a we have an assignable
list<string> Node::pre_code()
{
    list<string> empty;

    if (_assign != NULL)
    {
        return emit_code();
    }

    return _pre_code;
}

// Purpose: add pre_code to our pre_code list
void Node::add_pre_code(list<string> l)
{
    list<string>::iterator end;

    end = _pre_code.end();
    _pre_code.insert(end, l.begin(), l.end());
    //printf("add_pre_code:%s\n", join(l, "\n").c_str());
}

string Node::expr_code()
{
    list<string>::iterator i;
    list<string> code;
    string ret;

    if (_assign == NULL)
    {
        code = emit_code();

        for (i = code.begin(); i != code.end(); i++)
        {
            ret = ret + (*i);
        }
        return ret;
    }
    else return _assign->symbol();
}

// a do nothing emit_guard
string Node::emit_guard(string p)
{
	string ret;
	return ret;
}

void Node::set_depth(int lvl)
{
    depth = lvl;
}

void Node::add_symbol(SymbolTable * st, int df)
{
    // default is to do nothing
    return;
}

void Node::line(int l)
{
    _line = l;
}

int Node::line()
{
    return _line;
}

void Node::type(int t)
{
    if (debug) fprintf(stderr, "Node::type(%d)\n", t);
    _type = t;
}

int Node::type()
{
    return _type;
}

// PURPOSE: resolve my type with given type?
bool Node::typecheck(Etype t2, Etype restrict)
{
#if 0
    assert(_type != LT_none);
    assert(t2 != LT_none);

	if (t1->flag != CLT_unknown && t2->flag != CT_unknown && 
			t1->flag != t2->flag) {
		yyerrorf("Types (totally) incompatible in %s\n", where);
		return 0;
	}
#endif
    if (_type & t2) 
    {
        return true;
    }
#if 0
	switch (_type) 
    {
		case CT_basic:
			if ((t1->type.basic & t2->type.basic) && 
				(t1->type.basic & restrict)) return 1;
			break;
		case CT_class:
			if (t1->type.class == t2->type.class) return 1;
			break;
		case CT_unknown:
			if (t2->flag == CT_unknown) {
				t1->unknown = t2->unknown;
				add_resolve_list(t1->resolve, t2->resolve);
				t2->resolve = t1->resolve;
				return 1;
			}
			t1->flag = t2->flag;
			t1->type = t2->type;
			t1->depth = t2->depth;
			t1->unknown = 0;
			resolve_resolve_list(t1);
			t1->resolve = 0;
			t2->resolve = 0;
			return 1;
		default:
			fatal("Illegal internal compiler type encountered.\n");
			return 0;
	}
	yyerrorf("Types don't match in %s\n", where);
#endif
    fprintf(stderr, "type error at (%d)\n", _line);
	return false;
}

string Node::emit_typecheck(string str, int restrict, int param, string where)
{
    if (_type & restrict) 
    {
        return "";
    }
    else
    {
        return "TC("+str+","+itoa(restrict)+","+itoa(param)+","+
                "\"" + where + "\"),";
    }
}

void Node::parent(Node * n)
{
    _parent = n;
}

Node * Node::parent()
{
    return _parent;
}

void Node::assign(Symbol * x)
{
    _assign = x;
}

Symbol * Node::assign()
{
    return _assign;
}

void Node::freesym(Symbol * x)
{
    _freesym = x;
}

Symbol * Node::freesym()
{
    return _freesym;
}

int Node::stmt_type()
{
    return _stmt;
}

void Node::stmt_type(int x)
{
    _stmt = x;
}

void Node::extra(const string& str)
{
    _extra = str;
}

string Node::extra()
{
    // fprintf(stderr, "extra=%s\n", _extra.c_str());
    return _extra;
}

// NodeList Class
NodeList::NodeList()
{
}

NodeList::NodeList(Node * n)
{
	nodes.push_front(n);
}

NodeList::NodeList(Node * n, Node *p)
{
	nodes.push_front(p);
	nodes.push_front(n);
}

#if 0
NodeList::NodeList(Node * n, NodeList * tail)
{
}
#endif

list<string> NodeList::emit_code() 
{
	list<Node *>::iterator i;
	list<string> ret, code;

	for (i = nodes.begin(); i != nodes.end(); i++)
	{
        code=(*i)->emit_code();
        add_pre_code((*i)->pre_code());
		ret.insert(ret.end(), code.begin(), code.end());
        if (i != --(nodes.end()))
        {
            ret.push_back(_delimiter);
        }
	}
	return ret;
}

string NodeList::emit_pattern(string p)
{
	list<Node *>::iterator i;
	string ret;

	if (nodes.size() == 0)
	{
		ret = "1";
		return ret;
	}
	
	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		ret = ret + (*i)->emit_pattern(p);
	}
	return ret;
}

string NodeList::emit_guard(string p)
{
	list<Node *>::iterator i;
	string ret;

	if (nodes.size() == 0)
	{
		ret = "1";
		return ret;
	}
	
	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		ret = ret + (*i)->emit_guard(p);
	}
	return ret;
}

string NodeList::print()
{
	string ret;
	list<Node *>::iterator i;

	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		ret = ret + (*i)->print();
	}
	return ret;
}

void NodeList::add_symbol(SymbolTable * st, int df)
{
	list<Node *>::iterator i;

	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		(*i)->add_symbol(st, df);
	}
}

void NodeList::prepend(Node * n)
{
	nodes.push_front(n);
}

void NodeList::append(Node * n)
{
	nodes.push_back(n);
}

int NodeList::size()
{
    return nodes.size();
}

void NodeList::set_depth(int lvl)
{
    list<Node *>::iterator i;
    depth = lvl;
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        (*i)->set_depth(lvl);
    }
}

void NodeList::delimiter(const string& delim)
{
    _delimiter = delim;
}

// set the type of all elements in the list..
void NodeList::type_all(int t)
{
    list<Node *>::iterator i;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        (*i)->type(t);
    }
}

Symbol *  NodeList::assign()
{
    return _assign;
}

void NodeList::assign(Symbol * t)
{
#if 1
    Node::assign(t);
#else
    list<Node *>::iterator i;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        (*i)->assign(t);
    }
#endif
}


// return the or'ed set of all elements in the list..                 
int NodeList::type_all()
{
    list<Node *>::iterator i;
    int ret = LT_none;

    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        ret |= (*i)->type();
    }
    return ret;
}


Bracket::Bracket(Node * n)
{
	value = n;
}

list<string> Bracket::emit_code()
{
    add_pre_code(value->pre_code());
	return string2list("(" + value->expr_code() + ")");
}

string Bracket::print()
{
	return "(" + value->print() + ")";
}


