

//
// $Id: pattern.cc,v 1.58 2002/08/04 14:45:28 geoff Exp $
//
// $Source: /home/cvs/gerl/src/pattern.cc,v $
// $Revision: 1.58 $
// $Date: 2002/08/04 14:45:28 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "pattern.h"
#include "str.h"
#include "symbol.h"
#include "program.h"
#include "error.h"
#include "erlib.h"

// #define DEBUG 1
extern int debug;

PatternList::PatternList()
{
}

PatternList::PatternList(NodeList * n)
{
	n = n;
	//	nodes.add(n)
}

PatternList::PatternList(Node * n)
{
	prepend(n);
}

string PatternList::emit_pattern(string p)
{
	string ret;
	string pee = "p";
	list<Node *>::iterator i;
	int count = 0;

	// stop compiler complaining
	p = p;

	if (nodes.size() == 0)
	{
		ret = "(1)";
		return ret;
	}

	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		// Somehow we need to check for "identicals" in here..
		//  perhaps each node needs a named annotation?
		//  used in conjunction with a symbol table lookup / insertion
		// Also need to assign variable names to these parameters

        error.argument(count);
		ret = ret + ((Node *)(*i))->emit_pattern(pee + itoa(count++));
		if (i != --(nodes.end())) ret = ret + " && ";
	}

	return ret;
}

list<string> PatternList::emit_code()
{
	string ret;

#if 0
	list<Node *>::iterator i;
	for (i = nodes.begin(); i != nodes.end(); i++)
	{
		ret = ret + (*i)->emit_code();
	}
#endif

	return string2list(ret);
}

string PatternList::print()
{
    string ret;
    list<Node *>::iterator i;

	// printf("PatternList::print()\n");
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
		if (i != nodes.begin()) 
		{
			ret = ret + ",";
		}
        ret = ret + (*i)->print();
    }
    return ret;
}

PatternTail::PatternTail()
{
	tail = 0;
}

#if 0
PatternTail::PatternTail(NodeList * n)
{
}
#endif

PatternTail::PatternTail(Node * n)
{
	tail = n;
}

list<string> PatternTail::emit_code()
{
	list<string> ret;
	return ret;
}

string PatternTail::emit_pattern(string p)
{
	string ret, nxt, res;
	list<Node *>::iterator i;
	int count = 0;

	if (nodes.size() == 0)
	{
		ret = ret + "(Eis_empty_list(" + p + "))";
	}
	else
	{
		ret = ret + "(Eis_full_list(" + p + "))";
        if (tail == NULL) 
        {
            ret = ret + " && (Elist_size(" + p + ","
                + itoa(nodes.size()) + "))";
        }
        else
        {
            if (nodes.size() > 1)
                ret = ret + " && (Elist_gsize(" + p + ","
                + itoa(nodes.size()) + "))";
        }
		for (i = nodes.begin(); i != nodes.end(); i++)
		{
#ifdef _CONS_H
            res = (*i)->emit_pattern("("+p+".u.list" +nxt+"->e)");	
            if (res != "(1)") ret = ret + " && " + res;
            count++;
            nxt = nxt + "->next";
#else
			ret = ret + " && " + (*i)->emit_pattern("Eindex_list(" + p + "," +
					itoa(count++) + ")");	
#endif
		}
	}
	if (tail)
	{
#ifdef _CONS_H
		ret = ret+" && (";
        if (count == 1)
        {
            ret = ret + tail->emit_pattern("(Etail_list(__vvv,"
            +p+"," + itoa(count) + ",\""+error.function()+"\"))") + ")";
        }
        else
        {
            ret = ret + tail->emit_pattern("(Entail_list(__vvv,"
            + p + ".u.list" + nxt + "," + itoa(count+1) + 
            ",\""+error.function()+"\"))") + ")";
        }
#else
#ifdef _ALLOCS_H
		ret = ret+" && (" + 
            tail->emit_pattern("(Etail_list(__vvv,"
            +p+"," + itoa(error.argument()) + ",\""+error.function()+"\"))") + ")";
#else
        // LIST HACK:
        tail->extra("destructive");
		ret = ret+" && (" + 
            tail->emit_pattern("(Etail_listd("
            +p+"," + itoa(error.argument()) + ",\""+error.function()+"\"))") + ")";
#endif

#endif
	}
	return ret;
}

string PatternTail::print()
{
    string ret;
    list<Node *>::iterator i;

	ret = "[";
    for (i = nodes.begin(); i != nodes.end(); i++)
    {
        ret = ret + (*i)->print();
    }
	ret = ret + "]";
    return ret;
}

void PatternTail::add_symbol(SymbolTable * st, int df)
{
#ifdef DEBUG
    fprintf(stderr, "PatternTail::add_symbol\n");
#endif
    if (st == NULL) return;
    NodeList::add_symbol(st, df);
    if (tail != NULL)
    {
        if (debug) fprintf(stderr, "adding tail to symbol table (%d)\n",df);
        // tail->add_symbol(st, df);
        tail->add_symbol(st, LT_list | LT_on_heap | df);
    }
}

PatternTuple::PatternTuple(const string& tmp)
{
    storage = tmp;
}

PatternTuple::PatternTuple(const string& tmp, NodeList * n)
{
    storage = tmp;
	value = n;
}

string PatternTuple::emit_pattern(string p)
{
	list<Node *>::iterator i;
    // int count = value->size() - 1;
    int count = 0;

    if (debug)
    {
        fprintf(stderr, "PT:emit_pattern %s (type=%d)\n", p.c_str(),
            type());
    }

    if (value == NULL)
    {
        return "Eis_tuple(0,"+p+")";
    }

    // Even though we don't add these patterns into the string
    // we need to emit_pattern() on all the elements of the tuple
    // so they can build up the initial assignment
	string ret("(Eis_tuple(");
	ret = ret + itoa(value->size()) + "," + p + "))";

	for (i = value->nodes.begin(); i != value->nodes.end(); i++)
	{
        string tmp;
        if (type() & LT_returned)
            tmp = "Ecindex_tuple(" + p + "," + itoa(count++) + ")";
        else
            tmp = "Eindex_tuple(" + p + "," + itoa(count++) + ")";
        tmp = (*i)->emit_pattern(tmp);
        if (tmp != "(1)") ret = ret + " && " + tmp; 
    }

	return ret;
}

list<string> PatternTuple::emit_code()
{
	string ret;
	list<Node *>::iterator i;

    if (value == NULL)
    {
#ifdef _ALLOCS_H
        return string2list("(tuplealloc(" + storage 
            + ",0),Emake_tuple(" + storage + "))");
#else
        return string2list("Emake_tuple(" + storage + ")");
#endif
    }
    if (value->nodes.size())
    {
#ifdef _ALLOCS_H
        ret = ret + "(tuplealloc(" + storage + ", " 
                    + itoa(value->nodes.size()) + "),";
#endif
    }
    else
    {
#ifdef _ALLOCS_H
        return string2list("(tuplealloc(" + 
            storage + ",0),Emake_tuple(" + storage + "))");
#else
        return string2list("Emake_tuple(" + storage + ")");
#endif
    }
    ret = ret + "Emake_tuple(" + storage;

    if (value->nodes.size()) ret = ret + ",";
	for (i = value->nodes.begin(); i != value->nodes.end(); i++)
	{
        add_pre_code((*i)->pre_code());
      	if (i != value->nodes.begin())
        {
            // shouldn't do this if it's "tail" item..
            // where should this be FIXED?
            ret = ret + ",";
        }
		ret = ret + (*i)->expr_code();
	}
#ifdef _ALLOCS_H
	ret = ret + ")";
#endif
	ret = ret + "))";

	return string2list(ret);
}

string PatternTuple::print()
{
    if (value == NULL) 
    {
        return "{ }";
    }
	return "{ " + value->print() + " }";
}

void PatternTuple::add_symbol(SymbolTable * st, int df)
{
    if (st == NULL) return;
    if (value == NULL) return;

    value->add_symbol(st, df);
}                           

#if 0
// ALL BELOW IS if'ed out 
// basic types work fine?
PatternAtom::PatternAtom(const string * at)
{
	value = *at;
}

list<string> PatternAtom::emit_code()
{
	// CHANGED HERE
	return "EValatom(\""+value+"\")";
}

string PatternAtom::emit_pattern(string p)
{
	string ret;
	
	if (value == "_")
	{
        // this is just 'ignore variable'
		ret = "(1)";
	}
	else
	{
		ret = "Eequalvs(" + p + ", \"" + value + "\")";
	}
	return ret;
}

string PatternAtom::print()
{
	return value;
}
	

PatternNumber::PatternNumber(int num)
{
	value = num;
}

list<string> PatternNumber::emit_code()
{
	return itoa(value);
}

string PatternNumber::emit_pattern(string p)
{
	string ret;
	
	ret = ret + "(Eequalvi(" + p + ", " + itoa(value) + "))";
	return ret;
}

string PatternNumber::print()
{
	return itoa(value);
}


Patternstring::Patternstring(const string * str)
{
	value = *str;
}

list<string> Patternstring::emit_code()
{
	return value;
}

string Patternstring::emit_pattern(string p)
{
	string ret;
	
	ret = ret + "(Estrcmp(" + p + "," + value + ") == 0)";
	return ret;
}

string Patternstring::print()
{
	return value;
}


PatternVariable::PatternVariable()
{
}


PatternVariable::PatternVariable(const string * str)
{
	value = *str;
}


list<string> PatternVariable::emit_code()
{
	return value;
}


string PatternVariable::emit_pattern(string p)
{
	string ret = "(1)";
	SymbolTable * st;
	Symbol * sym;

	st = CurrentTable;

#if 1
	fprintf(stderr, "emitparameter: %s = %s\n", value.chars(), p.chars());
#endif

	if ((sym = st->find(value, S_Variable)))
	{
		if (sym->symbol() == value && sym->parameter() != "")
		{
			ret = "(Eequal(" + sym->parameter() + ", " + p + "))";
		}
		else
		{
#if  1
     fprintf(stderr,"PatternVariable::emit_pattern: %s = %s\n", 
            sym->symbol().chars(), p.chars());
#endif
			sym->setparameter(p);
		}
	}
#if 1
    else
    {
        fprintf(stderr,"Can't find: %s\n", value.chars());
    }
#endif
	return ret;
}


string PatternVariable::print()
{
	return value;
}


PatternBoolean::PatternBoolean()
{
}

PatternBoolean::PatternBoolean(int x)
{
	value = x;
}

list<string> PatternBoolean::emit_code()
{
	// return "EValbool("+itoa(value)+")";
	return string2list(itoa(value));
}

string PatternBoolean::emit_pattern(string p)
{
	return itoa(value);
}

string PatternBoolean::print()	
{
	string ret;
	return ret;
}
#endif
