
#ifndef _PATTERNLIST_H
#define _PATTERNLIST_H

#include "node.h"
#include "symbol.h"

class PatternList : public NodeList
{
public:
	PatternList();
	PatternList(NodeList *);
	PatternList(Node *);

	string emit_pattern(string p);
		// PURPOSE: emit pattern matching code for parameter headings
	list<string> emit_code();
		// PURPOSE: emit code for patterns
	string print();
};

class PatternTail : public NodeList
{
	Node * tail;
public:

	PatternTail();
//	PatternTail(NodeList *);
	PatternTail(Node *);
	string print();	
	list<string> emit_code();
	string emit_pattern(string p);
    void add_symbol(SymbolTable *, int df = LT_none);
};

class PatternTuple : public Node
{
    string storage;
	NodeList * value;

public:
	PatternTuple(const string& tmp);
	PatternTuple(const string& tmp, NodeList *);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();	
    void add_symbol(SymbolTable * st, int df = LT_none);
};


class PatternAtom : public Node
{
	string value;
public:
	PatternAtom(const string * at);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();
	
};

class PatternNumber : public Node
{
	int value;
public:
	PatternNumber(int num);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();
};

// Estring - Erlang string (to distinguish this from a C++ string)
class Patternstring : public Node
{
	string value;
public:
	Patternstring(const string * str);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();
};

class PatternVariable : public Node
{
	string value;
public:
	PatternVariable();
	PatternVariable(const string * str);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();
};


class PatternBoolean : public Node
{
	bool value;
public:
	PatternBoolean();
	PatternBoolean(int);
	list<string> emit_code();
	string emit_pattern(string p);
	string print();	
};


#endif
