//
// $Id: program.cc,v 1.23 2002/08/05 08:03:19 geoff Exp $
//
// $Source: /home/cvs/gerl/src/program.cc,v $
// $Revision: 1.23 $
// $Date: 2002/08/05 08:03:19 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "program.h"
#include "str.h"
#include "function.h"
#include "form.h"

extern int debug;

// #include "erlib.h"

// #define DEBUG 1

// UGLY global - needs to be put in Program or something

int Program::val_count = 0;

Program::Program()
{
    errors = 0;
    imports = NULL;
    exports = NULL;
    local = new FormList();
    local->set_program(this);
}

Program::~Program()
{
}

void Program::initialise()
{
    // import the standard modules
    open_module("liberl");
    import_module("bif");
    import_module("erlang");
}

void Program::set_module(string str)
{
    _module = str;
}

string Program::Module()
{
    return _module;
}

// set the current clause we're working on
void Program::set_clause(SymbolTable * c)
{
    _clause = c;
}

// return the current clause
SymbolTable * Program::Clause()
{
    return _clause;
}

void Program::set_paths(list<string>& pt)
{
    import_paths = pt;
}

void * Program::open_module(const char * module)
{
    void * handle;
    string smodule;

    // by default look locally
    smodule = smodule + module + ".so";

#if 0
    string testmod;
    list<string>::iterator i;
    struct stat fbuf;

    for (i = import_paths.begin(); i != import_paths.end(); i++)
    {
#ifdef DEBUG
    fprintf(stderr,"Checking %s for %s\n", (*i).c_str(), module);
#endif
        testmod = (*i) + "/" + module + ".so";
        if (stat(testmod.c_str(), &fbuf) == 0)
        {
            smodule = testmod;
        }
    }
#endif

    handle = dlopen (smodule.c_str(), RTLD_NOW | RTLD_GLOBAL);
    if (!handle)
    {
        fprintf(stderr, "Loading module (%s) failed.\n", smodule.c_str());
        return 0;
    }
    return handle;
}


void Program::add_imports(Attribute * attr)
{
#if 0
    // FIX: should add to the list rather than replace it..
    imports = attr;
    import_module(attr->name.c_str());
#else
    list<Node *>::iterator i;
    string module = attr->module();

    // go through the farity list and add each one to the symbol table
    for (i = attr->vlist->nodes.begin(); i != attr->vlist->nodes.end(); i++)
    {
        // FIX: need to use a farity list if it's available..
        Farity * farity = (Farity *)(*i);
        string imp = farity->name();
        int params = farity->params();
        string typesig;

        if (Table.find(imp, S_Function) == 0)
        {
            Symbol * s;
#ifdef DEBUG
            fprintf(stderr, "adding (%s <-> %s, %d)\n", module.c_str(), 
                            imp.c_str(), params);
#endif
            s = new Symbol(imp, S_Function, typesig, params);
            s->module(module);
            Table.add(s);                             
        }
    }
#endif
}

void Program::add_record(string at, NodeList * pat)
{
    RecordTable.add_element(at.c_str(), pat);
}

NodeList * Program::find_record(const string &r)
{
    NodeList * nl;

    nl = RecordTable[r.c_str()];

    return nl;
}


int Program::find_record(const string &r, const string &f)
{
    NodeList * nl;
    list<Node *>::iterator ni;

    int x = 0;

    nl = RecordTable[r.c_str()];
    if (!nl) return -1;
    if (debug)
        fprintf(stderr, "found record(%s)\n", r.c_str());

    // else find the field in the list ...
    for (ni = nl->nodes.begin(); ni != nl->nodes.end(); ni++)
    {
        string tmp;
    
        tmp = (*(ni))->print();
        if (debug)
            fprintf(stderr, "record field(%s)\n", tmp.c_str());
        if (tmp == f) return x;
        x++;
    }

    return -1;
}

// FIX: this actually queries a modules exports for 
//  the imports and doesn't use a specifically supplied import list
int Program::import_module(const char * module)
{
    _Symbol * (*func)();
    void * handle;
    string Smodule(module);
    _Symbol * imports;
    const char * error;

    handle = open_module(module);

    func = (_Symbol * (*)())(dlsym(handle, (Smodule + EXPORTS).c_str()));
    if ((error = dlerror()) != NULL)
    {
        fprintf(stderr, "Unable to find %s__exports().\n", module);
        fprintf(stderr, "%s\n", error);
        return 0;
    }

    imports = func();

    assert(imports != NULL);

    for (int i = 0; imports[i].module != NULL; i++)
    {
        string imp = imports[i].module;
        int params = imports[i].parameters;
        string typesig;

        if (Table.find(imp, S_Function) == 0)
        {
            Symbol * s;
#ifdef DEBUG
            fprintf(stderr, "adding (%s <-> %s/%d)\n", module,
                            imp.c_str(), params);
#endif
            s = new Symbol(imp, S_Function, typesig, params);
            s->module(module);
            Table.add(s);                             
        }
        else
        {
            // we have a problem (name clash)
            fprintf(stderr, "name clash problem (%s)\n", (imp.c_str()));

        }
    }
    return 1;
}

void Program::add_exports(Attribute * attr)
{
    // FIX: should add to the list rather than replace it..
    exports = attr;
}


string Program::EmitHead()
{
    string ret;
    Dictionary<Symbol *>::iterator i;
    Symbol * sym;
    string modname;
    string fname;

    ret = "#include \"erlib.h\"\n\n";

    for (i = Table.Table.begin(); i != Table.Table.end(); i++)
    {
        sym = (*i);
        if (sym == NULL) continue;

        fname = sym->symbol();

//fprintf(stderr, "EH: %s\n", sym->symbol().c_str());
        if (sym->symboltype == S_Function)
        {
            modname = sym->module();
            if (modname.length() == 0)
            {
                // it's a local function
                modname = Module();
            }
            else
            {
                // or it's a remote one and it's not used
                if (sym->used() == 0) continue;
            }
            ret = ret + "Val " + modname + "_" + fname + "(";
            for (int x = 0; x < sym->params(); x++)
            {
                if (x != 0) 
                {
                    ret = ret + ",";
                }
                ret = ret + "Val p" + itoa(x);
            }
            ret = ret + ");\n";
        }
    }

    // we should also output the __exports() function here
    // FIX: should really come from the exports() list not the function list!
    ret = ret + "_Symbol * " + Module() + 
            EXPORTS + "()\n{\n\tstatic _Symbol tmp[] = {";
    for (i = Table.Table.begin(); i != Table.Table.end(); i++)
    {
        string mod;
        int params;
        sym = (*i);

        if (sym == NULL) continue;
        fname = sym->symbol();
        mod = sym->module();
        params = sym->params();

        if (sym->symboltype == S_Function && mod.length() == 0)
        {
            ret = ret + "{\"" + fname + "\"," + itoa(params) + "},\n\t";
        }
    }

    ret = ret + "{ NULL, 0 }  };\n\treturn tmp;\n}\n";

    return ret;
}


string Program::EmitStatic(string& name)
{
    string module;
    string fun, funstr;
    string ret;
    int params;
    int i, p;

    module = Module();
    p = name.find('/');
    if (p != -1)
    {
        params = atoi(name.substr(p+1).c_str());
    }
    else
    {
        p = name.length();
    }
    fun = name.substr(0,p);

    funstr = Namer::mangle(module,fun,params);

    ret = "main(int argc, char * argv[])\n{\n";
    ret = ret + "\tVal _ret;\n";
    ret = ret + "\tEinitserver(0,0);\n";
    ret = ret + "\t_ret=" + funstr + "(";
    for (i = 0; i < params; i++)
    {
        if (i != 0) ret = ret + ",";
        ret = ret + "EParam(argv["+itoa(i)+"])";
    }
    ret = ret + ");\n\t";
    ret = ret + "EPrintvalue(stdout,_ret);\n";
    ret = ret + "}\n";
    return ret;
}


