
//
// $Id: program.h,v 1.15 2004/09/05 02:42:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/program.h,v $
// $Revision: 1.15 $
// $Date: 2004/09/05 02:42:56 $
// $State: Exp $
//
// Author: Geoff Wong
//


#ifndef _PROGRAM_H
#define _PROGRAM_H

#include <stdio.h>
#include <dlfcn.h>
#include <sys/stat.h>
#include <list>
#include "node.h"
#include "symbol.h"
#include "dictionary.h"
#include "farity.h"
#include "erlib.h"
#include "namer.h"
#include "expr.h"

//
// Name:
//     Program 
// 
// Purpose:
//    Provide a "global" name space for parsing
//  Maintains:
//        the module name
//        module symbol table
//        symbol table for each function within the module
//
// Notes:
//      should this be a singleton?
//
//

class FormList;

class Program 
{
    SymbolTable * _clause;

    Dictionary<NodeList *> RecordTable;

    string _module;
        // module name

    list<string> import_paths;
        // list of paths to look for imported modules

    Attribute * imports;

    Attribute * exports;

    void * open_module(const char *);
        // open a dynamically loadable module and return the handle

    int import_module(const char *);
        // actually import a module; 0 - failed

public:
    static int val_count;
        // a temporary variable counter for clauses (ugly)

    int errors;
        // number of errors encountered in parsing

    FormList * local;
        // pointer to the root of local (anonymous) functions
        
    Node * ast;
        // pointer to the root of the abstract syntax tree

    SymbolTable Table;

    Program();
    ~Program();

    void initialise();
        // start up a program by importing standard modules (etc)
    
    string Module();

    void set_module(string str);

    void set_clause(SymbolTable * c);
        // set the current clause we're working on

    void set_paths(list<string>& pt);

    void add_imports(Attribute * attr);
        // add imported functions (entire modules) into the
        // symbol table (checking for clashes).

    void add_exports(Attribute * attr);
        // all symbols to the list of symbols to export 
        // these will be checked against existing functions on
        // output in EmitHead()

    void add_record(string a, NodeList * r);
        // pre-processor (in compiler) support

    int find_record(const string& r, const string& f);
        // find a r.f match
        // Returns: -1 on failure otherwise the what?
        
    NodeList * find_record(const string& r);
        // Returns: the actual record structure (or NULL)

    SymbolTable * Clause();
        // Returns: the current clause

    string EmitHead();
        // Purpose: 
        //    Emit the "head" of the C file for this module
        //     Also emit the Module()__exports() function

    string EmitStatic(string& m);
        // Purpose: 
        //    Emit the static "main()" function so we can statically
        //  link this module (should this be in a separate file)? 

};
 
extern Program * yy_program;
extern SymbolTable  * CurrentTable;
extern list<SymbolTable *> TableStack;      
// ugly global..

#define EXPORTS "__exports"


#endif
