
//
// $Id: receive.cc,v 1.20 2002/02/02 02:31:33 geoff Exp $
//
// $Source: /home/cvs/gerl/src/receive.cc,v $
// $Revision: 1.20 $
// $Date: 2002/02/02 02:31:33 $
// $State: Exp $
//
// Copyright (C) Geoff Wong (1996, 1997, 1998)
// geoff@serc.rmit.edu.au
//
// This file is distributed under the terms of the
// GNU Public License, details of which can be found
// in the file "Copying" which accompanies this code.        
//

#include "receive.h"
#include "error.h"

// int Receive::_labelcount = 0;

Receive::Receive(CaseList * l, Node * e, Node * b) : Case(b, l)
{
	clauses = l;
	after = e;
	body = b;

    _label = "cR"+itoa(_labelcount);
    _labelcount++;
    stmt_type(ST_stmt);
}

list<string> Receive::emit_code()
{
    string ret;

    ret = _label + " = Ereceive(";
    if (after)
    {
        add_pre_code(after->pre_code());
        ret = ret + after->expr_code();
    }
    else
    {
        ret = ret + "zero";
    }
    ret = ret + ");\n\t";

    ret = ret + "if (" + _label + ".tag != LT_none)\n\t{\n\t";
    if (clauses != NULL)
    {
        add_pre_code(clauses->pre_code());
        ret = ret + clauses->emit_code(_label);
    }
    else
    {
        ret = ret + error.emit_exit("receive_clause") + ";";
    }
    ret = ret + "\n\t}\n\t";

    // after bit..
    if (body != NULL)
    {
        add_pre_code(body->pre_code());
        ret = ret + "else\n\t{\n" + body->expr_code() + "\t}\n";
    }
#if 0
    else
    {
        ret = ret + "else\n\t{\n\t" + error.emit_exit("receive_clause") + ";\n\t}\n"; 

    }
#endif

    return string2list(ret);
}

string Receive::print()
{
	return "receive " + clauses->print() + " after " + after->print() 
		+ body->print() + "end";
}

string Receive::label()
{
    return _label;
}

void Receive::set_depth(int lvl)
{
    depth = lvl;
    if (clauses != NULL) clauses->set_depth(lvl);
    if (after != NULL) after->set_depth(lvl);
    if (body != NULL) body->set_depth(lvl);
}

Send::Send(Node * l, Node * r)
{
	left = l;
	right = r;
}

list<string> Send::emit_code()
{
    list<string> ret;

    add_pre_code(left->pre_code());
    add_pre_code(right->pre_code());

    ret.push_back("Esend(" + left->expr_code() + "," + right->expr_code() + ")");
	return ret;
}

string Send::print()
{
	return left->print() + " ! " + right->print();
}

void Send::set_depth(int lvl)
{
    left->set_depth(1);
    right->set_depth(1);
}
