
//
//	Erlang Compiler with Monitoring Support
//
//	Geoff Wong (C) Copyright 1996,1997,1998.
//	C/O RMIT University
//	Swanston St. Melbourne 3001.
//	Australia
//
//	email: geoff@serc.rmit.edu.au
//
//	This file is distributed under the terms of the 
//	GNU Public License, details of which can be found
//	in the file "Copying" which accompanies this code.
//

#include "node.h"
#include "program.h"
#include "str.h"
#include "record.h"
#include "yy.h"

Record::Record(Node * v, string r, string f)
{
    NodeList * nl;

    var = v;
    _index = yy_program->find_record(r,f);
}

list<string> Record::emit_code()
{
    list<string> ret;

    ret = var->emit_code();
    ret.push_front("Eindex_tuple(");
    ret.push_back(","+itoa(_index)+")"); 
    return ret;
}

string Record::print()
{
    return "";
}


NewRecord::NewRecord(string a, string r, NodeList * f)
{
    atom = a;
    record = r;
    fields = f;
}

NewRecord::NewRecord(string r, NodeList * f)
{
    record = r;
    fields = f;
}

string NewRecord::print()
{
    return "";
}

#define MAX_FIELDS 256

list<string> NewRecord::emit_code()
{
    NodeList * rec;
    list<Node *>::iterator i;
    list<string> ret, nod;
    string outfields[MAX_FIELDS];
    int num, count = 0;

    rec = yy_program->find_record(record);
    if (!rec)
    {
        fprintf(stderr,"Unable to find record: %s\n", record.c_str());
        exit(1);
    }
    num = rec->nodes.size();

    ret.push_back("Emake_tuple(");
#ifdef _ALLOCS_H
    ret.push_back("tuplemalloc("+itoa(num)+")");
#else
    ret.push_back(itoa(num));
#endif
    if (num != 0) ret.push_back(",");

    // for named fields - work out where to emit them ..
    // for unnamed fields - in order emission
    //printf("emitting record %s (size %d)\n", record.c_str(), num);
    for (i = fields->nodes.begin(); i != fields->nodes.end(); i++)
    {
        Field * ele;
        string fld;
        int fldno;

        ele = (Field *) (*i);
        fld = ele->field();

        if (fld.empty())
        {
            outfields[count++] = (ele->expr())->print();
        }
        else
        {
            fldno = yy_program->find_record(record, fld);
            if (fldno == -1)
            {
                /* error */
                yyerrorf("Unknown field (%s) in record (%s)\n",
                                fld.c_str(), record.c_str());
            }
            else
            {
                list<string>::iterator i;
                string nodstr;
                nod = (ele->expr())->emit_code();
                for (i = nod.begin(); i != nod.end(); i++)
                    nodstr = nodstr + (*i);
                outfields[fldno] = nodstr;
            }
        }
    }

    // spit them out now
    for (count = 0; count < num; count++)
    {
        if (count != 0) ret.push_back(",");
        if (outfields[count].empty())
        {
            if (atom.empty() == false)
               ret.push_back("Eindex_tuple("+atom +","+itoa(count)+")");  
            else
               ret.push_back("0");
        }
        else
            ret.push_back(outfields[count]);
    }

    ret.push_back(")");
    return ret;
}


#if 0
Field::Field()
{
}
#endif

Field::Field(const string &f, Node * e)
{
    _field = f;
    _expr = e;
}

list<string> Field::emit_code()
{
    list<string> empty;
    return empty;
}

string Field::print()
{
    return _field;
}

