

//
// $Id: socket.cc,v 1.4 2002/10/08 06:51:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/socket.cc,v $
// $Revision: 1.4 $
// $Date: 2002/10/08 06:51:56 $
// $State: Exp $
//
// Author: Geoff Wong
//

#include "socket.h"
#include "str.h"

Socket::Socket()
{
    socket_no = 0;
    socket_type = 0;
    protocol = 0;
}
 
Socket::Socket(int domain, int tipe, int prot)
{
    socket_no = socket(domain, tipe, prot);
    socket_type = tipe;
    protocol = prot;
}
 
// NOTE: should we should TCP_NODELAY for TCP connections in here?
Socket * Socket::Accept()
{
    struct sockaddr_in addr;
    int             length, new_socket;
    Socket * o;
 
    /* make sure we're bound to a socket */
    if (socket_no == 0) return 0;
 
    /* can only accept on stream sockets */
    if (socket_type != SOCK_STREAM) return 0;
 
    /* accept connections */
    length = sizeof addr;
    new_socket = accept(socket_no,(struct sockaddr *) &addr, &length);

    if (new_socket > 0) 
    {
        struct protoent * prot;
        /* get a null socket */
        o = new Socket();     
        o->socket_no = new_socket;
        o->int_address = addr.sin_addr.s_addr;        /* set remote address */
        o->port = addr.sin_port;
        //      printf("finally got an accept\n"); 

        prot = getprotobyname("TCP");
        o->Setsockopt(prot->p_proto, TCP_NODELAY, 0);
//        o->Setsockopt(SOL_SOCKET, SO_REUSEADDR, 0))
//        { perror("setsockopt"); //    exit(1); }
        return o;
    }
    return 0;
}
 
/*
 * Bind to an address on the local machine
 * Internet bind
 */
int Socket::Bind(int port_number)
{
    struct sockaddr_in sin;        

    if (!Setsockopt(SOL_SOCKET, SO_REUSEADDR, 0))
    {
        perror("setsockopt");
        exit(1);
    }

    sin.sin_port = htons(port_number);
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
 
    if (bind(socket_no, (struct sockaddr *) &sin, sizeof sin) == -1)
    {
        Set_error("Bind failed");
//        printf("ERR: %d\n", errno);
        return 0;
    }
    return 1;
}
 
// Bind to a local socket
int Socket::Bind(const char * name)
{
    struct sockaddr_un sin;        
//    char            host_name[100];
    int len;
 
    memset((char *) &sin, '\0', sizeof sin);
    sin.sun_family = AF_UNIX;
    strcpy(sin.sun_path, name);
    len = strlen(sin.sun_path) + sizeof(sin.sun_family);
 
    if (bind(socket_no, (struct sockaddr *) &sin, len) == -1)
    {
        Set_error("Bind failed");
        // printf("ERR: %d\n", errno);
        return 0;
    }
    return 1;
}

int Socket::Close()
{
        // if it's a unix domain socket we should unlink too..
        return close(socket_no);
}
 
/*
 * Connect to a remote address
 */
int Socket::Connect(const char * address)
{
    int len;
        /* set up identification for server's inwards socket */
    if (socket_type == AF_UNIX) 
    {
        struct sockaddr_un server;
           server.sun_family = AF_UNIX;
        strcpy(server.sun_path, address);
        len = strlen(server.sun_path) + sizeof(server.sun_family);
 
             /* block waiting for a connection */
        if (connect(socket_no, (struct sockaddr *)&server, len) == -1) 
        {
//            perror("client connect()");
               return 0;
        }
        return 1;
    }
    else     // inet stuff
    {
        unsigned int port;
        char * loc;

        loc = index(address, ':');
        if (loc == NULL) return 0;
        port = atoi(&(loc[1]));    
        return Connect(address, port);
    }
}

/*
 * Connect to a remote inet address
 */
int Socket::Connect(const char * address, unsigned int port)
{
    char x;

    if (socket_type == AF_INET)
    {
        struct sockaddr_in sin;
 
        sin.sin_port = htons(port);
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = inet_addr(address);
        if (connect(socket_no, (struct sockaddr *) &sin, sizeof sin) == -1) 
        {
//            perror("connect");
            return 0;
        }
        return 1;
    }
    return 0;
}

/*
 * Connect to a remote inet address
 */
int Socket::Connect(long addr, unsigned int port)
{
    char x;

    if (socket_type == AF_INET)
    {
        struct sockaddr_in sin;
 
        sin.sin_port = htons(port);
        sin.sin_family = AF_INET;
        sin.sin_addr.s_addr = addr;
        if (connect(socket_no, (struct sockaddr *) &sin, sizeof sin) == -1) 
        {
//            perror("connect");
            return 0;
        }
        return 1;
    }
    return 0;
}

 
int Socket::Fileno()
{
    return 0;
}
 
char * Socket::Getpeername()
{
    return 0;
}
 
char * Socket::Getsockopt(int level, int optname, int buflen)
{
    return 0;
}
 
int Socket::Listen(int backlog)
{
    return listen(socket_no, backlog);
}
 
char * Socket::Recv(int bufsize)
{
        int x;
        char * buf, * ret;      
        buf = (char *) new char[bufsize + 1];
        x = recv(socket_no, buf, bufsize, 0);
        if (x == -1) 
        {
                delete [] buf;
                return 0;
        }
        buf[x] = 0;
        ret = string_copy(buf);
        delete [] buf;
        return ret;
}
 
/*
 * return # of characters sent (or -1 on failure)
 * /int send(char * what, int flags) - for out of band data?
 */
int Socket::Send(const char * what)
{
    int x;
        
    if (what == NULL) return 0;
    x = send(socket_no, what, strlen(what), 0);
    return x;
}
 
int Socket::Sendto(const char * what, int flags, int address)
{
    return 0;
}
 
int Socket::Setblocking(int flags)
{
    if (flags == 0) 
    {
#if 0
        if (ioctl(socket_no, FIONBIO, &tmp) == -1) {
//                perror("ioctl socket FIONBIO");
                return 0;
        }
#else                           /* sun */
        if (fcntl(socket_no, F_SETFL, O_NDELAY) == -1) {
//                perror("ioctl socket FIONBIO");
                return 0;
        }
#endif                          /* sun */
        return 1;
    }
    else  
    {
        return 0;
    }
}
 
int Socket::Setsockopt(int level, int optname, char * value)
{
//        char tmp[1];
    int tmp = 1;
    if (setsockopt(socket_no, level, optname, (char *)&tmp, sizeof(tmp))==-1) 
    {
            Set_error("setsockopt failed");
            return 0;
    }
    return 1;
}
 
int Socket::Shutdown(int how)
{
    return shutdown(socket_no, how);
}
 
 
void Socket::Set_error(const char * how)
{
//        printf("Socket error: %s (%d)\n", how, errno);
    if (error) delete [] error;
    error = string_copy(how);
}
 
char * Socket::Query_error()
{
    return error;
}

char * Socket::Ip_number()
{
    struct in_addr xx;
    xx.s_addr = int_address;
    if (socket_no == 0) 
        return string_copy("");
    return string_copy(inet_ntoa(xx));
}
 
 
#if 0
/*
 * Note: if the address char * is "a.b.c.d" the address number is a * 256^3 +
 * b * 256^2 + c * 256 + d
 * 
 */
 
char * ip_number()
{
    int         a, b, c, d;
    char *      addr;   
 
    if (!socket_no) return "";
 
    d = int_address & 0xff;
    c = (int_address & 0xff00) >> 8;
    b = (int_address & 0xff0000) >> 16;
    a = (int_address & 0xff000000) >> 24;
    addr = a+"."+b+"."+c+"."+d;
    return addr;
}
#endif                          /* INET_NTOA_OK */
 
#define HOST_BUFF 1024
char * Socket::Hostname()
{
    static char     name[HOST_BUFF];
 
    gethostname(name, sizeof name);
    name[sizeof name - 1] = '\0';       /* Just to make sure */
    return string_copy(name);
}

