
//
// $Id: str.cc,v 1.14 2002/07/08 00:35:42 geoff Exp $
//
// $Source: /home/cvs/gerl/src/str.cc,v $
// $Revision: 1.14 $
// $Date: 2002/07/08 00:35:42 $
// $State: Exp $
//
// Author: Geoff Wong
// Note: Standard Zikzak library code.
//


#include <algorithm>
#include <ctype.h>
#include <sys/stat.h>

#include "str.h"

#define MAXTMP 256 
#define MAX_LINE 4096


char * string_copy(const char * s)
{
	char * ret;
    ret = new char[(strlen(s)+1)*sizeof(char)];
	strcpy(ret,s);
	return ret;
}

//
// Note: 256 should be long enough for any integer we're printing! 
// 
string itoa(int x)
{
	char tmp[MAXTMP];
#ifdef WIN32
	sprintf(tmp, "%d", x);
#else
	snprintf(tmp, MAXTMP-1, "%d", x);
#endif
	string out(tmp);
	return out;
}

//
// 256 should be long enough for any integer we're printing! 
// Note: this is a nasty hack - we should use OS provided functions 
// 
string itodot(int x)
{
	/* assume that x is in network order. */
	char tmp[MAXTMP];
	int a,b,c,d;

	a = x & 0xff000000;
	b = x & 0x00ff0000;
	c = x & 0x0000ff00;
	d = x & 0x000000ff;

#ifdef WIN32
	sprintf(tmp, "%d.%d.%d.%d", a,b,c,d);
#else
	snprintf(tmp, MAXTMP-1, "%d.%d.%d.%d", a,b,c,d);
#endif
	string out(tmp);
	return out;
}

//
// itoa - pad it to a certain width with '0's.
// 
string itoa(int x, int width, char pad)
{
	char tmp[MAXTMP];
	int i, l;

	if (width > MAXTMP) width = MAXTMP;

#ifdef WIN32
	sprintf(tmp, "%*d", width, x);
#else
	snprintf(tmp, MAXTMP-1, "%*d", width, x);
#endif

	l = strlen(tmp);
	for (i = 0; i < l; i++)
	{
		if (tmp[i] == ' ') tmp[i] = pad;
	}

	string out(tmp);
	return out;
}

//
//
//
string randomstr(int width)
{
    char * tmp = new char[width];
    memset(tmp, width, sizeof(char));

    string retc;

    for (int i = 0; i < width; i++)
    {
        tmp[i] = (char)(random()%254 + 1);
    }
    retc = tmp;

	delete [] tmp;
    return retc;
}

//
// Name:    split
// Purpose: Split a string into elements
// Note: Separator is extracted.
//
int split(string instr, string * outarr, int max, string sep)
{
    int i = -1, oldi = 0;
    int count = 0;
    int len = sep.length();
    int len2 = instr.length();
    string tmpcord;

    if (instr.empty() == true) return 0;

    while ( (oldi < len2) 
        && ((i = (int)instr.find(sep, oldi)) != -1) && (count < (max-1)))
    {
        tmpcord = instr.substr(oldi, i-(oldi));
        outarr[count] = tmpcord;
        count++;
        // outarr[count++] = instr.substr(oldi+1, i);
        oldi = i+len;
    }

    // urgh
    outarr[count++] =  instr.substr(oldi, instr.length());
    return count;
}

//
//
//
int gsub(string& str, string& pat, string& repl)
{
#ifdef FreeBSD
	return str.gsub(pat, repl);
#else
	int x = 0;
	int count = 0;

	x = str.find(pat, 0);
	while (x != -1)
	{
		count++;
		str.replace(x, pat.length(), repl);
		x = str.find(pat, x+1);
	}
	return count;
#endif
}


#ifdef WIN32

int regfind(string& str, string& pat)
{
	// Unimplemented
	return 0;
}

int reggsub(string& str, string& regpat, string& repl)
{
	// Unimplemented
	return 0;
}

#else
int reggsub(string& str, string& regpat, string& repl)
{
#ifdef FreeBSD
	Regex rstr = regpat.c_str();
	return str.gsub(rstr, repl);
#else
	int count = 0;
    int status;
	int len;
	int x = 0;
    regex_t   re;
	regmatch_t pmatch[1];
	const char * cstr;

    if (regcomp(&re, regpat.c_str(), REG_EXTENDED) != 0) 
	{
			/* report error? */
    		return(0);      
    }

	cstr = str.c_str();

	while (1)
	{
    	status = regexec(&re, &(cstr[x]), (size_t) 1, pmatch, 0);
    	if (status != 0) 
		{
    		break;
    	}
		if (pmatch[0].rm_so == -1) break;

		len = pmatch[0].rm_eo - pmatch[0].rm_so;
		str.replace(pmatch[0].rm_eo, len, repl);
		x = pmatch[0].rm_so + 1;
		count++;
	}

    regfree(&re);

    return count;
#endif
}


int regfind(string& str, string& pat)
{
    int status;
    regex_t   re;
	regmatch_t pmatch[1];
	const char * cstr;

    if (regcomp(&re, pat.c_str(), REG_EXTENDED) != 0) 
	{
			/* report error? */
    		return -1;      
    }

	cstr = str.c_str();
   	status = regexec(&re, cstr, (size_t) 1, pmatch, 0);
    if (status != 0) 
	{
   		return -1; 
    }

	return pmatch[0].rm_so;
}

#endif

string strip_whitespace(string str, string whitespace = " \t\n")
{
    string ret;
    char * buf = (char *)malloc((str.length() + 1) * sizeof(char));
    int i, count = 0;

    for (i = 0; i < (int)str.length(); i++)
    {
        if ((int)whitespace.find(str[i]) == -1)
            buf[count++] = str[i];
    }
    buf[count] = '\0';

    ret = buf;
    free(buf);
    return ret;
}


// Author: Zik Saleeba
string strip_leading_whitespace(string str)
{
    int Count;
    
    for (Count = 0; Count < (int)str.length() && isspace(str[Count]); Count++)
    {};
    
    if (Count >= (int)str.length())
        return "";
    else
        return str.substr(Count);
}


// Author: Zik Saleeba
string strip_trailing_whitespace(string str)
{
    int Count;
    
    for (Count = str.length()-1; Count >= 0 && isspace(str[Count]); Count--)
    {};
    
    if (Count < 0)
        return "";
    else
        return str.substr(0, Count+1);
}


// Author: Zik Saleeba
string strip_surrounding_whitespace(string str)
{
    return strip_trailing_whitespace(strip_leading_whitespace(str));
}


// Author: Zik Saleeba
string lowercase(const string& instr)
{
    string InStr = instr;
    transform(InStr.begin(), InStr.end(), InStr.begin(), tolower);
    return InStr;
}


// get a line from a file
bool get_line(FILE *fp, string &result)
{
    char buf[MAX_LINE+1];

    if (fp == NULL)
        return false;

    buf[MAX_LINE] = '\0';
    if (fgets(buf, MAX_LINE, fp) == NULL)
        return false;

    result = (string)buf;
    
    return true;
}


// get a line from a cord
bool get_line(const string &source, string &result, int &pos)
{
    int Count;
    int Length = source.length();
    if (pos >= Length)
        return false;

    // search for the end of a line
    for (Count = pos; Count < Length; Count++)
    {
        char ThisChar = source[Count];
        
        if (ThisChar == '\n')
        {
            result = source.substr(pos, Count-pos);
            pos = Count+1;
            return true;
        }
        else if (ThisChar == '\r')
        {
            if (Count < Length-1)
            {
                // check forward one char for cr/lf
                if (source[Count+1] == '\n')
                    Count++;
            }

            result = source.substr(pos, Count-pos);
            pos = Count+1;
            return true;
        }
    }
    
    // up to the end
    result = source.substr(pos, Length-pos);
    pos = Length;
    return true;
}



//
// NAME:        get_part
// ACTION:      for reading a cord in parts separated by a string
// PARAMETERS:  string &result - the result is placed here
//              int &pos - keeps track of the position
//              string str - the separator
// RETURNS:     true if a line was fetched
//

bool get_part(const string &source, string &result, int &pos, string str)
{
    int Length = source.length();
    if (pos >= Length)
        return false;
        
    int FoundAt = source.find(str, pos);
    
    if (FoundAt >= 0)
    {
        result = source.substr(pos, FoundAt-pos);
        pos = FoundAt + str.length();
    }
    else
    {
        // at end
        result = source.substr(pos, Length-pos);
        pos = Length;
    }
    
    return true;
}



//
// NAME:        get_part
// ACTION:      for reading a cord in parts separated by a character
// PARAMETERS:  string &result - the result is placed here
//              int &pos - keeps track of the position
//              char c- the separator
// RETURNS:     true if a line was fetched
//

bool get_part(const string &source, string &result, int &pos, char c)
{
    int Length = source.length();
    if (pos >= Length)
        return false;
        
    int FoundAt = source.find(c, pos);
    
    if (FoundAt >= 0)
    {
        result = source.substr(pos, FoundAt-pos);
        pos = FoundAt+1;
    }
    else
    {
        // at end
        result = source.substr(pos, Length-pos);
        pos = Length;
    }
    
    return true;
}


// Purpose: turn a buffer of range 0-255 to a cord
// with characters of ascii range 65-97
string binary2cord(unsigned char * buf, int len)
{
    char * newbuf;
    string ret;
    int i, c = 0;
    int low, high;

    newbuf  = (char *)malloc((2*len+1) * sizeof(char));

    for (i = 0; i < len; i++)
    {
        low = (buf[i] & 0xf0) >> 4;
        high = buf[i] & 0x0f;
        newbuf[c++] = low + 65;
        newbuf[c++] = high + 65;
    }

    newbuf[c++] = '\0';
    ret = newbuf;

    free(newbuf);

    return ret;
}


// read an entire file in one hit
bool read_file(string FileName, string &Result)
{
    struct stat FileStat;
    char *Buf;
    FILE *InFile;
    const char *FName = FileName.c_str();
    
    // get the file length
    if (stat(FName, &FileStat) < 0)
        return false;
    
    Buf = (char *)malloc(FileStat.st_size+1);
    
    // read it
    InFile = fopen(FName, "r");
    if (InFile == NULL)
        return false;
    
    if ((int)fread(Buf, 1, FileStat.st_size, InFile) != FileStat.st_size)
    {
        fclose(InFile);
        return false;
    }
    fclose(InFile);
    
    // set it as a cord
    Buf[FileStat.st_size] = '\0';
    Result = Buf;
    free(Buf);
    return true;
}


// get the right hand part of a string
string string_right(const string &source, int length)
{
    int SrcLength = source.length();
    if (length <= SrcLength)
        return source.substr(SrcLength - length);
    else
        return source;
}


// get the left hand part of a string
string string_left(const string &source, int length)
{
    return source.substr(0, length);
}

string join(string * inarr, int max, string sep)
{
    int i;
    string ret;

    for (i = 0; i < max; i++)
    {
        if (i != 0)
            ret = ret + sep + inarr[i];
        else
            ret = inarr[i];
    }
    return  ret;
}

string join(list<string>& inlist, string sep)
{
    list<string>::iterator i;
    string ret;

    for (i = inlist.begin(); i != inlist.end(); i++)
    {
        if (i != inlist.begin())
            ret = ret + sep + (*i);
        else
            ret = (*i);
    }

    return ret;
}


string topntail(const string& str)
{
    string ret, whitespace = " \t\n";
    char * buf = (char *)malloc((str.length() + 1) * sizeof(char));
    int i = 0, count = 0, start, end;

    while (isspace(str[i])) i++;

    start = i;

    i = str.length() - 1;
    while (isspace(str[i])) i--;
    end = i;

    for (i = start; i <= end; i++)
    {
        buf[count++] = str[i];
    }

    buf[count] = '\0';
    ret = buf;
    free(buf);
    return ret;
}

list<string> string2list(const string& str)
{
    list<string> ret;
    ret.push_front(str);
    return ret;
}

