
//
// String Support Stuff
//
// $Id: str.h,v 1.11 2004/09/05 02:42:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/str.h,v $
// $Revision: 1.11 $
// $Date: 2004/09/05 02:42:56 $
// $State: Exp $
//
// Author: Geoff Wong
// Note: Standard Zikzak library code.
// Copyright (C) 1999 Geoff Wong.
//
// Provides a miscellaneous collection of string
// support functions which are sometimes useful.
//

#ifndef _ISTRING_H
#define _ISTRING_H

#include <unistd.h>
#include <regex.h>
#include <stdio.h>
#include <sys/types.h>
#include <string>
#include <list>

using namespace std;

char * string_copy(const char *);
    // Purpose: allocate memory and copy a C string, essentially strdup()
    // Returns: the newly allocated string

string itoa(int x);
    // Purpose: convert a number to a C++ string
    // Returns: a string containing the number

string itoa(int num, int width, int pad);
    // Purpose: convert a number to a C++ string

string itodot(int);
	// Purpose: Convert an IP number to dot notation (C++ string).

string randomstr(int width);
    // Purpose: return a C++ string full of random garble
    // Returns: a string for of funny chars.
    
string lowercase(const string& instr);
    // Purpose: transform a string which may include uppercase letters
    //      entirely into lowercase.
    // Returns: the new string.

int split(string, string *, int max, string sep);
    // Purpose: split a string into an array of string given a
    //      separator. All C++ string.
    //     2nd arg is an array of string (size max) which the "split"
    //     items will be written into
    // Returns: a value indicating the number of items split into

string join(string * inarr, int max, string sep);
    // Purpose: "reverse" of split(). Joins a number of strings in
    //  an array of cords with a given separator.
    // Returns: the joined string.
    
string join(list<string>&, string sep);
    // Purpose: reverse of split using a list
    // Returns: the joined string.

int gsub(string& str, string& pat, string& repl);
	// Purpose: global substitution of patterns found in a string
    // Returns: the number of substitutions made

int reggsub(string& str, string& regpat, string& repl);
    // Purpose: global substitution of patterns from in a string
    //      given a regular expression to replace.
    // Returns: the number of substitutions made

int regfind(string& str, string& pat);
    // Purpose: return the index of a character in a string given
    //      a regular expression to match.
    
string strip_whitespace(string str, string optws);
    // Purpose: strip any whitespace chars from the in string.
    // Returns: the stripped input string.

string strip_leading_whitespace(string);
    // Purpose: strip any leading whitespace chars from the in string.
    // Returns: the stripped input string.

string strip_trailing_whitespace(string);
    // Purpose: strip any trailing whitespace chars from the in string.
    // Returns: the stripped input string.

string strip_surrounding_whitespace(string);
    // Purpose: strip any leading or trailing whitespace chars from the in string.
    // Returns: the stripped input string.

string binary2cord(unsigned char * buf, int len);
    // Purpose: turn a buffer of range 0-255 to a cord
    //      with characters of ascii range 32-159

string topntail(const string& str);
    // Purpose: return leading and trailing whitespace.

// get a line from a file
bool get_line(FILE *fp, string &result);

// get a line from a cord
bool get_line(const string &source, string &result, int &pos);
bool get_part(const string &source, string &result, int &pos, string str);
bool get_part(const string &source, string &result, int &pos, char c);

// read an entire file in one hit
bool read_file(string FileName, string &Result);

// get the right hand part of a string
string string_right(const string &source, int length);
string string_left(const string &source, int length);

list<string> string2list(const string& str);

#endif
