

//
// $Id: symbol.cc,v 1.23 2002/09/08 06:54:31 geoff Exp $
//
// $Source: /home/cvs/gerl/src/symbol.cc,v $
// $Revision: 1.23 $
// $Date: 2002/09/08 06:54:31 $
// $State: Exp $
//
// Author: Geoff Wong
//


/*
 * SymbolTable class
 * 
 * PURPOSE: maintain information about a particular symbol
 *
 */

#include <stdio.h>
#include "symbol.h"
#include "namer.h"
#include "erlib.h"

#define DEBUG 0
extern int debug;

//	S_Illegal, S_Function, S_Variable, S_SymbolTable

Symbol::Symbol()
{
	_symbol = "X";
	symboltype = S_Illegal;
	typesig = "";
	_params = 0;
    _used = 0;
    _type = 0;
    _destructive = false;
}

Symbol::~Symbol()
{
}

Symbol::Symbol(const string& s, SymbolType t, string& ts, int noparams)
{
	symboltype = t;
	switch (t)
	{
		case S_Function:
			_symbol = "F" + s;
			break;
		case S_Variable:
			_symbol = "V" + s;
			break;
		case S_Clause:
			_symbol = "C" + s;
            break;
		case S_Boolean:
			_symbol = "B" + s;
            break;
		default:
			_symbol = "I" + s;
			break;
	}
	typesig = ts;
	_params = noparams;
    _destructive = false;
    _type = 0;
    _used = 0;
}

string Symbol::symbol()
{
	string ret;
	ret = _symbol.substr(1);
	return ret;
}

string Symbol::full_symbol()
{
	return _symbol;
}

void Symbol::symbol(const string& newsym)
{
    // X - no type set, used to abuse symbol by EList
    _symbol = "X" + newsym;
}

string Symbol::parameter()
{
	return param;
}

void Symbol::setparameter(const string& str)
{
	param = str;
}

int Symbol::params()
{
	return _params;
}

void Symbol::params(int x)
{
	_params = x;
}

string Symbol::module()
{
    return _module;
}

void Symbol::module(const string& mod)
{
    _module = mod;
}

void Symbol::used(int x)
{
    _used = x;
}

int Symbol::used()
{
    return _used;
}

void Symbol::type(int x)
{
    if (debug)
        fprintf(stderr, "ST: %s type=%d\n", symbol().c_str(), x);
    _type = x;
}

int Symbol::type()
{
    return _type;
}

Node * Symbol::associated()
{
    return _associated;
}

void Symbol::associated(Node * n)
{
    _associated = n;
}

bool SymbolTable::add(Symbol * s)
{
	if (Table[s->full_symbol().c_str()] == NULL)
	{
        if (debug)
            fprintf(stderr, "ST: adding (%s)\n", s->full_symbol().c_str());

		Table.add_element(s->full_symbol().c_str(), s);

#if 0
if (Table[s->full_symbol().c_str()] == NULL)
			fprintf(stderr, "ST: failed to add (%s)\n", s->full_symbol().c_str());
#endif

        if (_chain) _chain->add(s);

		return true;
	}
	else 
	{
#if 0
	    fprintf(stderr, "ST: false add (%s)\n", s->_symbol.c_str());
#endif
		return false;
	}
}

string SymbolTable::add_temp()
{
    Symbol * s;
    string tmpval, ts;

    tmpval = "_tx" + itoa(tmp_count++);

    s = new Symbol(tmpval, S_Variable, ts);
    add(s);

    return tmpval;
}

Symbol * SymbolTable::add_freetemp()
{
    Symbol * s;
    string tmpval, ts;

    tmpval = "_fx" + itoa(tmp_count++);

    s = new Symbol(tmpval, S_Variable, ts);
    // s->type(LT_normal | LT_on_heap);
    add(s);

    return s;
}

Symbol * SymbolTable::find(string& s, SymbolType t)
{
	string tmp;
	switch (t)
	{
		case S_Function:
			tmp = "F" + s;
			break;
		case S_Variable:
			tmp = "V" + s;
			break;
		case S_SymbolTable:
			tmp = "T" + s;
			break;
		case S_Clause:
			_symbol = "C" + s;
            break;
		case S_Boolean:
			_symbol = "B" + s;
            break;
		default:
			tmp = "I" + s;
			break;
	}
	if (Table[tmp.c_str()] == NULL)
	{
        if (debug)
            fprintf(stderr, "ST: NOT found (%s)\n", tmp.c_str());
	}
	return Table[tmp.c_str()];
}

SymbolTable::SymbolTable()
{
    _chain = NULL;
	_symbol = "";
	symboltype = S_Illegal;
	typesig = "";
    tmp_count = 0;
}

SymbolTable::SymbolTable(string& s)
{
    _chain = NULL;
	_symbol = "T" + s;
	symboltype = S_SymbolTable;
	_params = 0;
    tmp_count = 0;
}

SymbolTable::SymbolTable(string& s, SymbolType st, string& ts, int par)
{
    _chain = NULL;
	_symbol = "T" + s;
	symboltype = S_SymbolTable;
	switch (st)
	{
		case S_Function:
			_symbol = "F" + s;
			break;
		case S_Variable:
			_symbol = "V" + s;
			break;
		case S_SymbolTable:
			_symbol = "T" + s;
			break;
		case S_Clause:
			_symbol = "C" + s;
            break;
		case S_Boolean:
			_symbol = "B" + s;
            break;
		default:
			_symbol = "I" + s;
			break;
	}
//	printf("ST: %s\n", _symbol.c_str());
	typesig = ts;
	_params = par;
}

string SymbolTable::name()
{
    return _name;
}

void SymbolTable::name(const string& str)
{
    _name = str;
}

void SymbolTable::chain(SymbolTable * st)
{
    _chain = st;
}

string SymbolTable::full_symbol()
{
    return Symbol::full_symbol();
	return Namer::semi_mangle(_symbol,_params);
}

SymbolTable * SymbolTable::chain()
{
    return _chain;
}

//HACK
int operator==(const Dictionary<Symbol *>::iterator& it1,
                               const Dictionary<Symbol *>::iterator& it2)
{
       if ((it1.where == it2.where)
                && (it1.ihash == it2.ihash)) return 1;
       return 0;
}
                                
