//
// $Id: symbol.h,v 1.17 2004/09/05 02:42:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/symbol.h,v $
// $Revision: 1.17 $
// $Date: 2004/09/05 02:42:56 $
// $State: Exp $
//
// Author: Geoff Wong
//

#ifndef _SYMBOL_H
#define _SYMBOL_H

#include <list>
#include "expr.h"
#include "dictionary.h"
#include "str.h"

//
// Types of symbols in Erlang.
//
enum SymbolType
{
    S_Illegal, 
    S_Function, 
    S_Variable, 
    S_SymbolTable, 
    S_Clause, 
    S_Boolean, 
    S_Clash
};

class Node;

//
// Name:
//        Symbol
// Purpose:
//        Class Symbol represents a language symbol, it may be
//        a variable, a paramater, etc.
//        
class Symbol
{
protected:

    string param;                    
        // parameter assignment (if necessary)
    string _module;
        // if we're not "" then we're imported from elsewhere
    string _symbol;
        // our actually (modified) symbol

    int _used;

    int _type;
        // do I have a particular type?

    bool _destructive;
        // am I destructive to my variable?

    Node * _associated;
        // Am I associated with a particular node?

public:

    SymbolType symboltype;
    string typesig;    
    int _params;                    
        // number of parameters
    
    Symbol();
    Symbol(const string& s, SymbolType t, string& ts, int noparams = 0);
        // create a symbol with symbol name, symbol type, type sig, etc

    string symbol();
        // Return my (actual) identifier
        
    virtual string full_symbol();
        // Return my (modified) symbol
        
    void symbol(const string&);
        // A hack to allow direct setting of the symbol (ugh)

    string parameter();
        // Purpose: emit any parameter information set in setparameter
        //      used when emitting lists of variable initialisation
    int params();
        // Purpose: query # of parameters

    void params(int x);
        // Purpose: set # of parameters
    void setparameter(const string& str);

    string module();
        // Purpose: return the module I came from
    void module(const string& str);
        // Purpose: to set the module

    int used();
        // Is this symbol actually used?
    void used(int x);
        // Set the used flag

    bool destructive() { return _destructive; };
        // am I destructive to my variable?
    void destructive(bool x) { _destructive = x; }
        // set it ..
        
    Node * associated();
        // returns an associated Node

    void associated(Node *);
        // sets an associated Node ..

    int type();
        // return my type
    void type(int x);
        // set my type

    virtual ~Symbol();

};


//
// Name:
//        SymbolClash
// Purpose:
//        To have a list of Symbols which clash with the same name
//        in a symbol table.
//
class SymbolClash : public Symbol
{
public:
    list<Symbol *> clashes;

    SymbolClash();
    SymbolClash(Symbol *, Symbol *);    
        // start with two clashing symbols
    
    bool add(Symbol *);
        // a another clashing symbol in

    int size();
        // return number of clashing parameters
};


//
// Name:
//        SymbolTable
// Purpose:
//        A SymbolTable is used to handle symbols within a particular
//        scope. SymbolTable itself is a Symbol so scopes can be nested.
//
class SymbolTable : public Symbol
{
    string _name;
        // actual clause name (if we have one)
        // separate from our id#

    SymbolTable * _chain;

    int tmp_count;
        // temporary symbol counter

public:
    Dictionary <Symbol *> Table;

    SymbolTable();
    SymbolTable(string& s);
    SymbolTable(string& s, SymbolType t, string& ts, int noparams = 0);
        // Constructors

    bool add(Symbol * s);
        // Add a symbol into the symboltable

    string add_temp();
        // Add a temporary symbol (and return it);
        // should return a Symbol *?

    Symbol * add_freetemp();
        // Add a temporary symbol that will need freeing (and return it);

    Symbol * find(string& s, SymbolType t);
        // returns 0 is it doesn't exist

    string name();
        // Purpose return the function name

    void name(const string& str);
        // Purpose return the function name

    void chain(SymbolTable * t);
        // Chained to another symbol table? Additions go to both tables
        // this one and the chained one

    string full_symbol();
        // Return my (modified) symbol

    SymbolTable * chain();

};

//
// Operator to compare two iterators from a SymbolTable
//
extern
int operator==(const Dictionary<Symbol *>::iterator& it1,
                               const Dictionary<Symbol *>::iterator& it2);

#endif

