//
// $Id: tuple.cc,v 1.23 2002/08/03 01:53:15 geoff Exp $
//
// $Source: /home/cvs/gerl/src/tuple.cc,v $
// $Revision: 1.23 $
// $Date: 2002/08/03 01:53:15 $
// $State: Exp $
//
// Copyright (C) Geoff Wong (1996, 1997, 1998)
// geoff@cs.rmit.edu.au
//
// This file is distributed under the terms of the
// GNU Public License, details of which can be found
// in the file "Copying" which accompanies this code.        
//


#include "program.h"
#include "tuple.h"

extern int debug;
extern Program * yy_program;

Tuple::Tuple(const string& s)
{
    storage = s;
    vlist = NULL;
}

Tuple::Tuple(const string& s, NodeList * l)
{
    storage = s;
	vlist = l;
}

list<string> Tuple::emit_code()
{
    string pre, RHS, ret;
    list<string> lret, code;
    list<Node *>::iterator i;
    Symbol * assocsym, * s;

    if (vlist == NULL)
    {
#ifdef _ALLOCS_H
        ret = ret + "(tuplealloc(" + storage + "," + itoa(0) + "),";
#endif
        ret = ret + "Emake_tuple("+itoa(0)+")";
#ifdef _ALLOCS_H
        ret = ret + ")";
#endif
        return string2list(ret);
    }

#ifdef _ALLOCS_H
    lret.push_back("(tuplealloc(" + storage + "," +
                    itoa(vlist->nodes.size()) + "),");
    lret.push_back("Emake_tuple(" + storage);
#else
    lret.push_back("Emake_tuple(" + itoa(vlist->nodes.size()));
#endif
    if (vlist->nodes.size())
    {
        ret = ",";
        for (i = vlist->nodes.begin(); i != vlist->nodes.end(); i++)
        {
            if (i != vlist->nodes.begin())
            {
                    ret = ret + ",";
            }
            assert((*i) != NULL);

            // should also emit EValcopy() for appropriate types
            // set assocaited freesym so it's not freed..
            // perhaps we should have a separate LT_returned flag?
            // assocsym->type(assocsym->type() & ~LT_on_heap);
            (*i)->freesym(NULL);
            
            add_pre_code((*i)->pre_code());
            RHS = (*i)->expr_code();

            if ((((*i)->type() & LT_on_heap))
                            || (((*i)->type() & LT_anyvar) < LT_list 
                                    && ((*i)->type() & LT_anyvar) > LT_none))
            {
                ret = ret + RHS;
            }
            else
            {
                ret = ret + "EValcopy("+RHS+")";
            }
        }
    }
    lret.push_back(ret);
#ifdef _ALLOCS_H
    lret.push_back("))");
#else
    lret.push_back(")");
#endif
    return lret;
}

string Tuple::print()
{
	return "{ " + vlist->print() + " }";
}


//
// Name: emit_pattern
// Purpose: set up indexing patterns for tuple - does not get emitted directly
//
string Tuple::emit_pattern(string p)
{
	string ret;
    list<Node *>::iterator i;
	//int count = vlist->size() - 1;
	int count = 0;

    if (debug)
    {
        fprintf(stderr, "Tuple:emit_pattern %s (type=%d)\n", p.c_str(),
            type());
    }

    for (i = vlist->nodes.begin(); i != vlist->nodes.end(); i++)
    {
#if 0
        if (count != 0) 
        {
            ret = "," + ret ;
        }
#endif
        ret = ret + emit_typecheck(p, LT_tuple, 1, yy_program->Module());
        if (type() & LT_returned)
        {
		    ret = (*i)->emit_pattern("Ecindex_tuple("+  p + "," +
					itoa(count) + ")"); // + ret;
        }
        else
        {
		    ret = (*i)->emit_pattern("Eindex_tuple("+  p + "," +
					itoa(count) + ")"); // + ret;
        }
		count++;
	}
#if 0
    fprintf(stderr, "Tuple::emit_pattern result(%s)\n", ret.c_str());
#endif
//    ret = "(" + ret + ")";
	return ret;
}

void Tuple::add_symbol(SymbolTable * st, int df)
{
    if (st == NULL) return;

    if (debug)
    {
        fprintf(stderr, "Tuple::add_symbol (type=%d)\n", df);
    }
    // a little hack here..
    if (df & LT_returned) 
    {
        type(type() | LT_returned);
    }

	if (vlist != NULL) vlist->add_symbol(st, df);
}

