
/*
** $Id: type_sup.c,v 1.1 1998/07/09 01:52:32 geoff Exp $
**
** $Source: /home/cvs/gerl/src/type_sup.c,v $
** $Revision: 1.1 $
** $Date: 1998/07/09 01:52:32 $
** $State: Exp $
**
** Copyright (C) Geoff Wong (1995, 1996, 1997, 1998)
** geoff@cs.rmit.edu.au
**
** This file is distributed under the terms of the
** GNU Public License, details of which can be found
** in the file "Copying" which accompanies this code.        
**
** Internal compiler type support stuff.
**
**  A word on the type hacks below
**  o designed to minimise changes the rest of the system
**    (this will be fixed in due course).
**  o Assume 32 bit ints.
**  top 8 bits struct no     (max 255 structs)
**  next 8 bits        (pointer depth max 255)
**  next 8 bits        (special types static, private etc)
**  next 8 bits        (basic types)
*/

#include "proto.h"
#include "gecko.h"
#include "type_sup.h"
#include <assert.h>

static int current_unknown = 1;

Compile_type
type_new(int type, int depth, int constant)
{
	Compile_type ret;
	ret = (Compile_type) malloc(sizeof(struct COMPILE_TYPE));
	ret->flag = CT_basic;
	ret->type.basic = type;
	ret->depth = depth;
	ret->constant = constant;
	ret->unknown = 0;
	ret->misc = 0;
	ret->resolve = 0;
	return ret;
}

Compile_type
type_unknown(EmitList param, EmitList utype)
{
	Compile_type ret;
	ret = (Compile_type) malloc(sizeof(struct COMPILE_TYPE));
	ret->flag = CT_unknown;
	ret->type.basic = 0;
	ret->depth = 0;
	ret->constant = 0;
	ret->unknown = current_unknown++;
	ret->misc = 0;
	ret->resolve = add_resolve_list(make_resolve_list(param, utype), 0);
	return ret;
}

Compile_type
type_copy(Compile_type t)
{
	Compile_type ret;
    
    assert(t != NULL);

	ret = (Compile_type) malloc(sizeof(struct COMPILE_TYPE));
	ret->flag = t->flag;
	ret->type = t->type;
	ret->depth = t->depth;
	ret->constant = t->constant;
	ret->unknown = t->unknown;
	ret->misc = t->misc;
	ret->resolve = t->resolve;
	return ret;
}

int
resolve_types(Compile_type t1, Compile_type t2, int restrict, char * where)
{
	Compile_type x;
	if (!t2) t2 = t1;
	if (t2->flag == CT_unknown) {
		x = t1; t1 = t2; t2 = x;
	}
	if (!restrict) restrict = T_ANY;
	if (t1->depth != t2->depth) {
		yyerrorf("Arrays different depths in %s(%d,%d)\n", where,
			t1->depth, t2->depth);
		return 0;
	}
	if (t1->flag != CT_unknown && t2->flag != CT_unknown && 
			t1->flag != t2->flag) {
		yyerrorf("Types (totally) incompatible in %s\n", where);
		return 0;
	}
	switch (t1->flag) {
		case CT_basic:
			if ((t1->type.basic & t2->type.basic) && 
				(t1->type.basic & restrict)) return 1;
			break;
		case CT_class:
			if (t1->type.class == t2->type.class) return 1;
			break;
		case CT_unknown:
			if (t2->flag == CT_unknown) {
				t1->unknown = t2->unknown;
				add_resolve_list(t1->resolve, t2->resolve);
				t2->resolve = t1->resolve;
				return 1;
			}
			t1->flag = t2->flag;
			t1->type = t2->type;
			t1->depth = t2->depth;
			t1->unknown = 0;
			resolve_resolve_list(t1);
			t1->resolve = 0;
			t2->resolve = 0;
			return 1;
		default:
			fatal("Illegal internal compiler type encountered.\n");
			return 0;
	}
	yyerrorf("Types don't match in %s\n", where);
	return 0;
}

/*  	A checking function */
int
basic_type(struct COMPINFO * ci, int x)
{
	Compile_type t1 = ci->type;
	if (t1->flag != CT_basic) return 0; 
	if (t1->type.basic & x && t1->depth == 0) return 1;
	return 0;
}

/* 
	Convert internal type information into
	a string representation (handy storage)
*/
char * make_typesig()
{
	return 0;
}

char *make_type_name(int t)
{
	if (ARRAY_DEPTH(t) > 0) return string_copy("Array");
        switch (ARRAY_TYPE(t)) {
        case T_VOID:
                return string_copy("void");
        case T_NUMBER:
                return string_copy("int");
        case T_REAL:
                return string_copy("float");
        case T_STRING:
                return string_copy("String");
        case T_OBJECT:
                return string_copy("Obj");
        default:
                fatal("Unknown type in make_type_name()\n");
                exit(1);
        }
}

char *make_itype_name(int t)
{
	if (ARRAY_DEPTH(t) > 0) return string_copy("array");
        switch (ARRAY_TYPE(t)) {
        case T_VOID:
                return string_copy("nothing");
        case T_NUMBER:
                return string_copy("number");
        case T_REAL:
                return string_copy("real");
        case T_STRING:
                return string_copy("string");
        case T_OBJECT:
                return string_copy("ob");
        default:
                fatal("Unknown type in make_itype_name()\n");
                exit(1);
        }
}


char *make_ctype_name(Compile_type t)
{
	if (t->depth > 0) return string_copy("Array");
    switch (t->type.basic) {
        case T_VOID:
                return string_copy("void");
        case T_NUMBER:
                return string_copy("int");
        case T_REAL:
                return string_copy("float");
        case T_STRING:
                return string_copy("String");
        case T_OBJECT:
                return string_copy("Obj");
        default:
                fatal("Unknown type in make_ctype_name()\n");
                exit(1);
    }
}


char *type_to_str(Compile_type t)
{
    if (t->depth > 0) return string_copy("A");
    switch (t->type.basic) 
    {
        case T_NUMBER:
                return string_copy("i");
        case T_REAL:
                return string_copy("f");
        case T_STRING:
                return string_copy("s");
        case T_OBJECT:
                return string_copy("o");
        default:
                fatal("Unknown type (%d) in type_to_str()\n", t->type.basic);
                exit(1);
    }
}

/*
** Name: demangle_params
** Purpose: produce a string for the types in the (mangled) parameter list
** Notes: static return string.
*/
#define MAX_PARAML 65536
char * demangle_params(Func pr, Class_state state)
{
    char buf[MAX_PARAML];
    char * typesig = pr->typesig;
    int i = 0, max;
    
    assert(pr != NULL);
    /* empty buf */
    buf[0] = '\0';
    if (state == C_class) strcpy(buf, "Obj");

    if (typesig == NULL) return buf;

    max = strlen(typesig);
    for (i = 0; i < max; i++)
    {
        if (i != 0 || state == C_class) strcat(buf, ",");
        switch (typesig[i])
        {
        case 'i':
                strcat(buf, "int");
                break;
        case 'f':
                strcat(buf, "float");
                break;
        case 's':
                strcat(buf, "String");
                break;
        case 'o':
                strcat(buf, "Obj");
                break;
        case 'A':
                strcat(buf, "Array");
                break;
        default:
                fatal("Unknown type (%c) in demangle_params()\n", pr[i]);
                exit(1);
        }
    }
    return buf;
}


/*
 * A bit of ugly stuff to resolve unknown types.
 */

Rlist make_resolve_list(EmitList param, EmitList utype)
{
	Rlist tmp;
	tmp = (Rlist) malloc(sizeof(struct RLIST));
	tmp->param = param;
	tmp->utype = utype;
	tmp->next = 0;
	return tmp;
}

/* add lists together */
Rlist add_resolve_list(Rlist a, Rlist b)
{
	Rlist x = a;

	assert(a != NULL);

	while (x->next) {
		x = x->next;
	}
	x->next = b;
	return a;
}

/* fix up param/union type emits */
void resolve_resolve_list(Compile_type t1)
{
	Rlist x = t1->resolve;

	assert(t1 != NULL);

	/* do the ugly stuff */
	while (x) {
		switch(t1->flag) {
		case CT_basic:		
			if (x->param && x->utype) {
				creplacef(x->utype, ".%s", make_itype_name(T_ARRAY(t1->type.basic, t1->depth)));
				creplacef(x->param, "%s", x->param->str, T_ARRAY(t1->type.basic, t1->depth));
			}
			break;
		case CT_class:
			printf("CT_class found in resolve_resolve_list ???\n");
			break;
		default:
			fatal("resolve_resolve_list failed with unknown resolution\n");
		}
		x = x->next;
	}
}

void clean_up_resolve(Compile_type t1)
{
	t1->type.basic = T_VOID;
	t1->depth = 0;
	t1->flag = CT_basic;
	resolve_resolve_list(t1);
}
